/* authentication.c - authentication routines */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/* 
 * $Header$
 *
 *
 * $Log$
 */

/*
 *			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#include <stdio.h>
#include "objects.h"
#include "party.h"
#include "md5.h"
#include "logger.h"


#define LENOFDIGEST	16

/*  */

struct qbuf *generate_md5_digest ();
char   *get_digest_from_string ();

/*  */

int	authenticate_incoming (privmsg, authmsg, srcParty, dstParty, reason)
struct type_SNMP_SnmpPrivMsg *privmsg;
struct type_SNMP_SnmpAuthMsg *authmsg;
struct snmp_party *srcParty,
		  *dstParty;
int    *reason;
{
    if (reason)
	*reason = AUTH_UNKNOWN;
    if (privmsg == NULL) {
	(void) sprintf (PY_pepy, "null privmsg");
	return NOTOK;
    }
    if (authmsg == NULL) {
	(void) sprintf (PY_pepy, "null authmsg");
	return NOTOK;
    }
    if (srcParty == NULL) {
	(void) sprintf (PY_pepy, "null srcParty");
	return NOTOK;
    }
    if (dstParty == NULL) {
	(void) sprintf (PY_pepy, "null dstParty");
	return NOTOK;
    }

    if (oid_cmp (privmsg -> privDst, authmsg -> authData -> dstParty)) {
	(void) sprintf (PY_pepy,
			"destination parties different in privacy (%s) and authentication ",
			oid2ode (privmsg -> privDst));
	(void) sprintf (PY_pepy + strlen (PY_pepy), "(%s) messages",
			oid2ode (authmsg -> authData -> dstParty));
	if (reason)
	    *reason = AUTH_MISMATCH;
	return NOTOK;
    }

    switch (srcParty -> party_authProt) {
	case BADAUTH:
	    (void) sprintf (PY_pepy,
			    "unable to determine authentication protocol for %s",
			    oid2ode (srcParty -> party_Identity));
	    return NOTOK;

        case NOAUTH:
	    if (authmsg -> authInfo -> offset != type_SNMP_AuthInformation_1) {
		(void) sprintf (PY_pepy,
				"wrong kind of authInformation present (%d)",
				authmsg -> authInfo -> offset);
		if (reason)
		    *reason = AUTH_WRONGAUTH;
		return NOTOK;
	    }
	    return OK;

	case MD5AUTHPROT:
	    return do_md5_auth (privmsg, authmsg, srcParty, dstParty, reason);

	default:
	    (void) sprintf (PY_pepy,
			    "no support for authentication protocol (%d) for %s",
			    srcParty -> party_authProt,
			    oid2ode (srcParty -> party_Identity));
	    return NOTOK;
    }
}

/*  */

static int  do_md5_auth (privmsg, authmsg, srcParty, dstParty, reason)
struct type_SNMP_SnmpPrivMsg *privmsg;
struct type_SNMP_SnmpAuthMsg *authmsg;
struct snmp_party *srcParty,
		  *dstParty;
int    *reason;
{
    long    delta;
    u_long  current_time,
	    now,
	    timestamp;
    struct qbuf *digest,
		*newdigest = NULL;
    register struct snmp_access *sa;

    if (srcParty -> party_AuthPrivate == NULL) {
	(void) sprintf (PY_pepy,
			"cannot create MD5 authentication without a key");
	return NOTOK;
    }
    if (authmsg -> authInfo -> offset != type_SNMP_AuthInformation_2) {
	(void) sprintf (PY_pepy, "wrong kind of authInformation present (%d)",
			authmsg -> authInfo -> offset);
	if (reason)
	    *reason = AUTH_WRONGAUTH;
	return NOTOK;
    }

    current_time = (now = party_time_now ()) - srcParty -> party_AuthStart;
    timestamp = (u_long) authmsg -> authInfo -> un.choice_SNMP_1
							-> authSrcTimestamp;

    digest = authmsg -> authInfo -> un.choice_SNMP_1 -> authDigest;
    if (digest -> qb_forw -> qb_forw != digest) {
	(void) sprintf (PY_pepy, "constructed-encoded digest");
	if (reason)
	    *reason = AUTH_WRONGAUTH;
	goto what_you_talking_about;
    }

    if (current_time == (u_long) 0xffffffff) {
	if (debug)
	    fprintf (stderr, "clock (%lu) at maximum\n", current_time);

	if (timestamp < current_time) {
	    (void) sprintf (PY_pepy, "timestamp(%lu) < current time(%lu)",
			    timestamp, current_time);
	    if (reason)
		*reason = AUTH_LIFETIME;
	    goto what_you_talking_about;
	}
    }
    else
	if (current_time <= srcParty -> party_AuthLifetime) {
	    /* always authentic (think about it...) */
	}
	else
	    if (timestamp < current_time - srcParty -> party_AuthLifetime) {
		(void) sprintf (PY_pepy,
				"timestamp(%lu) < current time(%lu) - lifetime(%d)",
				timestamp, current_time,
				srcParty -> party_AuthLifetime);
		if (reason)
		    *reason = AUTH_LIFETIME;
		goto what_you_talking_about;
	    }

    if ((newdigest = generate_md5_digest (privmsg -> privData,
					  srcParty -> party_AuthPrivate))
	    == NULL)
	goto what_you_talking_about;

    if (digest -> qb_len != newdigest -> qb_len
	    || bcmp (digest -> qb_forw -> qb_data,
		     newdigest -> qb_forw -> qb_data,
		     digest -> qb_len)) {
	(void) sprintf (PY_pepy, "digest mismatch");
	if (reason)
	    *reason = AUTH_DIGEST;
	goto what_you_talking_about;
    }

    qb_free (newdigest);

    if ((sa = dstParty -> party_access) == NULL)
	goto no_acl_table;	/* manager... */
    while (sa) {
	if (sa -> sa_status == ROW_ACTIVE && sa -> sa_subject == srcParty)
	    break;
	if ((sa = sa -> sa_next) && sa -> sa_target != dstParty)
	    return OK;
    }
    if (!sa)
	return OK;
no_acl_table: ;

    if (current_time < timestamp) {
	if (srcParty -> party_AuthStart > (delta = timestamp - current_time))
	    srcParty -> party_AuthStart -= delta;
	else
	    srcParty -> party_AuthStart = 0;
	if (debug)
	    fprintf (stderr, "advance source clock by %ld\n", delta);
	party_changed = -1;
    }

    current_time = now - dstParty -> party_AuthStart;
    timestamp = (u_long) authmsg -> authInfo -> un.choice_SNMP_1
					              -> authDstTimestamp;
    if (current_time < timestamp) {
	if (dstParty -> party_AuthStart > (delta = timestamp - current_time))
	    dstParty -> party_AuthStart -= delta;
	else
	    dstParty -> party_AuthStart = 0;
	if (debug)
	    fprintf (stderr, "advance destination clock by %ld\n", delta);
	party_changed = -1;
    }

    return OK;

what_you_talking_about: ;
    if (newdigest)
	qb_free (newdigest);

    return NOTOK;
}

/*  */

struct qbuf *generate_outgoing_auth (msg, srcParty, dstParty, watch)
struct type_SNMP_SnmpAuthMsg *msg;
struct snmp_party *srcParty,
		  *dstParty;
caddr_t	watch;
{
    int	    len;
    u_long  now;
    struct qbuf *qb,
		*pb,
		*tmpqbuf;
    register struct element_SNMP_0 *s2;
    PE	    pe;
    static PS	qbufPS = NULLPS;

    if (msg == NULL) {
	(void) sprintf (PY_pepy, "null msg");
	return NULL;
    }
    if (srcParty == NULL) {
	(void) sprintf (PY_pepy, "null srcParty");
	return NULL;
    }
    if (dstParty == NULL) {
	(void) sprintf (PY_pepy, "null dstParty");
	return NULL;
    }

    if (msg -> authInfo)
	free_SNMP_AuthInformation (msg -> authInfo);
    if ((msg -> authInfo = (struct type_SNMP_AuthInformation *)
					calloc (1, sizeof *msg -> authInfo))
	    == NULL) {
no_mem: ;
	(void) sprintf (PY_pepy, "out of memory");
	return NULL;
    }

    switch (srcParty -> party_authProt) {
	case BADAUTH:
	    (void) sprintf (PY_pepy,
			    "unable to determine authentication protocol for %s",
			    oid2ode (srcParty -> party_Identity));
	    return NULL;

	case NOAUTH:
	    msg -> authInfo -> offset = type_SNMP_AuthInformation_1;
	    if ((msg -> authInfo -> un.choice_SNMP_0 = str2qb ("", 0, 1))
		    == NULL)
		goto no_mem;
	    break;

	case MD5AUTHPROT:
	    if (srcParty -> party_AuthPrivate == NULL) {
		(void) sprintf (PY_pepy,
				"cannot create MD5 authentication without a key");
		return NULL;
	    }

	    msg -> authInfo -> offset = type_SNMP_AuthInformation_2;
	    if ((s2 = (struct element_SNMP_0 *) calloc (1, sizeof *s2)) ==NULL)
		goto no_mem;
	    msg -> authInfo -> un.choice_SNMP_1 = s2;

	    s2 -> authSrcTimestamp = (now = party_time_now ())
						- srcParty -> party_AuthStart;

	    s2 -> authDstTimestamp = now - dstParty -> party_AuthStart;

	    if ((s2 -> authDigest =
		         str2qb ((char *) srcParty -> party_AuthPrivate,
				 srcParty -> party_AuthPrivate_len, 1))
		    == NULL)
		goto no_mem;
	    break;

	default:
	    (void) sprintf (PY_pepy,
			    "no support for authentication protocol (%d) for %s",
			    srcParty -> party_authProt,
			    oid2ode (srcParty -> party_Identity));
	    return NULL;
    }

    if (encode_SNMP_SnmpAuthMsg (&pe, 1, 0, NULLCP, msg) == NOTOK)
	return NULL;

    len = ps_get_abs (pe);
    if ((pb = (struct qbuf *) malloc ((unsigned) (sizeof *pb + len)))
	        == NULL
	    || (qb = (struct qbuf *) malloc (sizeof *qb)) == NULL) {
	(void) sprintf (PY_pepy, "out of memory");
oops: ;
	if (pb) {
	    free ((char *) pb);
	    if (qb)
		free ((char *) qb);
	}
	pe_free (pe);
        return NULL;
    }
    qb -> qb_forw = qb -> qb_back = qb;
    qb -> qb_data = NULL, qb -> qb_len = len;
    insque (pb, qb);

    pb -> qb_data = pb -> qb_base, pb -> qb_len = len;

    if (qbufPS == NULLPS) {
	if ((qbufPS = ps_alloc (str_open)) == NULLPS) {
	    (void) sprintf (PY_pepy, "ps_alloc(str_open) failed");
	    goto oops;
	}
	if (str_setup (qbufPS, NULLCP, 0, 0) == NOTOK) {
	    (void) sprintf (PY_pepy, "str_setup failed: %s",
			    ps_error (qbufPS -> ps_errno));
	    ps_free (qbufPS), qbufPS = NULL;
	    goto oops;
	}
    }

    qbufPS -> ps_base = qbufPS -> ps_ptr = pb -> qb_data;
    qbufPS -> ps_cnt = qbufPS -> ps_bufsiz = pb -> qb_len;
    
    if (pe2ps (qbufPS, pe) == NOTOK) {
	(void) sprintf (PY_pepy, "pe2ps: %s", ps_error (qbufPS -> ps_errno));
	goto oops;
    }

    qbufPS -> ps_base = NULL, qbufPS -> ps_cnt = 0;
    qbufPS -> ps_ptr = NULL, qbufPS -> ps_bufsiz = 0;

    switch (srcParty -> party_authProt) {
	case MD5AUTHPROT:
	    if ((tmpqbuf = generate_md5_digest (qb, (u_char *) 0)) == NULL)
		goto oops;
	    qb_free (s2 -> authDigest), s2 -> authDigest = tmpqbuf;
	    if (watch) {
		PE	q = pe -> pe_cons -> pe_cons -> pe_next -> pe_next;

		if (q -> pe_prim && !q -> pe_inline)
		    PEDfree (q -> pe_prim);
		q -> pe_prim = (PElementData) tmpqbuf -> qb_forw -> qb_data;
		q -> pe_len = tmpqbuf -> qb_forw -> qb_len;
		q -> pe_inline = 1;
	    }
	    break;

	default:
	    break;
    }

    if (watch) {
	if (watch == (caddr_t) -1)
	    (void) print_SNMP_SnmpAuthMsg (pe, 1, NULLIP, NULLVP, NULLCP);
	else
	    PLOGP ((LLog *) watch,SNMP_SnmpAuthMsg, pe, "SnmpAuthMsg", 0);
    }

    pe_free (pe);

    return qb;
}

/*  */

static struct qbuf *generate_md5_digest (privdata, secret)
struct qbuf *privdata;
u_char   *secret;
{
    int	    length;
    char   *oldigest_string;
    unsigned char digest[LENOFDIGEST],
		  oldigest[LENOFDIGEST];
    struct qbuf *qb,
		*tmpqbuf;
    MD5_CTX	mdContext;

    if ((oldigest_string = get_digest_from_string (privdata, &length)) == NULL)
	return NULL;
    if (length != LENOFDIGEST) {
	(void) sprintf (PY_pepy, "digest wrong length (got %d, expecting %d)",
			length, LENOFDIGEST);
	return NULL;
    }
    if (secret) {
	bcopy (oldigest_string, (char *) oldigest, LENOFDIGEST);
	bcopy ((char *) secret, oldigest_string, LENOFDIGEST);
    }

    qb = privdata -> qb_forw;

    MD5Init (&mdContext);
    MD5Update (&mdContext, (unsigned char *) qb -> qb_data,
	       (unsigned int) qb -> qb_len);
    MD5Final (digest, &mdContext);

    if (secret)
	bcopy ((char *) oldigest, oldigest_string, LENOFDIGEST);
    else
	bcopy ((char *) digest, oldigest_string, LENOFDIGEST);

    if (debug) {
	register int	i;

	fprintf (stderr, "MD5 generated digest: ");
	for (i = 0; i < LENOFDIGEST; i++)
	    fprintf (stderr, "%02x", digest[i] & 0xff);
	fprintf (stderr, "\n");
    }

    if ((tmpqbuf = str2qb ((char *) digest, LENOFDIGEST, 1)) == NULL)
	(void) sprintf (PY_pepy, "str2qb failed");

    return tmpqbuf;
}

/*  */

static char *get_digest_from_string (qb, length)
struct qbuf *qb;
int   *length;
{
    char   *digest;
    PElementClass   class;
    PElementForm    form;
    PElementID	    id;
    PElementLen	    len;
    static PS  stringPS = NULLPS;

    if (qb -> qb_forw -> qb_forw != qb) {
	(void) sprintf (PY_pepy, "constructed-encoded authenticated message");
	return NULL;
    }
    qb = qb -> qb_forw;

    if (stringPS == NULLPS) {
	if ((stringPS = ps_alloc (str_open)) == NULLPS) {
	    (void) sprintf (PY_pepy, "ps_alloc(str_open) failed");
	    return NULL;
	}
	if (str_setup (stringPS, NULLCP, 0, 0) == NOTOK) {
	    (void) sprintf (PY_pepy, "str_setup failed: %s",
			    ps_error (stringPS -> ps_errno));
	    ps_free (stringPS), stringPS = NULL;
	    return NULL;
	}
    }

    stringPS -> ps_base = stringPS -> ps_ptr = qb -> qb_data;
    stringPS -> ps_cnt = stringPS -> ps_bufsiz = qb -> qb_len;

/* now skip over for SNMPv2:
    [1] IMPLICIT SEQUENCE {
	[2] IMPLICIT SEQUENCE {
		authDigest OCTET STRING,
*/

    if (ps_read_id (stringPS, 0, &class, &form, &id) == NOTOK
	    || ps_read_len (stringPS, &len) == NOTOK
	    || form != PE_FORM_CONS
	    || len <= 0

	    || ps_read_id (stringPS, 0, &class, &form, &id) == NOTOK
	    || ps_read_len (stringPS, &len) == NOTOK
	    || form != PE_FORM_CONS
	    || len <= 0) {
no_digest: ;
	if (stringPS -> ps_errno)
	    (void) sprintf (PY_pepy, "%s while looking for digest",
			    ps_error (stringPS -> ps_errno));
	else
	    (void) sprintf (PY_pepy,
			    "invalid syntax (form=%d len=%d) while looking for digest",
			    form, id);

	digest = NULL, *length = 0;
    }
    else
	if (ps_read_id (stringPS, 0, &class, &form, &id) == NOTOK
	        || ps_read_len (stringPS, &len) == NOTOK
	        || form != PE_FORM_PRIM
	        || len == PE_LEN_INDF
	        || stringPS -> ps_cnt < len)
	    goto no_digest;
	else
	    digest = stringPS -> ps_ptr, *length = len;

    stringPS -> ps_base = NULL, stringPS -> ps_cnt = 0;
    stringPS -> ps_ptr = NULL, stringPS -> ps_bufsiz = 0;

/*
    if (debug && digest) {
	register int	i;

	fprintf (stderr, "    extracted digest: ");
	for (i = 0; i < len; i++)
	    fprintf (stderr, "%02x", digest[i] & 0xff);
	fprintf (stderr, "\n");
    }
 */

    return digest;
}
