: run this script through /bin/sh

# bootV1.sh manager 192.103.140.1 2001 "public" 100 read-only

DB="$1"
IP="$2"
PORT="$3"
if [ -z "$PORT" ]; then
    PORT=161
fi
COMMUNITY="$4"
OFFSET="$5"
PRIVS="$6"
case "$PRIVS" in
    read-only)	PRIVS="get get-next response get-bulk"
		;;

    read-write)	PRIVS="get get-next response set get-bulk"
		;;

    *)		PRIVS=""
		;;
esac

if [    -z "$DB" \
     -o -z "$IP" \
     -o -z "$COMMUNITY" \
     -o -z "$OFFSET" \
     -o -z "$PRIVS" ]; then
    echo "usage: bootV1.sh party-database IP-address UDP-port community offset privs" 1>&2
    exit 1
fi

if [ ! -f "$DB.public" ]; then
    if [ ! -f "/usr/local/etc/parties/$DB.public" ]; then
	echo "unable to find party database $DB" 2>&1
	exit 1
    fi
    DB=/usr/local/etc/parties/$DB
fi

if [    ! -w $DB.public \
     -o ! -w $DB.private \
     -o ! -w $DB.context \
     -o ! -w $DB.access ]; then
    echo "unable to write party database $DB" 2>&1
    exit 1
fi

echo \
"
A initialPartyId.$IP.$OFFSET \\
	rfc1157Domain 0x$IP.0.$PORT 8192 \\
	rfc1157noAuth 0 0 \\
	noPriv" \
    >> $DB.public

echo \
"
initialPartyId.$IP.$OFFSET		active \"$COMMUNITY\"" \
    >> $DB.private

echo \
"
A initialContextId.$IP.$OFFSET \"$IP-$COMMUNITY\" active \\
	view 1 \"\" currentTime" \
    >> $DB.context

echo \
"
initialPartyId.$IP.$OFFSET \\
	initialPartyId.$IP.$OFFSET \\
	initialContextId.$IP.$OFFSET \\
	active $PRIVS" \
    >> $DB.access

exit 0
