: run this script through /bin/sh

B="bootparty" D="party"
PUBLIC="$B.public" PRIVATE="$B.private" CONTEXT="$B.context" ACCESS="$B.access" VIEW="$B.view"
touch	   $PUBLIC $PRIVATE $ACCESS $VIEW $D.public $D.private $D.context $D.access
chmod 0600         $PRIVATE
chmod 0644 $PUBLIC          $ACCESS $VIEW $D.public $D.private $D.context $D.access

if [ "$1" = "-1" ];  then
    ONEKEY=T
    shift
fi

A="$1"
if [ -z "$A" ]; then
    echo "usage: bootparty.sh IP-address [port]" 1>&2
    exit 1;
fi
if echo "$A" | awk '
/^[1-9][0-9]*\.[1-9][0-9]*\.[1-9][0-9]*\.[1-9][0-9]*$/ { next; }
	{ exit(1); }
'; then
    true
else
    echo "invalid IP-address: $A" 1>&2
    echo "usage: bootparty.sh IP-address" 1>&2
    exit 1;
fi

if [ -z "$2" ]; then
    P="0.161"
else
    P="`echo $2 | awk '{ x = ($1 % 256); y = int($1 / 256); printf "%d.%d\n", y, x;}'`"
fi

F=/tmp/bootparty.$$
trap "rm -f $F" 0 1 2 3 13 15
touch $F
chmod 0600 $F

ps axl | sed -e 's% %%g' | awk '
	{
	    if ($NR == 1) next;
	    printf "Y%d=\"%cecho %c%s%c | /usr/lib/makekey | awk ", \
		flip, 96, 39, substr($1,9,10), 39;
	    printf "%c{ printf substr($1,1,8) }%c%c\"\n", 39, 39, 96;
	    if (flip) flip = 0; else flip = 1;
	    if (!flip) { k++; printf "K%d=\"$Y0$Y1\"\n", k; if (k == 8) exit; }
	}
' > $F
. $F

if [ ! -z "$ONEKEY" ]; then
    K2="$K1" K3="$K1" K4="$K1" K5="$K1" K6="$K1"
fi

echo \
"###############################################################################
#
# $PUBLIC - public party information
#
###############################################################################

###############################################################################
#
# Syntax:
#
#	<location> <identity> <tdomain> <taddress> <maxsize>
#	<authprotocol> <authstart> <authlifetime> 
#	<privprotocol> [ <authPublic> [ <privPublic> ]]
#
# where:
#
#	<location>	- one of "A", "L", "M", "P"
#	<identity>	- OID
#	<tdomain>	- OID
#	<taddress>	- <string>
#	<maxsize>	- 484..65507
#	<authprotocol>	- OID
#	<authclock>	- unsigned-long
#	<authlifetime>	- 0..2147483647
#	<privprotocol>	- OID
#	<authPublic>	- <string>
#	<privPublic>	- <string>
#
#	<string>	- 0x<octet>: ... :<octet>  (in hex)
#			- 0x<octet>. ... .<octet>  (in decimal)
#			- "ascii"
#
###############################################################################


L initialPartyId.$A.1 \\
	snmpUDPDomain 0x$A.$P 8192 \\
	noAuth 0 0 \\
	noPriv

M initialPartyId.$A.2 \\
	snmpUDPDomain 0x0.0.0.0.0.0 8192 \\
	noAuth 0 0 \\
	noPriv

L initialPartyId.$A.3 \\
	snmpUDPDomain 0x$A.$P 8192 \\
	v2md5AuthProtocol 0 300 \\
	noPriv

M initialPartyId.$A.4 \\
	snmpUDPDomain 0x0.0.0.0.0.0 8192 \\
	v2md5AuthProtocol 0 300 \\
	noPriv

L initialPartyId.$A.5 \\
	snmpUDPDomain 0x$A.$P 8192 \\
	v2md5AuthProtocol 0 300 \\
	desPrivProtocol

M initialPartyId.$A.6 \\
	snmpUDPDomain 0x0.0.0.0.0.0 8192 \\
	v2md5AuthProtocol 0 300 \\
	desPrivProtocol


# this next party is for proxy to the SNMP unix agent

P initialPartyId.$A.7 \\
	rfc1157Domain 0x$A.100.161 8192 \\
	rfc1157noAuth 0 0 \\
	noPriv" \
> $PUBLIC


echo \
"###############################################################################
#
# $D.public - public party information (noAuth/noPriv only)
#
###############################################################################

L initialPartyId.$A.1 \\
	snmpUDPDomain 0x$A.$P 8192 \\
	noAuth 0 0 \\
	noPriv

M initialPartyId.$A.2 \\
	snmpUDPDomain 0x0.0.0.0.0.0 8192 \\
	noAuth 0 0 \\
	noPriv" \
> $D.public


echo \
"###############################################################################
#
# $PRIVATE - private party information
#
###############################################################################

###############################################################################
#
# Syntax:
#
#	<identity> <status> [ <authPrivate> [ <privPrivate> ]]
#
# where:
#
#	<identity>	- OID
#	<status>	- "active" | "notInService" | "notReady"
#	<authPrivate>	- <string>
#	<privPrivate>	- <string>
#
#	<string>	- 0x<octet>: ... :<octet>  (in hex)
#			- 0x<octet>. ... .<octet>  (in decimal)
#			- "ascii"
#
###############################################################################


initialPartyId.$A.1 active
initialPartyId.$A.2 active

# these keys were automatically generated, change them right away!
initialPartyId.$A.3 active	\"$K1\"
initialPartyId.$A.4 active	\"$K2\"

# these keys were automatically generated, change them right away!
initialPartyId.$A.5 active	\"$K3\" \"$K4\"
initialPartyId.$A.6 active	\"$K5\" \"$K6\"

# this entry is for the SNMP unix agent
# this key was automatically generated, change it right away!
initialPartyId.$A.7 active	\"$K7\"" \
> $PRIVATE


echo \
"###############################################################################
#
# $D.private - private party information
#
###############################################################################

initialPartyId.$A.1 active
initialPartyId.$A.2 active" \
> $D.private


echo \
"###############################################################################
#
# $CONTEXT - party context information
#
###############################################################################

###############################################################################
#
# Syntax:
#
#	<location> <identity> <name> <status>
#       [  "view"  <index>  <entity> <time>
#        | "proxy" <dst-id> <src-id> <ctx-id> ]
#
# where:
#
#	<location>	- one of "A", "M"
#	<identity>	- OID
#	<name>		- <string>
#	<status>	- "active" | "notInService" | "notReady"
#	<index>		- 0..65535
#       <entity>        - <string>
#       <time>          - OID
#	<dst-id>	- OID
#	<src-id>	- OID
#	<ctx-id>	- OID
#
#	<string>	- 0x<octet>: ... :<octet>  (in hex)
#			- 0x<octet>. ... .<octet>  (in decimal)
#			- "ascii"
#
###############################################################################


A initialContextId.$A.1 \"$A-system\"   active \\
	view 1 \"\" currentTime
A initialContextId.$A.2 \"$A-internet\" active \\
	view 2 \"\" currentTime

# this entry is for the proxy to the SNMP unix agent
A initialContextId.$A.3 \"$A-unix\"     active \\
	proxy initialPartyId.$A.7 0.0 0.0" \
> $CONTEXT


echo \
"###############################################################################
#
# $D.context - party context information
#
###############################################################################

A initialContextId.$A.1 \"$A-system\"   active \\
        view 1 \"\" currentTime

# this entry is for the proxy to the SNMP unix agent
A initialContextId.$A.3 \"$A-unix\"     active \\
	proxy initialPartyId.$A.7 0.0 0.0" \
> $D.context


echo \
"###############################################################################
#
# $ACCESS - party acl information
#
###############################################################################

###############################################################################
#
# Syntax:
#
#	<target> <subject> <resources> <status> [ <operation> ... ]
#
# where:
#
#	<target>	- OID
#	<subject>	- OID
#	<resources>	- OID
#	<status>	- "active" | "notInService" | "notReady"
#	<operation>	-   "get"  | "get-next" | "response" | "set" | "trap"
#			  | "get-bulk" | "inform" | "notify"
#
###############################################################################


# this entry allows the agent to send traps to itself
initialPartyId.$A.1 \\
	initialPartyId.$A.1 \\
	initialContextId.$A.1 \\
	active notify

initialPartyId.$A.1 \\
	initialPartyId.$A.2 \\
	initialContextId.$A.1 \\
	active get get-next get-bulk
initialPartyId.$A.2 \\
	initialPartyId.$A.1 \\
	initialContextId.$A.1 \\
	active response

initialPartyId.$A.3 \\
	initialPartyId.$A.4 \\
	initialContextId.$A.2 \\
	active get get-next set get-bulk
initialPartyId.$A.4 \\
	initialPartyId.$A.3 \\
	initialContextId.$A.2 \\
	active response

initialPartyId.$A.5 \\
	initialPartyId.$A.6 \\
	initialContextId.$A.2 \\
	active get get-next set get-bulk
initialPartyId.$A.6 \\
	initialPartyId.$A.5 \\
	initialContextId.$A.2 \\
	active response

# these five entries are for the proxy to the SNMP unix agent
initialPartyId.$A.1 \\
	initialPartyId.$A.1 \\
	initialContextId.$A.3 \\
	active notify
initialPartyId.$A.1 \\
	initialPartyId.$A.2 \\
	initialContextId.$A.3 \\
	active get get-next get-bulk
initialPartyId.$A.2 \\
	initialPartyId.$A.1 \\
	initialContextId.$A.3 \\
	active response
initialPartyId.$A.3 \\
	initialPartyId.$A.4 \\
	initialContextId.$A.3 \\
	active get get-next set get-bulk
initialPartyId.$A.4 \\
	initialPartyId.$A.3 \\
	initialContextId.$A.3 \\
	active response" \
> $ACCESS


echo \
"###############################################################################
#
# $D.access - party acl information
#
###############################################################################

initialPartyId.$A.1 \\
	initialPartyId.$A.2 \\
	initialContextId.$A.1 \\
	active get get-next get-bulk
initialPartyId.$A.2 \\
	initialPartyId.$A.1 \\
	initialContextId.$A.1 \\
	active response
initialPartyId.$A.1 \\
	initialPartyId.$A.2 \\
	initialContextId.$A.3 \\
	active get get-next get-bulk
initialPartyId.$A.2 \\
	initialPartyId.$A.1 \\
	initialContextId.$A.3 \\
	active response" \
> $D.access


echo \
"###############################################################################
#
# $VIEW - party view information
#
###############################################################################

###############################################################################
#
# Syntax:
#
#	<index> <status> <type> <subtree> [<mask>]
#
# where:
#
#	<index>		- 1..65535
#	<status>	- "active" | "notInService" | "notReady"
#	<type>		- "included" | "excluded"
#	<subtree>	- OID
#	<mask>		- string
#
###############################################################################


1	active	included	system
1	active	included	snmpStats
1	active	included	snmpParties
# this entry allows the agent to send traps to itself
1	active	included	snmpTraps

2	active	included	internet" \
> $VIEW


exit
