/* snmp.c - SNMPv2 changes for gawk */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/* 
 * $Header$
 *
 *
 * $Log$
 */

/*
 *			  ISODE 8.0 NOTICE
 *
 *   Acquisition, use, and distribution of this module and related
 *   materials are subject to the restrictions of a license agreement.
 *   Consult the Preface in the User's Manual for the full terms of
 *   this agreement.
 *
 *
 *			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#ifdef	SNMP
#include "awk.h"
#ifdef	HUGE
#undef	HUGE
#endif
#define	free	_free_
#include <isode/snmp/objects.h>
#undef	free
#include <isode/pepsy/SNMP-types.h>
#include <isode/snmp/party.h>
#include <isode/dgram.h>
#include <isode/internet.h>
#include <isode/isoaddrs.h>
#include <isode/tailor.h>

/*    DATA */

int	debug = 0;

int	snmp_enabled = 1;

static	integer	snmp_id = 0;
static	int	snmp_portno = 0;
static	int	snmp_retries = 3;
static	int	snmp_timeout = 5;

static	char   *snmp_ctx = NULL;

NODE   *CONTEXT_node,           /* used to define management CONTEXT */
       *DIAGNOSTIC_node,
       *ERROR_node,
       *RETRIES_node,
       *TIMEOUT_node;
NODE   *Ndot_string;

NODE   *SET_node;
NODE   *SET_inst;

static	int	snmp_fd = NOTOK;
static	struct sockaddr_in in_socket;
static	PS	ps = NULLPS;

static	char   *pParty = "party";

static	struct snmp_party *srcParty,
			  *dstParty;
static	struct snmp_context *theContext;

static	PE	unSpecified;
static	struct type_SNMP_SnmpAuthMsg	authmsgs;
static	struct type_SNMP_Snmp1157Msg	msgs;
static	struct type_SNMP_SnmpMgmtCom	mgmtcoms;
static	struct type_SNMP_PDUs		pdus;
static	struct type_SNMP_PDU		parms;
static	struct type_SNMP_VarBindList	vps;
static	struct type_SNMP_VarBind	vs;


struct snmp_search {
    OT	    s_parent;

    struct type_SNMP_VarBindList *s_prototype;

#define	NREQ	10
    struct snmp_req {
	struct type_SNMP_VarBindList *r_bindings;
	PE	r_pb;

	integer	r_id;
	PE	r_pe;

	struct type_SNMP_SnmpAuthMsg *r_authmsg;

	int	r_toobig;
    }	    s_reqs[NREQ];

    struct snmp_search *s_prev;
    struct snmp_search *s_next;
};

static struct snmp_search *head = NULL;
static struct snmp_search *tail = NULL;


static	snmp_onceonly ();
static	snmp_get_next (), snmp_get_next_aux (), next_ready ();
int	snmp_ready (), snmp_build_msg (), snmp_build_v1_msg (), 
	snmp_recv_msg (), snmp_recv_v1_msg (), snmp_diag ();
char   *snmp_error (), *snmp_exception (), *snmp_variable ();



#ifndef	SYS5
long	random ();
#endif

/*    INIT */

int	snmp_init (cValue)
char  **cValue;
{
    char   *cp;

    snmp_onceonly ();

    cp = sprintoid (theContext -> sc_identity);
    emalloc (snmp_ctx, char *, strlen (cp) + 1, "snmp_init");
    (void) strcpy (*cValue = snmp_ctx, cp);

    return 0;
}


void	snmp_fin ()
{
    if (sync_party (pParty, MANAGER, 0) == NOTOK)
	fatal ("stable storage suspect: %s", PY_pepy);

    if (ps)
	ps_free (ps), ps = NULLPS;
    if (snmp_fd != NOTOK)
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;
}

/*  */

int	e_integer (), e_octets (), e_display (), e_objectID (), e_bits (),
    	e_null (), e_counter (), e_counter64 (), e_gauge (), e_uint (),
    	e_timeticks (), e_ipaddr (), e_nsapaddr ();
int	d_integer (), d_octets (), d_display (), d_objectID (), d_bits (),
	d_null (), d_ulong (), d_ipaddr (), d_nsapaddr ();


static struct pair {
    char   *pp_name;

    IFP	    pp_new_encode;
    IFP	    pp_new_decode;

    IFP	    pp_old_encode;
    IFP	    pp_old_decode;
    IFP	    pp_old_print;
}	pairs[] = {
    "Integer32",	e_integer,	d_integer,   NULLIFP, NULLIFP, NULLIFP,
    "OctetString",	e_octets,	d_octets,    NULLIFP, NULLIFP, NULLIFP,
    "DisplayString",	e_display,	d_display,   NULLIFP, NULLIFP, NULLIFP,
    "ObjectID",		e_objectID,	d_objectID,  NULLIFP, NULLIFP, NULLIFP,
    "BitString",	e_bits,		d_bits,	     NULLIFP, NULLIFP, NULLIFP,
#ifdef	notdef
    "NULL",		e_null,		d_null,	     NULLIFP, NULLIFP, NULLIFP,
#endif
    "Counter32",	e_counter,	d_ulong,     NULLIFP, NULLIFP, NULLIFP,
    "Counter64",	e_counter64,	d_ulong,     NULLIFP, NULLIFP, NULLIFP,
    "Gauge32",		e_gauge,	d_ulong,     NULLIFP, NULLIFP, NULLIFP,
    "UInteger32",	e_uint,		d_ulong,     NULLIFP, NULLIFP, NULLIFP,
    "TimeTicks",	e_timeticks,	d_ulong,     NULLIFP, NULLIFP, NULLIFP,
    "IpAddress",	e_ipaddr,	d_ipaddr,    NULLIFP, NULLIFP, NULLIFP,
    "NsapAddress",	e_nsapaddr,	d_nsapaddr,  NULLIFP, NULLIFP, NULLIFP,

    NULL
};


static	snmp_onceonly () {
    int	    i;
    char   *cp;
    register struct pair *pp;
    register struct type_SNMP_SnmpAuthMsg *authmsg = &authmsgs;
    register struct type_SNMP_Snmp1157Msg *msg = &msgs;
    register struct type_SNMP_SnmpMgmtCom *mgmtcom = &mgmtcoms;
    register struct type_SNMP_PDUs *pdu = &pdus;
    register struct type_SNMP_PDU *parm = &parms;
    register struct type_SNMP_VarBindList *vp = &vps;
    register struct type_SNMP_VarBind *v = &vs;
    OS	    first,
	    last,
	    os;
    register OT	    ot,
		    ot2;

    Ndot_string = make_string (".", 1);
    Ndot_string -> flags |= PERM;

    SET_node = SET_inst = NULL;

    _tailor_logging_only = 1;
    if (loadobjects (getenv ("MIBDEFS")) == NOTOK)
	fatal ("loadobjects: %s", PY_pepy);

				/* mark entries that are actually columns! */
    for (ot = anchor; ot; ot = ot -> ot_next) {
	if (ot -> ot_syntax
		|| (i = strlen (ot -> ot_text)) <= 5
	        || strcmp (ot -> ot_text + i - 5, "Entry"))
	    continue;
	for (ot2 = ot -> ot_children; ot2; ot2 = ot2 -> ot_sibling)
	    if (ot2 -> ot_children)
		break;
	if (ot2)
	    continue;
	for (ot2 = ot -> ot_children; ot2; ot2 = ot2 -> ot_sibling)
	    if (ot2 -> ot_syntax)
		ot2 -> ot_getfnx = (IFP) 1;
	    else
		if (debug > 0)
		    fprintf (stderr, "no syntax for columnar object \"%s\"\n",
			     ot -> ot_text);
    }

    for (pp = pairs; pp -> pp_name; pp++)
	if ((os = text2syn (pp -> pp_name)) == NULL)
	    fatal ("lost syntax for \"%s\"", pp -> pp_name);
        else {
	    pp -> pp_old_encode = os -> os_encode;
	    pp -> pp_old_decode = os -> os_decode;
	    pp -> pp_old_print = os -> os_print;
	}

    flsyntax (&first, &last);
    for (os = first; os <= last; os++) {
	for (pp = pairs; pp -> pp_name; pp++)
	    if (pp -> pp_old_print == os -> os_print) {		
		os -> os_encode = pp -> pp_new_encode;
		os -> os_decode = pp -> pp_new_decode;
		break;
	    }
	if (pp -> pp_name)
	    continue;

	for (pp = pairs; pp -> pp_name; pp++)
	    if (pp -> pp_old_encode == os -> os_encode) {		
		os -> os_encode = pp -> pp_new_encode;
		os -> os_decode = pp -> pp_new_decode;
		break;
	    }
	if (pp -> pp_name)
	    continue;

	fatal ("don't have syntax mapping for \"%s\"", os -> os_name);
    }
    
    if (!(cp = getenv ("PARTYDEFS")))
	cp = "party";
    if (init_party (cp, MANAGER, 0, &pParty) == NOTOK)
	fatal ("%s", PY_pepy);

    for (theContext = contextHead;
	     theContext;
	     theContext = theContext -> sc_next)
	if (theContext -> sc_status == ROW_ACTIVE
	        && theContext -> sc_location == AGENT)
	    break;
    if (!theContext)
	fatal ("unable to find any management contexts");

    if ((unSpecified = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_NULL))
		    == NULL)
	fatal ("pe_alloc: out of memory");

    bzero ((char *) authmsg, sizeof *authmsg);
    authmsg -> authData = mgmtcom;

    bzero ((char *) mgmtcom, sizeof *mgmtcom);
    mgmtcom -> pdu = pdu;

    bzero ((char *) msg, sizeof *msg);
    msg -> version = int_SNMP_version_version__1;
    msg -> data = pdu;

    bzero ((char *) pdu, sizeof *pdu);
    pdu -> offset = type_SNMP_PDUs_get__request;
    pdu -> un.get__request = parm;

    bzero ((char *) parm, sizeof *parm);
    parm -> variable__bindings = vp;

    bzero ((char *) vp, sizeof *vp);
    vp -> VarBind = v;

    bzero ((char *) v, sizeof *v);
    (v -> value = unSpecified) -> pe_refcnt++;

    ps_len_strategy = PS_LEN_LONG;

#ifndef	SYS5
    srandom (getpid ());
#else
    srand (getpid ());
#endif
}

/*    CHECK */

int	snmp_check (r, name)
NODE   *r;
char   *name;
{
    char    c;
    register char   *cp;
    OT	    ot;

    for (cp = name; is_identchar (*cp); cp++)
	continue;
    if (c = *cp)
	*cp = NULL;
    if (ot = text2obj (name)) {
	if (ot -> ot_syntax) {
	    r -> magic = (caddr_t) ot;
	    r -> type = Node_var_array;
	}
	else
	    fatal ("SNMP variable \"%s\" has unknown syntax", ot -> ot_text);
    }
    *cp = c;
}

/*    GET */

int	snmp_get (ptr, instname)
NODE   *ptr;
char   *instname;
{
    int	    gotone,
	    retries,
	    status = -1;
    struct type_SNMP_SnmpAuthMsg *authmsg = &authmsgs;
    struct type_SNMP_Snmp1157Msg *msg = &msgs;
    struct type_SNMP_SnmpPrivMsg *privmsg = NULL;
    register struct type_SNMP_PDU *parm =
				authmsg -> authData -> pdu -> un.get__request;
    register struct type_SNMP_VarBind *v =
					parm -> variable__bindings -> VarBind;
    struct qbuf *qb = NULL;
    struct snmp_party *sParty,
		      *dParty;
    PE	    pe = NULLPE,
	    p = NULLPE;
    OID	    oid;
    OT	    ot = (OT) ptr -> magic;
    NODE   *value = NULL;

    if (snmp_ready (1, PARTY_PRIV_GET) == NOTOK)
	goto out;

    parm -> request__id = snmp_id;
    if (v -> name)
	free_SNMP_ObjectName (v -> name), v -> name = NULL;

    if (instname == NULL) {
	register struct snmp_search *s;

	for (s = tail; s; s = s -> s_prev) {
	    register struct snmp_req *sr;

	    if (ot -> ot_name -> oid_nelem
		        != (oid = s -> s_parent -> ot_name) -> oid_nelem + 1
		    || bcmp ((char *) ot -> ot_name -> oid_elements,
			     (char *) oid -> oid_elements,
			     oid -> oid_nelem
			     	* sizeof ot -> ot_name -> oid_elements[0]))
		continue;
	    for (sr = s -> s_reqs; sr -> r_bindings; sr++) {
		register struct type_SNMP_VarBindList *vp;

		for (vp = sr -> r_bindings; vp; vp = vp -> next) {
		    if (ot -> ot_name -> oid_nelem
			    >= (v = vp -> VarBind) -> name -> oid_nelem)
			fatal ("snmp_get: internal error");
		    if (bcmp ((char *) v -> name -> oid_elements,
			      (char *) ot -> ot_name -> oid_elements,
			      ot -> ot_name -> oid_nelem
			          * sizeof ot -> ot_name -> oid_elements[0]))
			continue;

		    goto get_value;
		}
	    }
	    status = int_SNMP_error__status_noSuchName;
	    goto out;
	}
	if (ot -> ot_getfnx) {
	    snmp_diag (NULLCP,
		       "can't use SNMP array variable as scalar unless within for-in construct");
	    goto out;
	}

	if ((oid = v -> name = oid_extend (ot -> ot_name, 1)) == NULL) {
no_mem_for_inst: ;
	    snmp_diag (NULLCP, "oid_extend: out of memory");
	    goto out;
	}
	v -> name -> oid_elements[v -> name -> oid_nelem - 1] = 0;
    }
    else {
	register int	i;
	register unsigned int *ip,
			      *jp;
	OID	inst = str2oid (instname);

	if (inst == NULL) {
	    snmp_diag (NULLCP, "str2oid: bad instance identifier \"%s\"",
		       instname);
	    goto out;
	}
	if ((oid = v -> name = oid_extend (ot -> ot_name, inst -> oid_nelem))
	        == NULL)
	    goto no_mem_for_inst;
	ip = oid -> oid_elements + oid -> oid_nelem - inst -> oid_nelem;
	jp = inst -> oid_elements;
	for (i = inst -> oid_nelem; i > 0; i--)
	    *ip++ = *jp++;
    }

    if (dstParty -> party_tDomain == RFC1157DOMAIN) {
	if (snmp_build_v1_msg (msg, &pe) == NOTOK)
	    goto out;
    }
    else
	if (snmp_build_msg (authmsg, &pe) == NOTOK)
	    goto out;

    authmsg = NULL, msg = NULL, gotone = 0;
    for (retries = snmp_retries; retries > 0; ) {
	int	len;
	fd_set	rfds;

	if (debug > 1) {
	    if (dstParty -> party_tDomain == RFC1157DOMAIN)
		print_SNMP_Snmp1157Msg (pe, 1, NULLIP, NULLVP, NULLCP);
	    else
		print_SNMP_SnmpPrivMsg (pe, 1, NULLIP, NULLVP, NULLCP);
	}
	len = ps -> ps_byteno;
	if (pe2ps (ps, pe) == NOTOK) {
	    snmp_diag (NULLCP, "pe2ps: %s", ps_error (ps -> ps_errno));
	    goto error_x;
	}
	if (debug > 0
	        && (len = ps -> ps_byteno - len)
	    		> dstParty -> party_MaxMessageSize)
	    fprintf (stderr, "sent message of %d octets\n", len);

	FD_ZERO (&rfds);
	FD_SET (snmp_fd, &rfds);

	switch (xselect (snmp_fd + 1, &rfds, NULLFD, NULLFD, snmp_timeout)) {
	    case NOTOK:
	        snmp_diag ("failed", "xselect");
		goto error_x;

	    default:
		if (FD_ISSET (snmp_fd, &rfds))
		    break;
		/* else fall... */
	    case OK:
		if (debug > 0)
		    fprintf (stderr, "timeout\n");
		retries--;
		continue;
	}

	if (dstParty -> party_tDomain == RFC1157DOMAIN) {
	    if (snmp_recv_v1_msg (&msg, &p) == NOTOK)
		goto out;

	    if (msg -> data -> offset != type_SNMP_PDUs_response) {
		snmp_diag (NULLCP, "unexpected message type %d",
			   msg -> data -> offset);
		goto out;
	    }

	    if ((parm = msg -> data -> un.response) -> request__id == snmp_id)
		break;
	}
	else {
	    if (snmp_recv_msg (&privmsg, &authmsg, &sParty, &dParty, &p, &qb)
	            == NOTOK)
		goto out;

	    if (authmsg -> authData -> pdu -> offset
		    != type_SNMP_PDUs_response) {
		snmp_diag (NULLCP, "unexpected message type %d",
			   authmsg -> authData -> pdu -> offset);
		goto out;
	    }

	    if ((parm = authmsg -> authData -> pdu -> un.response)
						    -> request__id == snmp_id)
		break;
	}

	if (debug > 0)
	    fprintf (stderr, "bad ID (got %ld, wanted %ld)\n",
		     (long) parm -> request__id, (long) snmp_id);

	if (p)
	    pe_free (p), p = NULL;
	if (qb)
	    qb_free (qb), qb = NULL;
	if (privmsg)
	    free_SNMP_SnmpPrivMsg (privmsg), privmsg = NULL;
	if (authmsg)
	    free_SNMP_SnmpAuthMsg (authmsg), authmsg = NULL;
	if (msg)
	    free_SNMP_Snmp1157Msg (msg), msg = NULL;

	gotone++;
    }
    if (retries <= 0) {
	snmp_diag (NULLCP,
		   "no %sresponse within %d retries of %s%d second%s each",
		   gotone ? "acceptable " : "", snmp_retries + gotone,
		   gotone ? "upto " : "",
		   snmp_timeout, snmp_timeout != 1 ? "s" : "");
	goto out;
    }

    if ((status = parm -> error__status) != int_SNMP_error__status_noError) {
	char   *cp = snmp_variable (parm, parm -> error__index);

	snmp_diag (NULLCP, cp ? "%s at position %d (%s)" : "%s at position %d",
		   snmp_error (status), parm -> error__index, cp);
	goto out;
    }

    if (parm -> variable__bindings == NULL
	    || (v = parm -> variable__bindings -> VarBind) == NULL) {
	snmp_diag (NULLCP, "missing variable in response");
	goto out;
    }
    if (debug > 0 && parm -> variable__bindings -> next)
	fprintf (stderr, "too many responses starting with: %s\n",
		 oid2ode (parm -> variable__bindings -> next -> VarBind -> name));

    if (oid_cmp (oid, v -> name)) {
	char    buffer[BUFSIZ];

	(void) strcpy (buffer, oid2ode (v -> name));
	snmp_diag (NULLCP, "wrong variable returned (got %s, wanted %s)",
		 buffer, oid2ode (oid));
	goto out;
    }
		 
get_value: ;
    if (v -> value -> pe_class == PE_CLASS_CONT) {
	snmp_diag (NULLCP, "exception for variable %s: %s",
		 oid2ode (v -> name), snmp_exception (v -> value -> pe_id));
	goto out;
    }
    if ((*ot -> ot_syntax -> os_decode) (&value, v -> value) == NOTOK) {
	snmp_diag (NULLCP, "decode error for variable %s: %s",
		 oid2ode (v -> name), PY_pepy);
	goto out;
    }

    goto out;

error_x: ;
    if (ps)
	ps_free (ps), ps = NULLPS;
    if (snmp_fd != NOTOK)
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;

out: ;
    if (p)
	pe_free (p);
    if (qb)
	qb_free (qb);
    if (privmsg)
	free_SNMP_SnmpPrivMsg (privmsg);
    if (authmsg && authmsg != &authmsgs)
	free_SNMP_SnmpAuthMsg (authmsg);
    if (msg && msg != &msgs)
	free_SNMP_Snmp1157Msg (msg);
    if (pe)
	pe_free (pe);

    unref (ptr -> var_value);
    ptr -> var_value = value ? value : Nnull_string;

    ERROR_node -> var_value = make_number ((AWKNUM) status);
}

/*    SCAN */

void	snmp_assoc_scan (symbol, lookat, instance)
NODE   *symbol,
       *instance;
struct search *lookat;
{
    register struct snmp_search *s;
    OID	    inst;
    register OT	    ot = (OT) symbol -> magic;
    register struct type_SNMP_VarBindList **vp,
					  **vp2;

    lookat -> retval = NULL;
    if (!ot -> ot_getfnx && instance)
	fatal ("can't use SNMP scalar variable \"%s\" as control for for-in",
	       ot -> ot_text);

    if (instance) {
	char   *instname;
	NODE   *r;

	if ((r = instance) -> type != Node_val)
	    r = r_tree_eval (instance);

	if ((inst = str2oid (instname = force_string (r) -> stptr)) == NULL) {
	    snmp_diag (NULLCP, "str2oid: bad instance identifier \"%s\"",
		       instname);
	    free_temp (r);
	    return;
	}

	if (r != instance)
	    free_temp (r);
    }
    else
	inst = NULL;
    emalloc (s, struct snmp_search *, sizeof *s, "snmp_assoc_scan1");
    bzero ((char *) s, sizeof *s);

    ot -> ot_name -> oid_nelem--;
    s -> s_parent = name2obj (ot -> ot_name);
    ot -> ot_name -> oid_nelem++;

    if ((ot = s -> s_parent) == NULL)
	fatal ("unable to find parent for \"%s\"", snmp_name (symbol));

    s -> s_prototype = NULL;

    vp = &s -> s_reqs[0].r_bindings, vp2 = &s -> s_prototype;
    for (ot = ot -> ot_children; ot; ot = ot -> ot_sibling) {
	register int    i;
	register unsigned int *ip,
			      *jp;
	register struct type_SNMP_VarBindList *bind;
	register struct type_SNMP_VarBind *v,
					  *v2;

	if (!ot -> ot_syntax)
	    continue;
	emalloc (bind, struct type_SNMP_VarBindList *, sizeof *bind,
		 "snmp_assoc_scan2");
	*vp = bind, vp = &bind -> next;
	bind -> next = NULL;

	emalloc (v, struct type_SNMP_VarBind *, sizeof *v, "snmp_assoc_scan3");
	bind -> VarBind = v;
	if (inst) {
	    if ((v -> name = oid_extend (ot -> ot_name, inst -> oid_nelem))
		    == NULL)
		fatal ("oid_extend: out of memory");
	    ip = v -> name -> oid_elements + v -> name -> oid_nelem 
					   - inst -> oid_nelem;
	    jp = inst -> oid_elements;
	    for (i = inst -> oid_nelem; i > 0; i--)
		*ip++ = *jp++;
	}
	else
	    if ((v -> name = oid_cpy (ot -> ot_name)) == NULL)
		fatal ("oid_cpy: out of memory");
	v -> value = NULL;

	emalloc (bind, struct type_SNMP_VarBindList *, sizeof *bind,
		 "snmp_assoc_scan4");
	*vp2 = bind, vp2 = &bind -> next;
	bind -> next = NULL;

	emalloc (v2, struct type_SNMP_VarBind *, sizeof *v2,
		 "snmp_assoc_scan5");
	bind -> VarBind = v2;
	if ((v2 -> name = oid_cpy (v -> name)) == NULL)
	    fatal ("oid_cpy: out of memory");
	v2 -> value = NULL;
    }

    if (head == NULL)
	head = tail = s;
    else {
	tail -> s_next = s;
	s -> s_prev = tail;
	tail = s;
    }

    lookat -> snmp = (caddr_t) s;
    snmp_assoc_next (lookat, 0);
}

/*  */

void	snmp_assoc_next (lookat, alldone)
struct search *lookat;
int	alldone;
{
    int	    i;
    char   *cp;
    register struct snmp_search *s = (struct snmp_search *) lookat -> snmp;
    struct OIDentifier	oids;
    OID	    oid;
    OT	    ot = s -> s_parent;
    register struct type_SNMP_VarBind *v;

    unref (lookat -> retval);
    lookat -> retval = NULL;

    if (alldone
	    || snmp_get_next (s) == NOTOK
	    || s -> s_reqs[0].r_bindings == NULL) {
	register struct snmp_req *sr;

	if (s -> s_prototype)
	    free_SNMP_VarBindList (s -> s_prototype);
	for (sr = s -> s_reqs; sr -> r_bindings; sr++) {
	    free_SNMP_VarBindList (sr -> r_bindings);
	    if (sr -> r_pb)
		pe_free (sr -> r_pb);
	    if (sr -> r_authmsg) {
		if (dstParty -> party_tDomain == RFC1157DOMAIN)
		    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *)
							    sr -> r_authmsg);
		else
		    free_SNMP_SnmpAuthMsg (sr -> r_authmsg);
	    }
	    if (sr -> r_pe)
		pe_free (sr -> r_pe);
	}

	if (tail != s)
	    fatal ("snmp_assoc_next: internal error1");
	if (tail = s -> s_prev)
	    tail -> s_next = NULL;
	else
	    head = NULL;

	free ((char *) s);
	return;
    }

    if ((v = s -> s_reqs[0].r_bindings -> VarBind) == NULL
	     || (oid = v -> name) == NULL)
	fatal ("snmp_assoc_next: internal error2");
    if (ot -> ot_name -> oid_nelem >= oid -> oid_nelem
	    || bcmp ((char *) ot -> ot_name -> oid_elements,
		     (char *) oid -> oid_elements,
		     ot -> ot_name -> oid_nelem
		             * sizeof oid -> oid_elements[0]))
	fatal ("snmp_assoc_next: internal error3");

    oids.oid_nelem = oid -> oid_nelem - (i = ot -> ot_name -> oid_nelem + 1);
    oids.oid_elements = oid -> oid_elements + i;

    cp = sprintoid (&oids);
    lookat -> retval = make_string (cp, strlen (cp));
}

/*  */

static int  snmp_get_next (s)
register struct snmp_search *s;
{
    register struct type_SNMP_VarBindList  *vp,
					   *vp2,
					  **vpp,
					  **vpp2;
    register struct snmp_req *sr,
			     *sp;

    if (snmp_ready (0, PARTY_PRIV_NEXT) == NOTOK
	    || snmp_get_next_aux (s) == NOTOK)
	return NOTOK;

    vpp = &s -> s_prototype, vp = NULL;
    for (sr = s -> s_reqs; sr -> r_bindings; sr++) {
	if (dstParty -> party_tDomain == RFC1157DOMAIN)
	    vpp2 = &(((struct type_SNMP_Snmp1157Msg *) sr -> r_authmsg)
			    -> data -> un.get__request -> variable__bindings);
	else
	    vpp2 = &sr -> r_authmsg -> authData -> pdu -> un.get__request
							-> variable__bindings;

	while ((vp = *vpp) && (vp2 = *vpp2)) {
	    OID	    v = vp -> VarBind -> name,
		    v2 = vp2 -> VarBind -> name;

	    if (v -> oid_nelem > v2 -> oid_nelem
		    || bcmp ((char *) v -> oid_elements,
			     (char *) v2 -> oid_elements,
			     v -> oid_nelem
			         * sizeof v -> oid_elements[0])
		    || vp2 -> VarBind -> value -> pe_class == PE_CLASS_CONT) {
		*vpp = vp -> next;
		vp -> next = NULL;
		free_SNMP_VarBindList (vp);

		*vpp2 = vp2 -> next;
		vp2 -> next = NULL;
		free_SNMP_VarBindList (vp2);
	    }
	    else
		vpp = &vp -> next, vpp2 = &vp2 -> next;
	}

	if (vp == NULL && (vp2 = *vpp)) {
	    snmp_diag (NULLCP, "too many responses starting with: %s\n",
		       oid2ode (vp2 -> VarBind -> name));
	    return NOTOK;
	}
    }
    if (vp) {
	snmp_diag (NULLCP, "missing variable in response");
	return NOTOK;
    }

    for (sp = s -> s_reqs; sp < sr; sp++) {
	register struct type_SNMP_PDUs *pdu;

	if (sp -> r_bindings)
	    free_SNMP_VarBindList (sp -> r_bindings);
	if (sp -> r_pb)
	    pe_free (sp -> r_pb);

	if (dstParty -> party_tDomain == RFC1157DOMAIN)
	    pdu = ((struct type_SNMP_Snmp1157Msg *) sp -> r_authmsg) -> data;
	else
	    pdu = sp -> r_authmsg -> authData -> pdu;
	if (sp -> r_bindings && pdu -> un.get__request -> variable__bindings) {
	    sp -> r_bindings = pdu -> un.get__request -> variable__bindings;
	    pdu -> un.get__request -> variable__bindings = NULL;

	    sp -> r_pb = sp -> r_pe, sp -> r_pe = NULL;

	    if (dstParty -> party_tDomain == RFC1157DOMAIN)
		free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *)
							    sp -> r_authmsg);
	    else
		free_SNMP_SnmpAuthMsg (sp -> r_authmsg);
	    sp -> r_authmsg = NULL;
	    continue;
	}

	if (dstParty -> party_tDomain == RFC1157DOMAIN)
	    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *)
							    sp -> r_authmsg);
	else
	    free_SNMP_SnmpAuthMsg (sp -> r_authmsg);
	if (sp -> r_pe)
	    pe_free (sp -> r_pe);

	bzero ((char *) sp, sizeof *sp);
    }

    {
	register struct snmp_req *sz;

	for (sp = sz = s -> s_reqs; sp < sr; sp++)
	    if (sp -> r_bindings)
		if (sz != sp)
		    *sz++ = *sp;	/* struct copy */
		else
		    sz++;

	bzero ((char *) sz, sizeof *sz);
    }

    return OK;
}

/*  */

static int  snmp_get_next_aux (s)
register struct snmp_search *s;
{
    int	    gotone,
	    result,
	    retries,
	    status = -1;
    struct type_SNMP_SnmpAuthMsg *authmsg;
    struct type_SNMP_Snmp1157Msg *msg;
    struct type_SNMP_SnmpPrivMsg *privmsg = NULL;
    register struct type_SNMP_PDU *parm;
    struct type_SNMP_VarBindList  *vp;
    register struct snmp_req *sr;
    struct qbuf *qb = NULL;
    struct snmp_party *sParty,
		      *dParty;
    PE	    p = NULLPE;

    vp = authmsgs.authData -> pdu -> un.get__request -> variable__bindings;
    authmsgs.authData -> pdu -> offset = type_SNMP_PDUs_get__next__request;
    for (sr = s -> s_reqs; sr -> r_bindings; sr++)
	if ((result = next_ready (sr, 1)) == NOTOK)
	    break;
    authmsgs.authData -> pdu -> un.get__request -> variable__bindings = vp;
    authmsgs.authData -> pdu -> offset = type_SNMP_PDUs_get__request;

    if (result == NOTOK)
	return NOTOK;

    authmsg = NULL, msg = NULL, gotone = 0;
    for (retries = snmp_retries; retries > 0; ) {
	int	len;
	fd_set	rfds;

	for (sr = s -> s_reqs; sr -> r_bindings; sr++) {
	    if (sr -> r_authmsg)
		continue;

	    if (debug > 1) {
		if (dstParty -> party_tDomain == RFC1157DOMAIN)
		    (void) print_SNMP_Snmp1157Msg (sr -> r_pe, 1, NULLIP,
						   NULLVP, NULLCP);
		else
		    (void) print_SNMP_SnmpPrivMsg (sr -> r_pe, 1, NULLIP,
						   NULLVP, NULLCP);
	    }
	    len = ps -> ps_byteno;
	    if (pe2ps (ps, sr -> r_pe) == NOTOK) {
		snmp_diag (NULLCP, "pe2ps: %s", ps_error (ps -> ps_errno));
		goto error_x;
	    }
	    if ((sr -> r_toobig = ((len = ps -> ps_byteno - len)
				       > dstParty -> party_MaxMessageSize))
		    && debug > 0)
		fprintf (stderr, "sent message of %d octets\n", len);
	}

	FD_ZERO (&rfds);
	FD_SET (snmp_fd, &rfds);

	switch (xselect (snmp_fd + 1, &rfds, NULLFD, NULLFD, snmp_timeout)) {
	    case NOTOK:
	        snmp_diag ("failed", "xselect");
		goto error_x;

	    default:
		if (FD_ISSET (snmp_fd, &rfds))
		    break;
		/* else fall... */
	    case OK:
		for (sr = s -> s_reqs; sr -> r_bindings; sr++)
		    if (sr -> r_toobig)
			goto make_smaller;
		retries--;
		continue;
	}

again: ;
	if (dstParty -> party_tDomain == RFC1157DOMAIN) {
	    if (snmp_recv_v1_msg (&msg, &p) == NOTOK)
		goto out;

	    if (msg -> data -> offset != type_SNMP_PDUs_response) {
		snmp_diag (NULLCP, "unexpected message type %d",
			   msg -> data -> offset);
		goto out;
	    }
	    gotone++;

	    parm = msg -> data -> un.response;
	}
	else {
	    if (snmp_recv_msg (&privmsg, &authmsg, &sParty, &dParty, &p, &qb)
	            == NOTOK)
		goto out;

	    if (authmsg -> authData -> pdu -> offset
		    != type_SNMP_PDUs_response) {
		snmp_diag (NULLCP, "unexpected message type %d",
			   authmsg -> authData -> pdu -> offset);
		goto out;
	    }
	    gotone++;

	    parm = authmsg -> authData -> pdu -> un.response;
	}
	for (sr = s -> s_reqs; sr -> r_bindings; sr++)
	    if (parm -> request__id == sr -> r_id)
		break;
	if (!sr -> r_bindings || sr -> r_authmsg) {
	    if (debug > 0)
		fprintf (stderr, "%s response ID (got %ld)\n",
			 sr -> r_bindings ? "duplicate" : "unexpected",
			 (long) parm -> request__id);

	    if (p)
		pe_free (p), p = NULL;
	    if (qb)
		qb_free (qb), qb = NULL;
	    if (privmsg)
		free_SNMP_SnmpPrivMsg (privmsg), privmsg = NULL;
	    if (authmsg)
		free_SNMP_SnmpAuthMsg (authmsg), authmsg = NULL;
	    if (msg)
		free_SNMP_Snmp1157Msg (msg), msg = NULL;
	    goto check;
	}

	switch (status = parm -> error__status) {
	    case int_SNMP_error__status_noError:
	        break;

	    case int_SNMP_error__status_tooBig:
		{
		    register int    i;
		    register struct type_SNMP_VarBindList **vpp;
		    register struct snmp_req *sp;
		    struct snmp_req *sz;

make_smaller: ;
		    i = 0;
		    for (vp = sr -> r_bindings; vp; vp = vp -> next)
			i++;
		    if ((i >>= 1) < 1) {
			snmp_diag (NULLCP,
				   "%s for request with single variable",
				   snmp_error (status));
			goto out;
		    }
		    sz =  s -> s_reqs + NREQ;
		    for (sp = sr + 1; sp -> r_bindings; sp++)
			if (sp >= sz) {
			    snmp_diag (NULLCP,
				       "too many operations needed (%d max)",
				       NREQ - 1);
			    goto out;
			}
		    for (sz = sr + 1; sp > sz; sp--)
			*sp = *(sp - 1);	/* struct copy */
		    bzero ((char *) sp, sizeof *sp);

		    for (vpp = &sr -> r_bindings;
			     i-- > 0;
			     vpp = &((*vpp) -> next))
			continue;
		    sp -> r_bindings = *vpp, *vpp = NULL;

		    vp = authmsgs.authData -> pdu -> un.get__request
						  -> variable__bindings;
		    authmsgs.authData -> pdu -> offset =
					    type_SNMP_PDUs_get__next__request;
		    if ((result = next_ready (sr, 0)) != NOTOK)
			result = next_ready (sp, 0);
		    authmsgs.authData -> pdu -> un.get__request
					     -> variable__bindings = vp;
		    authmsgs.authData -> pdu -> offset =
						type_SNMP_PDUs_get__request;
		    if (result == NOTOK)
			goto out;

		    retries = snmp_retries;
		}
		goto check;

	    default:
		{
		    char   *cp = snmp_variable (parm, parm -> error__index);

		    snmp_diag (NULLCP, cp ? "%s at position %d (%s)"
			       		  : "%s at position %d",
			       snmp_error (status), parm -> error__index, cp);
		}
		goto out;
	}

	pe_free (sr -> r_pe);
	sr -> r_pe = p, p = NULLPE;
	if (qb)
	    qb_free (qb), qb = NULL;
	if (privmsg)
	    free_SNMP_SnmpPrivMsg (privmsg), privmsg = NULL;

	if (dstParty -> party_tDomain == RFC1157DOMAIN)
	    sr -> r_authmsg = (struct type_SNMP_SnmpAuthMsg *) msg, msg = NULL;
	else
	    sr -> r_authmsg = authmsg, authmsg = NULL;

	for (sr = s -> s_reqs; sr -> r_bindings; sr++)
	    if (!sr -> r_authmsg)
		break;
	if (!sr -> r_bindings) {
	    ERROR_node -> var_value = make_number ((AWKNUM) status);

	    return OK;
	}

check: ;
	FD_ZERO (&rfds);
	FD_SET (snmp_fd, &rfds);

	if (select_dgram_socket (snmp_fd + 1, &rfds, NULLFD, NULLFD, 1) > OK)
	    goto again;
    }
    snmp_diag (NULLCP,
	       "no %sresponse within %d retries of %s%d second%s each",
	       gotone ? "acceptable " : "", snmp_retries + gotone,
	       gotone ? "up to " : "", snmp_timeout,
	       snmp_timeout != 1 ? "s" : "");
    return NOTOK;

error_x: ;
    if (ps)
	ps_free (ps), ps = NULLPS;
    if (snmp_fd != NOTOK)
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;

out: ;
    if (p)
	pe_free (p);
    if (qb)
	qb_free (qb);
    if (privmsg)
	free_SNMP_SnmpPrivMsg (privmsg);
    if (authmsg)
	free_SNMP_SnmpAuthMsg (authmsg);
    if (msg)
	free_SNMP_Snmp1157Msg (msg);

    return NOTOK;
}

/*  */

static int  next_ready (sr, do_val)
register struct snmp_req *sr;
int	do_val;
{
    register struct type_SNMP_PDU *parm = authmsgs.authData
						  -> pdu -> un.get__request;

    if (do_val) {
	register struct type_SNMP_VarBindList *vp;

	for (vp = sr -> r_bindings; vp; vp = vp -> next) {
	    register struct type_SNMP_VarBind *v = vp -> VarBind;

	    if (v -> value)
		pe_free (v -> value);
	    (v -> value = unSpecified) -> pe_refcnt++;
	}
    }

    if (snmp_id >= 0x7fffffff)
	snmp_id = 0;
    snmp_id++;

    parm -> request__id = sr -> r_id = snmp_id;
    parm -> variable__bindings = sr -> r_bindings;

    if (sr -> r_authmsg) {
	if (dstParty -> party_tDomain == RFC1157DOMAIN)
	    free_SNMP_Snmp1157Msg ((struct type_SNMP_Snmp1157Msg *)
							    sr -> r_authmsg);
	else
	    free_SNMP_SnmpAuthMsg (sr -> r_authmsg);
	sr -> r_authmsg = NULL;
    }
    if (sr -> r_pe)
	pe_free (sr -> r_pe), sr -> r_pe = NULL;

    if (dstParty -> party_tDomain == RFC1157DOMAIN)
	return snmp_build_v1_msg (&msgs, &sr -> r_pe);
    else
	return snmp_build_msg (&authmsgs, &sr -> r_pe);
}

/*    SET */

void	snmp_set ()
{
    int	    gotone,
	    result,
	    retries,
	    status = -1;
    struct type_SNMP_SnmpAuthMsg *authmsg = &authmsgs;
    struct type_SNMP_Snmp1157Msg *msg = &msgs;
    struct type_SNMP_SnmpPrivMsg *privmsg = NULL;
    register struct type_SNMP_PDU *parm =
				authmsg -> authData -> pdu -> un.get__request;
    register struct type_SNMP_VarBind *v =
					parm -> variable__bindings -> VarBind;
    struct qbuf *qb = NULL;
    struct snmp_party *sParty,
		      *dParty;
    PE	    pe = NULLPE,
	    p = NULLPE,
	   *pep = NULLPEP;
    OID	    oid;
    NODE   *ptr = SET_node,
	   *value = NULL;
    OT	    ot = (OT) ptr -> magic;

    SET_node = NULL;
    if (snmp_ready (1, PARTY_PRIV_SET) == NOTOK)
	goto out;

    parm -> request__id = snmp_id;
    if (v -> name)
	free_SNMP_ObjectName (v -> name), v -> name = NULL;

    if (!SET_inst) {
	register struct snmp_search *s;

	for (s = tail; s; s = s -> s_prev) {
	    register struct snmp_req *sr;

	    if (ot -> ot_name -> oid_nelem
		        != (oid = s -> s_parent -> ot_name) -> oid_nelem + 1
		    || bcmp ((char *) ot -> ot_name -> oid_elements,
			     (char *) oid -> oid_elements,
			     oid -> oid_nelem
			         * sizeof ot -> ot_name -> oid_elements[0]))
		continue;
	    for (sr = s -> s_reqs; sr -> r_bindings; sr++) {
		register struct type_SNMP_VarBindList *vp;

		for (vp = sr -> r_bindings; vp; vp = vp -> next) {
		    register struct type_SNMP_VarBind *vv;

		    if (ot -> ot_name -> oid_nelem
			    >= (vv = vp -> VarBind) -> name -> oid_nelem)
			fatal ("snmp_set: internal error");
		    if (bcmp ((char *) vv -> name -> oid_elements,
			      (char *) ot -> ot_name -> oid_elements,
			      ot -> ot_name -> oid_nelem
			          * sizeof ot -> ot_name -> oid_elements[0]))
			continue;

		    if ((oid = v -> name = oid_cpy (vv -> name)) == NULL)
			goto no_mem_for_inst;
		    pep = &vv -> value;
		    goto set_value;
		}
	    }
	    status = int_SNMP_error__status_noSuchName;
	    goto out;
	}
	if (ot -> ot_getfnx) {
	    snmp_diag (NULLCP,
		       "can't use SNMP array variable as scalar unless within for-in construct");
	    goto out;
	}

	if ((oid = v -> name = oid_extend (ot -> ot_name, 1)) == NULL) {
no_mem_for_inst: ;
	    snmp_diag (NULLCP, "oid_extend: out of memory");
	    goto out;
	}
	v -> name -> oid_elements[v -> name -> oid_nelem - 1] = 0;
    }
    else {
	register int	i;
	register unsigned int *ip,
			      *jp;
	char   *instname = force_string (SET_inst) -> stptr;
	OID	inst = str2oid (instname);

	if (inst == NULL) {
	    snmp_diag (NULLCP, "str2oid: bad instance identifier \"%s\"",
		       instname);
	    goto out;
	}
	if ((oid = v -> name = oid_extend (ot -> ot_name, inst -> oid_nelem))
	        == NULL)
	    goto no_mem_for_inst;
	ip = oid -> oid_elements + oid -> oid_nelem - inst -> oid_nelem;
	jp = inst -> oid_elements;
	for (i = inst -> oid_nelem; i > 0; i--)
	    *ip++ = *jp++;
    }
set_value: ;

    if ((*ot -> ot_syntax -> os_encode) (ptr -> var_value, &pe) == NOTOK) {
	snmp_diag (NULLCP, "encode error for variable %s: %s",
		 oid2ode (v -> name), PY_pepy);
	goto out;
    }

    authmsg -> authData -> pdu -> offset = type_SNMP_PDUs_set__request;
    p = v -> value, v -> value = pe, pe = NULL;
    if (dstParty -> party_tDomain == RFC1157DOMAIN)
	result = snmp_build_v1_msg (msg, &pe);
    else
	result = snmp_build_msg (authmsg, &pe);
    authmsg -> authData -> pdu -> offset = type_SNMP_PDUs_get__request;
    pe_free (v -> value), v -> value = p, p = NULLPE;
    if (result == NOTOK)
	goto out;

    authmsg = NULL, msg = NULL, gotone = 0;
    for (retries = 1 /* snmp_retries */; retries > 0; ) {
	int	len;
	fd_set	rfds;

	if (debug > 1) {
	    if (dstParty -> party_tDomain == RFC1157DOMAIN)
		print_SNMP_Snmp1157Msg (pe, 1, NULLIP, NULLVP, NULLCP);
	    else
		print_SNMP_SnmpPrivMsg (pe, 1, NULLIP, NULLVP, NULLCP);
	}
	len = ps -> ps_byteno;
	if (pe2ps (ps, pe) == NOTOK) {
	    snmp_diag (NULLCP, "pe2ps: %s", ps_error (ps -> ps_errno));
	    goto error_x;
	}
	if (debug > 0
	        && (len = ps -> ps_byteno - len)
	    		> dstParty -> party_MaxMessageSize)
	    fprintf (stderr, "sent message of %d octets\n", len);

	FD_ZERO (&rfds);
	FD_SET (snmp_fd, &rfds);

	switch (xselect (snmp_fd + 1, &rfds, NULLFD, NULLFD, snmp_timeout)) {
	    case NOTOK:
	        snmp_diag ("failed", "xselect");
		goto error_x;

	    default:
		if (FD_ISSET (snmp_fd, &rfds))
		    break;
		/* else fall... */
	    case OK:
		if (debug > 0)
		    fprintf (stderr, "timeout...\n");
		retries--;
		continue;
	}

	if (dstParty -> party_tDomain == RFC1157DOMAIN) {
	    if (snmp_recv_v1_msg (&msg, &p) == NOTOK)
		goto out;

	    if (msg -> data -> offset != type_SNMP_PDUs_response) {
		snmp_diag (NULLCP, "unexpected message type %d",
			   msg -> data -> offset);
		goto out;
	    }

	    if ((parm = msg -> data -> un.response) -> request__id == snmp_id)
		break;
	}
	else {
	    if (snmp_recv_msg (&privmsg, &authmsg, &sParty, &dParty, &p, &qb)
	            == NOTOK)
		goto out;

	    if (authmsg -> authData -> pdu -> offset
		    != type_SNMP_PDUs_response) {
		snmp_diag (NULLCP, "unexpected message type %d",
			   authmsg -> authData -> pdu -> offset);
		goto out;
	    }

	    if ((parm = authmsg -> authData -> pdu -> un.response)
						    -> request__id == snmp_id)
	    break;
	}

	if (debug > 0)
	    fprintf (stderr, "bad ID (got %ld, wanted %ld)\n",
		     (long) parm -> request__id, (long) snmp_id);

	if (p)
	    pe_free (p), p = NULL;
	if (qb)
	    qb_free (qb), qb = NULL;
	if (privmsg)
	    free_SNMP_SnmpPrivMsg (privmsg), privmsg = NULL;
	if (authmsg)
	    free_SNMP_SnmpAuthMsg (authmsg), authmsg = NULL;
	if (msg)
	    free_SNMP_Snmp1157Msg (msg), msg = NULL;

	gotone++;
    }
    if (retries <= 0) {
	snmp_diag (NULLCP,
		   "no %sresponse within %d retries of %s%d second%s each",
		   gotone ? "acceptable " : "", 1 /* snmp_retries */ + gotone,
		   gotone ? "upto " : "",
		   snmp_timeout, snmp_timeout != 1 ? "s" : "");
	goto out;
    }

    if ((status = parm -> error__status) != int_SNMP_error__status_noError) {
	char   *cp = snmp_variable (parm, parm -> error__index);

	snmp_diag (NULLCP, cp ? "%s at position %d (%s)" : "%s at position %d",
		   snmp_error (status), parm -> error__index, cp);
	goto out;
    }

    if (parm -> variable__bindings == NULL
	    || (v = parm -> variable__bindings -> VarBind) == NULL) {
	snmp_diag (NULLCP, "missing variable in response");
	goto out;
    }
    if (debug > 0 && parm -> variable__bindings -> next)
	fprintf (stderr, "too many responses starting with: %s\n",
		 oid2ode (parm -> variable__bindings -> next -> VarBind -> name));

    if (oid_cmp (oid, v -> name)) {
	char    buffer[BUFSIZ];

	(void) strcpy (buffer, oid2ode (v -> name));
	snmp_diag (NULLCP, "wrong variable returned (got %s, wanted %s)",
		 buffer, oid2ode (oid));
	goto out;
    }
		 
/* needed in case of x = y = value, where "y" is an SNMP variable */
    if (v -> value -> pe_class == PE_CLASS_CONT) {
	snmp_diag (NULLCP, "exception for variable %s: %s",
		 oid2ode (v -> name), snmp_exception (v -> value -> pe_id));
	goto out;
    }
    if ((*ot -> ot_syntax -> os_decode) (&value, v -> value) == NOTOK) {
	snmp_diag (NULLCP, "decode error for variable %s: %s",
		 oid2ode (v -> name), PY_pepy);
	goto out;
    }
    if (pep) {
	pe_free (*pep);
	*pep = v -> value, v -> value = NULL;
    }

    goto out;

error_x: ;
    if (ps)
	ps_free (ps), ps = NULLPS;
    if (snmp_fd != NOTOK)
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;

out: ;
    if (p)
	pe_free (p);
    if (qb)
	qb_free (qb);
    if (privmsg)
	free_SNMP_SnmpPrivMsg (privmsg);
    if (authmsg && authmsg != &authmsgs)
	free_SNMP_SnmpAuthMsg (authmsg);
    if (msg && msg != &msgs)
	free_SNMP_Snmp1157Msg (msg);
    if (pe)
	pe_free (pe);

    unref (ptr -> var_value);
/* needed in case of x = y = value, where "y" is an SNMP variable */
    ptr -> var_value = value ? value : Nnull_string;

    ERROR_node -> var_value = make_number ((AWKNUM) status);

    if (SET_inst) {
	unref (SET_inst);
	SET_inst = NULL;
    }
}

/*    ENCODE */

static int  e_integer (x, pe)
NODE   *x;
PE     *pe;
{
    if ((*pe = int2prim ((integer) force_number (x))) == NULLPE) {
	(void) strcpy (PY_pepy, "int2prim failed");
	return NOTOK;
    }

    return OK;
}


extern char 	hex2nib[];

static int  e_octets (x, pe)
NODE   *x;
PE     *pe;
{
    int	    i;
    register char *cp,
		  *ep;
    char   *tmp = force_string (x) -> stptr;
    PE	    p;
    PElementData d;

    if ((i = strlen (tmp)) == 0
	    || i % 3 != 2
	    || !isxdigit (tmp[0])
	    || !isxdigit (tmp[1])
	    || !isxdigit (tmp[i - 2])
	    || !isxdigit (tmp[i - 1]))
	return e_display (x, pe);
    for (cp = tmp + 3, ep = tmp + i - 3; cp < ep; cp += 3)
	if (!isxdigit (cp[0]) || !isxdigit (cp[1]) || cp[2] != ':')
	    return e_display (x, pe);

    if ((p = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_OCTS)) == NULL) {
	(void) strcpy (PY_pepy, "pe_alloc failed");
	return NOTOK;
    }
    if ((d = p -> pe_prim = PEDalloc (p -> pe_len = (i / 3) + 1)) == NULL) {
	(void) strcpy (PY_pepy, "PEDalloc failed");
	pe_free (p);
	return NOTOK;
    }
    for (ep = (cp = tmp) + i; cp < ep; cp += 3)
	*d++ = (hex2nib[cp[0] & 0x7f] << 4) | (hex2nib[cp[1] & 0x7f]);

    *pe = p;
    return OK;
}


static int  e_display (x, pe)
NODE   *x;
PE     *pe;
{
    char   *tmp = force_string (x) -> stptr;

    if ((*pe = str2prim (tmp, strlen (tmp), PE_CLASS_UNIV, PE_PRIM_OCTS))
	    == NULLPE) {
	(void) strcpy (PY_pepy, "str2prim failed");
	return NOTOK;
    }

    return OK;
}


static int  e_objectID (x, pe)
NODE   *x;
PE     *pe;
{
    char   *tmp = force_string (x) -> stptr;
    OID	    oid = str2oid (tmp);

    if (oid == NULLOID) {
	(void) sprintf (PY_pepy, "str2oid: bad instance identifier \"%s\"",
			tmp);
	return NOTOK;
    }
    
    if ((*pe = oid2prim (oid)) == NULLPE) {
	(void) strcpy (PY_pepy, "oid2prim failed");
	return NOTOK;
    }

    return OK;
}


static int  e_bits (x, pe)
NODE   *x;
PE     *pe;
{
    int	    i;
    register char *cp;
    char   *tmp = force_string (x) -> stptr;

    for (cp = tmp; *cp; cp++)
	if (*cp != '1' || *cp != '0') {
	    (void) strcpy (PY_pepy, "invalid bit string");
	    return NOTOK;
	}

    if ((*pe = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_BITS))
	    == NULLPE) {
	(void) strcpy (PY_pepy, "pe_alloc failed");
	return NOTOK;
    }
    if ((i = cp - tmp) > 0 && bit_off (*pe, i - 1) == NOTOK) {
	(void) strcpy (PY_pepy, "bit_off failed: %s",
		       pe_error ((*pe) -> pe_errno));
	return NOTOK;
    }
    for (cp = tmp; *cp; cp++)
	if (*cp == '1' && bit_on (*pe, cp - tmp) == NOTOK) {
	    (void) strcpy (PY_pepy, "bit_on failed: %s",
			   pe_error ((*pe) -> pe_errno));
	    return NOTOK;
	}

    return OK;
}


#ifdef	notdef
/* ARGSUSED */

static int  e_null (x, pe)
NODE   *x;
PE     *pe;
{
    if ((*pe = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_NULL))
	    == NULLPE) {
	(void) strcpy (PY_pepy, "pe_alloc failed");
	return NOTOK;
    }

    return OK;
}
#endif


static int  e_counter (x, pe)
NODE   *x;
PE     *pe;
{
    counter64	c64;

    c64.hiword = 0;
    c64.loword = (u_long) force_number (x);
    if ((*pe = _ulong2prim (&c64, PE_CLASS_APPL, 1)) == NULL) {
	(void) strcpy (PY_pepy, "_ulong2prim failed");
	return NOTOK;
    }

    return OK;
}


static int  e_counter64 (x, pe)
NODE   *x;
PE     *pe;
{
    counter64	c64;
    AWKNUM	d;

    if ((d = force_number (x)) < 0.0) {
	(void) strcpy (PY_pepy, "e_counter64: negative number %.f", d);
	return NOTOK;
    }
    
    if (d > 4294967295.0) {
	register int	i;
	double	e = d;

	for (i = sizeof c64.loword * 8; i > 0; i--)
	    d /= 2.0;
	c64.hiword = d;
	for (i = sizeof c64.loword * 8; i > 0; i--)
	    d *= 2.0;
	c64.loword = e - d;	
    }
    else
	c64.hiword = 0L, c64.loword = d;

    if ((*pe = _ulong2prim (&c64, PE_CLASS_APPL, 6)) == NULL) {
	(void) strcpy (PY_pepy, "_ulong2prim failed");
	return NOTOK;
    }

    return OK;
}


static int  e_gauge (x, pe)
NODE   *x;
PE     *pe;
{
    counter64	c64;

    c64.hiword = 0;
    c64.loword = (u_long) force_number (x);
    if ((*pe = _ulong2prim (&c64, PE_CLASS_APPL, 2)) == NULL) {
	(void) strcpy (PY_pepy, "_ulong2prim failed");
	return NOTOK;
    }

    return OK;
}


static int  e_uint (x, pe)
NODE   *x;
PE     *pe;
{
    counter64	c64;

    c64.hiword = 0;
    c64.loword = (u_long) force_number (x);
    if ((*pe = _ulong2prim (&c64, PE_CLASS_APPL, 7)) == NULL) {
	(void) strcpy (PY_pepy, "_ulong2prim failed");
	return NOTOK;
    }

    return OK;
}


static int  e_timeticks (x, pe)
NODE   *x;
PE     *pe;
{
    counter64	c64;

    c64.hiword = 0;
    c64.loword = (u_long) force_number (x);
    if ((*pe = _ulong2prim (&c64, PE_CLASS_APPL, 3)) == NULL) {
	(void) strcpy (PY_pepy, "_ulong2prim failed");
	return NOTOK;
    }

    return OK;
}


static int  e_ipaddr (x, pe)
NODE   *x;
PE     *pe;
{
    char   *tmp = force_string (x) -> stptr;
    struct hostent *hp;
    struct sockaddr_in lo_socket;
    register struct sockaddr_in *lsock = &lo_socket;

    if (!(hp = gethostbystring (tmp))) {
	(void) sprintf (PY_pepy, "invalid IP-address: \"%s\"", tmp);
	return NOTOK;
    }
    bzero ((char *) lsock, sizeof *lsock);
    lsock -> sin_family = hp -> h_addrtype;
    inaddr_copy (hp, lsock);
    
    if ((*pe = str2prim ((char *) &lsock -> sin_addr, 4, PE_CLASS_APPL, 0))
	    == NULLPE) {
	(void) strcpy (PY_pepy, "str2prim failed");
	return NOTOK;
    }

    return OK;
}


static int  e_nsapaddr (x, pe)
NODE   *x;
PE     *pe;
{
    if (e_octets (x, pe) == NOTOK)
	return NOTOK;
    (*pe) -> pe_class = PE_CLASS_APPL, (*pe) -> pe_id = 5;

    return OK;
}

/*    DECODE */

static NODE *make_octet_node (base, len)
char   *base;
int	len;
{
    register char *bp,
		  *cp,
		  *ep;
    char   *s = "";
    register NODE *r;

    getnode (r);
    r -> type = Node_val;
    emalloc (r -> stptr, char *, len * 3 + 1, "make_octet_node");
    bp = r -> stptr;
    for (ep = (cp = base) + len; cp < ep; cp++, s = ":") {
	(void) sprintf (bp, "%s%02x", s, *cp & 0xff);
	bp += strlen (bp);
    }
    *bp = NULL;		/* in case len == 0 */
    r -> stlen = bp - r -> stptr;
    r -> stref = 1;
    r -> flags |= STR | MALLOC;

    return r;
}


static int  d_integer (x, pe)
NODE  **x;
PE	pe;
{
    integer	i = prim2num (pe);

    if (i == NOTOK && pe -> pe_errno != PE_ERR_NONE) {
	(void) strcpy (PY_pepy, pe_error (pe -> pe_errno));
	return NOTOK;
    }

    *x = make_number ((AWKNUM) i);

    return OK;
}


static int  d_octets (x, pe)
NODE  **x;
PE	pe;
{
    struct qbuf *qb = prim2qb (pe);

    if (qb == NULL || qb_pullup (qb) == NOTOK) {
	(void) strcpy (PY_pepy, qb ? "qb_pullup: out of memory"
				   : pe_error (pe -> pe_errno));
	return NOTOK;
    }

    *x = make_octet_node (qb -> qb_forw -> qb_data, qb -> qb_forw -> qb_len);

    qb_free (qb);

    return OK;
}


static int  d_display (x, pe)
NODE  **x;
PE	pe;
{
    struct qbuf *qb = prim2qb (pe);

    if (qb == NULL || qb_pullup (qb) == NOTOK) {
	(void) strcpy (PY_pepy, qb ? "qb_pullup: out of memory"
				   : pe_error (pe -> pe_errno));
	return NOTOK;
    }

    *x = make_string (qb -> qb_forw -> qb_data, qb -> qb_forw -> qb_len);

    qb_free (qb);

    return OK;
}


static int  d_objectID (x, pe)
NODE  **x;
PE	pe;
{
    char   *cp;
    OID	    oid = prim2oid (pe);

    if (oid == NULLOID) {
	(void) strcpy (PY_pepy, pe_error (pe -> pe_errno));
	return NOTOK;
    }
    cp = sprintoid (oid);

    *x = make_string (cp, strlen (cp));

    return OK;
}


static int  d_bits (x, pe)
NODE  **x;
PE	pe;
{
    register int i;
    register char *bp;
    register NODE *r;

    if (prim2bit (pe) == NULL) {
	(void) strcpy (PY_pepy, pe_error (pe -> pe_errno));
	return NOTOK;
    }

    getnode (r);
    r -> type = Node_val;
    emalloc (r -> stptr, char *, pe -> pe_nbits + 1, "d_bits");
    for (bp = r -> stptr, i = 0; i < pe -> pe_nbits; i++)
	*bp++ = bit_test (pe, i) == 1 ? '1' : '0';
    *bp = NULL;	
    r -> stlen = bp - r -> stptr;
    r -> stref = 1;
    r -> flags |= STR | MALLOC;

    return OK;
}


#ifdef	notdef
/* ARGSUSED */

static int  d_null (x, pe)
NODE  **x;
PE	pe;
{
    *x = make_string ("NULL", 4);

    return OK;
}
#endif


static int  d_ulong (x, pe)
NODE  **x;
PE	pe;
{
    register int    i;
    AWKNUM	d;
    counter64	c64;

    if (_prim2ulong (pe, &c64, 2 * sizeof (u_long)) == NOTOK) {
	(void) strcpy (PY_pepy, pe_error (pe -> pe_errno));
	return NOTOK;
    }

    if (c64.hiword == 0)
	d = c64.loword;
    else {
	d = c64.hiword;
	for (i = sizeof c64.loword * 8; i > 0; i--)
	    d *= 2.0;
	d += c64.loword;
    }

    *x = make_number (d);

    return OK;
}


static int  d_ipaddr (x, pe)
NODE  **x;
PE	pe;
{
    char    ipaddr[16];
    struct type_SNMP_IpAddress *ip;
    struct qbuf *qb;

    if ((ip = prim2qb (pe)) == NULL) {
	(void) strcpy (PY_pepy, pe_error (pe -> pe_errno));
	return NOTOK;
    }
    if (qb_pullup (ip) == NOTOK) {
	(void) strcpy (PY_pepy, "qb_pullup: out of memory");
	free_SNMP_IpAddress (ip);
	return NOTOK;
    }
    if ((qb = ip -> qb_forw) -> qb_len != 4) {
	(void) sprintf (PY_pepy,
			"IpAddress is wrong length (got %d, wanted 4)",
			qb -> qb_len);
	free_SNMP_IpAddress (ip);
	return NOTOK;
    }
    (void) sprintf (ipaddr, "%d.%d.%d.%d",
		    qb -> qb_data[0] & 0xff, qb -> qb_data[1] & 0xff,
		    qb -> qb_data[2] & 0xff, qb -> qb_data[3] & 0xff);

    *x = make_string (ipaddr, strlen (ipaddr));

    free_SNMP_IpAddress (ip);

    return OK;
}


static int  d_nsapaddr (x, pe)
NODE  **x;
PE	pe;
{
    int	    len;
    struct type_SNMP_NsapAddress *nsap;
    struct qbuf *qb;

    if ((nsap = prim2qb (pe)) == NULL) {
	(void) strcpy (PY_pepy, pe_error (pe -> pe_errno));
	return NOTOK;
    }
    if (qb_pullup (nsap) == NOTOK) {
	(void) strcpy (PY_pepy, "qb_pullup: out of memory");
	free_SNMP_NsapAddress (nsap);
	return NOTOK;
    }
    qb = nsap -> qb_forw;
    if ((len = qb -> qb_data[0] & 0xff) >= qb -> qb_len)
	len = qb -> qb_len - 1;

    *x = make_octet_node (qb -> qb_data + 1, len);

    free_SNMP_NsapAddress (nsap);

    return OK;
}

/*    MISC */

#define	QOS_NONE	0x00
#define	QOS_AUTH	0x01
#define	QOS_PRIV	0x02
#define	QOS_ALL		(QOS_AUTH | QOS_PRIV)


static	snmp_ready (do_id, mask)
int	do_id,
	mask;
{
    int	    changed = 0,
	    qos;
    char   *pp;
    struct sockaddr_in lo_socket;
    register struct sockaddr_in *lsock = &lo_socket;
    register struct sockaddr_in *isock = &in_socket;
    register struct hostent *hp;
    register struct servent *sp;
    register struct snmp_access *sa;
    struct snmp_party *p;
    register NODE   *t0;

    unref (DIAGNOSTIC_node -> var_value);
    DIAGNOSTIC_node -> var_value = Nnull_string;

    if ((snmp_retries = (int) RETRIES_node -> var_value -> numbr) <= 0)
	snmp_retries = 3;
    if ((snmp_timeout = (int) TIMEOUT_node -> var_value -> numbr) <= 0)
	snmp_timeout = 10;

    if (do_id) {
	if (snmp_id >= 0x7fffffff)
	    snmp_id = 0;
	snmp_id++;
    }

    if (snmp_fd == NOTOK || ps == NULLPS)
	changed++;

    t0 = force_string (CONTEXT_node -> var_value);
    if (snmp_ctx == NULL || strcmp (snmp_ctx, t0 -> stptr)) {
	int	i;
	register char *dp;
	struct snmp_context *sc;
	OID	oid;

	for (dp = pp = t0 -> stptr; isdigit (*dp); dp++)
	    continue;
	if (*dp == NULL && sscanf (pp, "%d", &i) == 1) {
	    if (!(sc = index_context (i))) {
		snmp_diag (NULLCP, "no context with index %d", i);
		return NOTOK;
	    }

	    goto test_context;
	}

	i = strlen (pp);
	for (sc = contextHead; sc; sc = sc -> sc_next)
	    if (sc -> sc_name
		    && sc -> sc_name_len == i
		    && !lexequ ((char *) sc -> sc_name, pp))
		goto test_context;

	if (!(oid = text2oid (pp))) {
	    snmp_diag (NULLCP, "no such object as %s", pp);
	    return NOTOK;
	}

	sc = lookup_context (oid);
	oid_free (oid);

	if (!sc) {
	    snmp_diag (NULLCP, "no such context as %s", pp);
	    return NOTOK;
	}

test_context: ;
	if (sc -> sc_location != AGENT) {
	    snmp_diag (NULLCP, "context %s is local", pp);
	    return NOTOK;
	}

	if (snmp_ctx)
	    free (snmp_ctx), snmp_ctx = NULL;
	emalloc (snmp_ctx, char *, strlen (pp) + 1, "snmp_ready");
	(void) strcpy (snmp_ctx, pp);

	theContext = sc;
    }

    for (qos = 0; qos <= QOS_ALL; qos <<= 1, qos |= 1) {
	for (sa = accessHead; sa; sa = sa -> sa_next)
	    if (sa -> sa_status == ROW_ACTIVE
	            && sa -> sa_resources == theContext
	            && (sa -> sa_privileges & mask)
		    && (p = sa -> sa_target)
		    && p -> party_location == AGENT
	            && p -> party_tDomain == SNMPUDPDOMAIN
	            && ((qos & QOS_PRIV) ? p -> party_privProt > NOPRIV
				         : p -> party_privProt <= NOPRIV)
	            && (p = sa -> sa_subject)
		    && p -> party_location == MANAGER
	            && ((qos & QOS_AUTH) ? p -> party_authProt > NOAUTH
				         : p -> party_authProt <= NOAUTH))
		break;
	if (sa)
	    break;
    }
    if (!sa)
	for (sa = accessHead; sa; sa = sa -> sa_next)
	    if (sa -> sa_status == ROW_ACTIVE
	            && sa -> sa_resources == theContext
	            && (sa -> sa_privileges & mask)
	            && (p = sa -> sa_target)
	            && p -> party_location == AGENT
	            && p -> party_tDomain == RFC1157DOMAIN)
		break;
    if (!sa) {
	snmp_diag (NULLCP,
		   "no access control entry allowing operation on context");
	return NOTOK;
    }
    if (sa -> sa_target != dstParty) {
	dstParty = sa -> sa_target;
	changed++;
    }
    srcParty = sa -> sa_subject;

    if (!changed)
	return OK;
    if (ps)
	ps_free (ps), ps = NULLPS;
    if (snmp_fd != NOTOK)
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;

    bzero ((char *) lsock, sizeof *lsock);
    if ((hp = gethostbystring (pp = getlocalhost ())) == NULL) {
	snmp_diag (NULLCP, "%s: unknown host", pp);
	return NOTOK;
    }
    lsock -> sin_family = hp -> h_addrtype;
    inaddr_copy (hp, lsock);

    bzero ((char *) isock, sizeof *isock);
    isock -> sin_family = AF_INET;
    bcopy ((char *) dstParty -> party_TAddress, (char *) &isock -> sin_addr,
	   sizeof isock -> sin_addr);
    bcopy ((char *) (dstParty -> party_TAddress + sizeof isock -> sin_addr),
	   (char *) &isock -> sin_port, sizeof isock -> sin_port);

    if ((snmp_fd = start_udp_client (lsock, 0, 0, 0)) == NOTOK) {
	snmp_diag ("failed", "start_udp_server");
	return NOTOK;
    }

    if (join_udp_server (snmp_fd, isock) == NOTOK) {
	snmp_diag ("failed", "join_udp_server");
	return NOTOK;
    }

    if ((ps = ps_alloc (dg_open)) == NULLPS
	    || dg_setup (ps, snmp_fd, dstParty -> party_MaxMessageSize,
			 read_udp_socket, write_udp_socket,
			 check_udp_socket) == NOTOK) {
	if (ps == NULLPS)
	    snmp_diag (NULLCP, "ps_alloc: failed");
	else
	    snmp_diag (NULLCP, "dg_setup: %s", ps_error (ps -> ps_errno));

	return NOTOK;
    }

#ifndef	SYS5
    snmp_id = ((int) random ()) & 0x7fffffff;
#else
    snmp_id = ((int) rand ()) & 0x7fffffff;
#endif

    return OK;
}

/*  */

static int  snmp_build_msg (authmsg, pe)
register struct type_SNMP_SnmpAuthMsg *authmsg;
PE     *pe;
{
    int	    result = NOTOK;
    struct type_SNMP_SnmpPrivMsg privmsgs;
    struct type_SNMP_SnmpPrivMsg *privmsg = &privmsgs;

    if (*pe)
	pe_free (*pe), *pe = NULL;
    authmsg -> authInfo = NULL;
    authmsg -> authData -> srcParty = srcParty -> party_Identity;
    authmsg -> authData -> dstParty = dstParty -> party_Identity;
    authmsg -> authData -> context = theContext -> sc_identity;

    bzero ((char *) privmsg, sizeof *privmsg);
    privmsg -> privDst = dstParty -> party_Identity;
    if ((privmsg -> privData =
	     generate_outgoing_auth (authmsg, srcParty, dstParty,
				     (caddr_t) (debug > 1 ? -1 : 0))) == NULL){
	snmp_diag (NULLCP, "error generating outgoing authentication: %s",
		   PY_pepy);
	goto out;
    }

    if (do_encryption (privmsg, dstParty) == NOTOK) {
	snmp_diag (NULLCP, "error encrypting: %s", PY_pepy);
	goto out;
    }

    if (encode_SNMP_SnmpPrivMsg (pe, 1, 0, NULLCP, privmsg) != NOTOK)
	result = OK;
    else
	snmp_diag (NULLCP, "encode_SNMP_SnmpPrivMsg: %s", PY_pepy);

out: ;
    if (*pe && result != OK)
	pe_free (*pe), *pe = NULL;

    if (authmsg -> authInfo)
	free_SNMP_AuthInformation (authmsg -> authInfo);
    if (privmsg -> privData)
	qb_free (privmsg -> privData), privmsg -> privData = NULL;

    return result;
}

/*  */

static int  snmp_build_v1_msg (msg, pe)
register struct type_SNMP_Snmp1157Msg *msg;
PE     *pe;
{
    int	    result = NOTOK;

    if (*pe)
	pe_free (*pe), *pe = NULL;

    if (msg -> community)
	qb_free (msg -> community);
    if ((msg -> community = str2qb ((char *) dstParty -> party_AuthPrivate,
				    dstParty -> party_AuthPrivate_len, 1))
	    == NULL) {
	snmp_diag (NULLCP, "unable to allocate community");
	goto out;
    }
   

    if (encode_SNMP_Snmp1157Msg (pe, 1, 0, NULLCP, msg) != NOTOK)
	result = OK;
    else
	snmp_diag (NULLCP, "encode_SNMP_Snmp1157Msg: %s", PY_pepy);

out: ;
    if (*pe && result != OK)
	pe_free (*pe), *pe = NULL;

    return result;
}

/*  */

static int  snmp_recv_msg (privmsg, authmsg, sParty, dParty, pe, qb)
struct type_SNMP_SnmpPrivMsg **privmsg;
register struct type_SNMP_SnmpAuthMsg **authmsg;
register struct snmp_party **sParty,
			   **dParty;
PE	*pe;
struct qbuf **qb;
{
    PS	    qbufPS;

    *privmsg = NULL, *authmsg = NULL;
    *sParty = NULL, *dParty = NULL;
    *pe = NULL, *qb = NULL;

    if ((*pe = ps2pe (ps)) == NULLPE) {
	if (ps)
	    ps_free (ps), ps = NULLPS;
	if (snmp_fd != NOTOK)
	    (void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;

bad_parse: ;
        snmp_diag (NULLCP, "ps2pe: %s", ps_error (ps -> ps_errno));
	goto out;
    }

    if (decode_SNMP_SnmpPrivMsg (*pe, 1, NULLIP, NULLVP, privmsg) == NOTOK) {
	snmp_diag (NULLCP, "decode_SNMP_SnmpPrivMsg: %s", PY_pepy);
	goto out;
    }

    if (debug > 1)
	(void) print_SNMP_SnmpPrivMsg (*pe, 1, NULLIP, NULLVP, NULLCP);

    if ((*dParty = lookup_party ((*privmsg) -> privDst, MANAGER))
	    == NULLPARTY) {
	snmp_diag (NULLCP, "badParty: destination=%s",
		   oid2ode ((*privmsg) -> privDst));
	goto out;
    }

    if ((*qb = do_decryption (*privmsg, *dParty)) == NULL) {
	snmp_diag (NULLCP, "error decrypting: %s", PY_pepy);
	goto out;
    }

    if ((qbufPS = ps_alloc (qbuf_open)) == NULLPS
	    || qbuf_setup (qbufPS, *qb) == NOTOK) {
	if (qbufPS == NULLPS)
	    snmp_diag (NULLCP, "ps_alloc(qbuf_open): failed");
	else
	    snmp_diag (NULLCP, "qbuf_setup: %s", ps_error (ps -> ps_errno));

	goto out;
    }

    if (*pe)
	pe_free (*pe), *pe = NULLPE;

    if ((*pe = ps2pe (qbufPS)) == NULLPE)
	goto bad_parse;
    do_decryption_fixup (*privmsg, *qb);

    if (decode_SNMP_SnmpAuthMsg (*pe, 1, NULLIP, NULLVP, authmsg) == NOTOK) {
	snmp_diag (NULLCP, "decode_SNMP_SnmpAuthMsg: %s", PY_pepy);
	goto out;
    }

    if (debug > 1)
	(void) print_SNMP_SnmpAuthMsg (*pe, 1, NULLIP, NULLVP, NULLCP);

    if ((*sParty = lookup_party ((*authmsg) -> authData -> srcParty, AGENT))
	    == NULLPARTY) {
	snmp_diag (NULLCP, "badParty: source=%s",
		   oid2ode ((*authmsg) -> authData -> srcParty));
	goto out;
    }

    if (authenticate_incoming (*privmsg, *authmsg, *sParty, *dParty, NULLIP)
	    == NOTOK) {
	snmp_diag (NULLCP, "authentication failure: %s", PY_pepy);
	goto out;
    }

    ps_free (qbufPS);
    return OK;

out: ;
    if (*pe)
	pe_free (*pe), *pe = NULLPE;
    if (qbufPS)
	ps_free (qbufPS);
    if (*qb)
	qb_free (*qb), *qb = NULL;
    if (*privmsg)
	free_SNMP_SnmpPrivMsg (*privmsg), *privmsg = NULL;
    if (*authmsg)
	free_SNMP_SnmpAuthMsg (*authmsg), *authmsg = NULL;

    return NOTOK;
}

/*  */

static int  snmp_recv_v1_msg (msg, pe)
register struct type_SNMP_Snmp1157Msg **msg;
PE	*pe;
{
    PS	    qbufPS;

    *msg = NULL;
    *pe = NULL;

    if ((*pe = ps2pe (ps)) == NULLPE) {
	if (ps)
	    ps_free (ps), ps = NULLPS;
	if (snmp_fd != NOTOK)
	    (void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;

bad_parse: ;
        snmp_diag (NULLCP, "ps2pe: %s", ps_error (ps -> ps_errno));
	goto out;
    }

    if (decode_SNMP_Snmp1157Msg (*pe, 1, NULLIP, NULLVP, msg) == NOTOK) {
	snmp_diag (NULLCP, "decode_SNMP_Snmp1157Msg: %s", PY_pepy);
	goto out;
    }

    if (debug > 1)
	(void) print_SNMP_Snmp1157Msg (*pe, 1, NULLIP, NULLVP, NULLCP);

    return OK;

out: ;
    if (*pe)
	pe_free (*pe), *pe = NULLPE;
    if (*msg)
	free_SNMP_Snmp1157Msg (*msg), *msg = NULL;

    return NOTOK;
}

/*  */

#ifndef	lint
static	snmp_diag (va_alist)
va_dcl
{
    char   *what,
	    buffer[BUFSIZ];
    va_list ap;

    va_start (ap);

    what = va_arg (ap, char *);

    _asprintf (buffer, what, ap);

    va_end (ap);

    if (debug > 0)
	fprintf (stderr, "%s\n", buffer);

    unref (DIAGNOSTIC_node -> var_value);
    DIAGNOSTIC_node -> var_value = make_string (buffer, strlen (buffer));
}
#else
/* VARARGS */

static	snmp_diag (what, fmt)
char   *what,
       *fmt;
{
    snmp_diag (what, fmt);
}
#endif

/*  */

char   *snmp_name (ptr)
NODE   *ptr;
{
    return ((OT) (ptr -> magic)) -> ot_text;
}

/*  */

static char *errors[] = {
    "noError",
    "tooBig",
    "noSuchName",
    "badValue",
    "readOnly",
    "genErr",
    "noAccess",
    "wrongType",
    "wrongLength",
    "wrongEncoding",
    "wrongValue",
    "noCreation",
    "inconsistentValue",
    "resourceUnavailable",
    "commitFailed",
    "undoFailed",
    "authorizationError",
    "notWritable",
    "inconsistentName"
};

static char *snmp_error (i)
int	i;
{
    static char buffer[BUFSIZ];

    if (0 <= i && i < sizeof errors / sizeof errors[0])
	return errors[i];
    (void) sprintf (buffer, "error-status %d", i);

    return buffer;
}


static char *exceptions[] = {
    "noSuchObject", "noSuchInstance", "noNextInstance"
};

static char *snmp_exception (i)
PElementID i;
{
    static char buffer[BUFSIZ];

    if (i < sizeof exceptions / sizeof exceptions[0])
	return exceptions[i];
    (void) sprintf (buffer, "exception %u", (unsigned int) i);

    return buffer;
}


static char *snmp_variable (parm, idx)
register struct type_SNMP_PDU *parm;
integer	idx;
{
    register struct type_SNMP_VarBindList *vp;

    if (idx <= 0 || (vp = parm -> variable__bindings) == NULL)
	return NULL;
    for (idx--; idx > 0; idx--)
	if ((vp = vp -> next) == NULL)
	    return NULL;

    return oid2ode (vp -> VarBind -> name);
}


static	OID	str2oid (s)
char    *s;
{
    int	    i;
    static struct OIDentifier   oids;
    static unsigned int elements[NELEM + 1];

#ifdef	original
    if ((i = str2elem (s, elements)) < 1
	    || elements[0] > 2
	    || (i > 1 && elements[0] < 2 && elements[1] > 39))
	return NULLOID;
#else
    if ((i = str2elem (s, elements)) < 1)
	return NULLOID;
#endif

    oids.oid_elements = elements, oids.oid_nelem = i;

    return (&oids);
}
#endif	/* SNMP */
