/* party.c - SNMPv2 party databases */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/* 
 * $Header$
 *
 *
 * $Log$
 */

/*
 *			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#include <ctype.h>
#include <stdio.h>
#include "objects.h"
#include "party.h"
#include "tailor.h"
#include <sys/stat.h>


#define	PUBLIC	"public"
#define PRIVATE	"private"
#define CONTEXT	"context"
#define	ACCESS	"access"
#define	VIEW	"view"

/*  */

#define	ADVISE	if (o_advise) (*o_advise)


	OID	noAuth = NULLOID;
	OID	md5Auth = NULLOID;
	OID	rfc1157noAuth = NULLOID;

	OID	noPriv = NULLOID;
	OID	desPriv = NULLOID;

	OID	currentTime = NULLOID;

static	OID	nullSpecific = NULLOID;

	OID	snmpUDPDomain = NULLOID;
	OID	rfc1157Domain = NULLOID;

	int	party_changed = 0;
	int	context_changed = 0;
	int	access_changed = 0;
	int	view_changed = 0;
	int	view_granular = 0;

struct snmp_party   *partyHead = NULL;
struct snmp_context *contextHead = NULL;
struct snmp_access  *accessHead = NULL;
struct snmp_view    *viewHead = NULL;

static struct row {
    char   *r_name;
    int	    r_value;
}	rows[] = {
    "active",		    ROW_ACTIVE,
    "notInService",	    ROW_NOTINSVC,
    "notReady",		    ROW_NOTREADY,
    "createdAndGo",	    ROW_CREATEGO,
    "createAndWait",	    ROW_CREATEWAIT,
    "destroy",		    ROW_DESTROY,

	NULL
};

/*  */

static struct pair {
    char   *p_name;
    OID	   *p_value;
}	pairs[] = {
    "noAuth",		    &noAuth,
    "v2md5AuthProtocol",    &md5Auth,
    "rfc1157noAuth",	    &rfc1157noAuth,
    "noPriv",		    &noPriv,
    "desPrivProtocol",	    &desPriv,
    "currentTime",	    &currentTime,
    "0.0",		    &nullSpecific,
    "snmpUDPDomain",	    &snmpUDPDomain,
    "rfc1157Domain",	    &rfc1157Domain,

    NULL
};


int	init_party (area, who, rflag, prefix)
char   *area;
int	who,
	rflag;
char  **prefix;
{
    register char *cp;
    char    file[BUFSIZ],
	    party[BUFSIZ];
    register struct pair *p;
    static int once_only = 1;

    PY_pepy[0] = NULL;

    if (once_only) {
	once_only = 0;

	for (p = pairs; p -> p_name; p++)
	    if ((*p -> p_value = text2oid (p -> p_name)) == NULLOID) {
		(void) sprintf (PY_pepy, "unknown OID \"%s\"", p -> p_name);
		return NOTOK;
	    }
    }

    if (who == MANAGER && !*area)
	return OK;

    (void) sprintf (file, "%s.%s", area, PUBLIC);
    (void) sprintf (party, "parties/%s.%s", area, PUBLIC);
    if (access (file, 0x00) == NOTOK
	    && access (isodefile (party, 0), 0x00) != NOTOK) {
	(void) sprintf (party, "parties/%s.", area);
	(void) sprintf (file, "%s", isodefile (party, 0));
    }
    else
	(void) sprintf (file, "%s.", area);
    cp = file + strlen (file);

    (void) strcpy (cp, PUBLIC);
    if (get_public_data (file, who, rflag) == NOTOK)
	return NOTOK;

    (void) strcpy (cp, PRIVATE);
    if (get_private_data (file, who, rflag) == NOTOK)
	return NOTOK;

    (void) strcpy (cp, CONTEXT);
    if (get_context_data (file, rflag) == NOTOK)
	return NOTOK;

    (void) strcpy (cp, ACCESS);
    if (get_access_data (file, who, rflag) == NOTOK)
	return NOTOK;

    if (who == AGENT) {
	(void) strcpy (cp, VIEW);
	if (get_view_data (file, rflag) == NOTOK)
	    return NOTOK;
    }

    if (prefix) {
	*cp = NULL;
	if ((*prefix = strdup (file)) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory");
	    return NOTOK;
	}
    }

    return check_party_configurations (who);
}


static  int check_file_perms (file, fp)
char   *file;
FILE   *fp;
{
    struct stat st;

    if (fstat (fileno (fp), &st) == NOTOK) {
	(void) sprintf (PY_pepy, "unable to fstat %s: %s", file,
			sys_errname (errno));
	return NOTOK;
    }
    if (st.st_uid != getuid () && st.st_uid != 0) {
	(void) sprintf (PY_pepy, "%s: not owned by you!", file);
	return NOTOK;
    }

    return OK;
}

/*  */

int	fin_party ()
{
    while (partyHead)
	free_party (partyHead);

    while (contextHead)
	free_context (contextHead);

    while (accessHead)
	free_access (accessHead);

    while (viewHead)
	free_view (viewHead);

    party_changed = context_changed = access_changed = view_changed = 0;

    return OK;
}

/*  */

void	free_party_aux (p, linked)
register struct snmp_party *p;
int	linked;
{
    if (linked) {
	register struct snmp_party  *pp,
				   **ppp;

	for (ppp = &partyHead; pp = *ppp; ppp = &pp -> party_next)
	    if (pp == p)
		break;
	*ppp = p -> party_next;
    }

    if (p -> party_Identity)
	oid_free (p -> party_Identity);
    if (p -> party_TDomain)
	oid_free (p -> party_TDomain);
    if (p -> party_TAddress)
	free ((char *) p -> party_TAddress);
    if (p -> party_AuthProtocol)
	oid_free (p -> party_AuthProtocol);
    if (p -> party_AuthPrivate)
	free ((char *) p -> party_AuthPrivate);
    if (p -> party_AuthPublic)
	free ((char *) p -> party_AuthPublic);
    if (p -> party_PrivProtocol)
	oid_free (p -> party_PrivProtocol);
    if (p -> party_PrivPrivate)
	free ((char *) p -> party_PrivPrivate);
    if (p -> party_PrivPublic)
	free ((char *) p -> party_PrivPublic);
    if (p -> party_new_AuthPrivate)
	free ((char *) p -> party_new_AuthPrivate);
    if (p -> party_new_AuthPublic)
	free ((char *) p -> party_new_AuthPublic);
    if (p -> party_new_PrivPrivate)
	free ((char *) p -> party_new_PrivPrivate);
    if (p -> party_new_PrivPublic)
	free ((char *) p -> party_new_PrivPublic);

    free ((char *) p);
}


void  free_context_aux (sc, linked)
register struct snmp_context *sc;
int	linked;
{
    if (linked) {
	register struct snmp_context  *pc,
				     **ppc;

	for (ppc = &contextHead; pc = *ppc; ppc = &pc -> sc_next)
	    if (pc == sc)
		break;
	*ppc = sc -> sc_next;
    }

    if (sc -> sc_name)
	free ((char *) sc -> sc_name);
    if (sc -> sc_identity)
	oid_free (sc -> sc_identity);
    if (sc -> sc_entity)
	free ((char *) sc -> sc_entity);
    if (sc -> sc_time)
	oid_free (sc -> sc_time);
    if (sc -> sc_dst_oid)
	oid_free (sc -> sc_dst_oid);
    if (sc -> sc_src_oid)
	oid_free (sc -> sc_src_oid);
    if (sc -> sc_ctx)
	oid_free (sc -> sc_ctx);
    if (sc -> sc_new_name)
	free ((char *) sc -> sc_new_name);
    if (sc -> sc_new_entity)
	free ((char *) sc -> sc_new_entity);
    if (sc -> sc_new_time)
	oid_free (sc -> sc_new_time);
    if (sc -> sc_new_dst_oid)
	oid_free (sc -> sc_new_dst_oid);
    if (sc -> sc_new_src_oid)
	oid_free (sc -> sc_new_src_oid);
    if (sc -> sc_new_ctx)
	oid_free (sc -> sc_new_ctx);

    free ((char *) sc);
}


void	free_access_aux (sa, linked)
register struct snmp_access *sa;
int	linked;
{
    register struct snmp_party *p;

    if (linked) {
	register struct snmp_access  *ap,
				    **app;

	for (app = &accessHead; ap = *app; app = &ap -> sa_next)
	    if (ap == sa)
		break;
	*app = sa -> sa_next;
    }

    if (sa -> sa_instance.oid_elements)
	free ((char *) sa -> sa_instance.oid_elements);

    free ((char *) sa);

    for (p = partyHead; p; p = p -> party_next)
	if (p -> party_clocksync == sa)
	    p -> party_clocksync = NULL;
}


void	free_view_aux (vu, linked)
register struct snmp_view *vu;
int	linked;
{
    if (linked) {
	register struct snmp_view  *vp,
				  **vpp;

	for (vpp = &viewHead; vp = *vpp; vpp = &vp -> vu_next)
	    if (vp == vu)
		break;
	*vpp = vu -> vu_next;
    }

    if (vu -> vu_instance.oid_elements)
	free ((char *) vu -> vu_instance.oid_elements);
    if (vu -> vu_subtree)
	oid_free (vu -> vu_subtree);
    if (vu -> vu_mask)
	qb_free (vu -> vu_mask);
    if (vu -> vu_new_mask)
	qb_free (vu -> vu_new_mask);

    free ((char *) vu);
}

/*  */

static int  get_public_data (file, who, rflag)
char   *file;
int	who,
	rflag;
{
    register char *cp;
    char    buffer[BUFSIZ],
	    line[BUFSIZ],
	   *vec[NVEC + NSLACK + 1];
    register struct snmp_party *p;
    struct snmp_party **pp;
    FILE   *fp;

    if ((fp = fopen (file, "r")) == NULL) {
	(void) sprintf (PY_pepy, "unable to read %s: %s", file,
			sys_errname (errno));
	return NOTOK;
    }
    if (!rflag && check_file_perms (file, fp) == NOTOK) {
	(void) fclose (fp);
	return NOTOK;
    }

    pp = &partyHead, p = NULL;
    while (fgets (buffer, sizeof buffer, fp)) {
again: ;
	if (cp = index (buffer, '\n'))
	    *cp = NULL;
	else {
line_too_long: ;
	    (void) sprintf (PY_pepy, "%s: line too long (%d character max)",
			    file, sizeof buffer - 1);
	    goto you_lose;
	}
	if (cp > buffer && *--cp == '\\') {
	    int	    i = sizeof buffer - (cp - buffer);

	    if (i <= 0)
		goto line_too_long;
	    if (fgets (cp, i, fp) == NULL) {
		(void) sprintf (PY_pepy, "%s: eof after continuation line",
				file);
		goto you_lose;
	    }
	    goto again;
	}
	if (*buffer == '#')
	    continue;
	(void) strcpy (line, buffer);

	if (p == NULL) {
	    static int	i = 0;

	    if ((p = (struct snmp_party *) calloc (1, sizeof *p)) == NULL) {
		(void) sprintf (PY_pepy, "out of memory");
you_lose: ;
		(void) fclose (fp);
		return NOTOK;
	    }

	    p -> party_index = ++i;
	    p -> party_status = ROW_NOTREADY;
	    p -> party_storage = STORAGE_NONVOLATILE;
	}

	bzero ((char *) vec, sizeof vec);
	switch (str2vec (buffer, vec)) {
	    case 0:
	        continue;

	    case 11:
		if (get_string (vec[10], &p -> party_PrivPublic,
				&p -> party_PrivPublic_len) == NOTOK)
		    goto you_lose;
		/* and fall... */
	    case 10:
		if (get_string (vec[9], &p -> party_AuthPublic,
				&p -> party_AuthPublic_len) == NOTOK)
		    goto you_lose;
		/* and fall... */
	    case 9:
		switch (*vec[0]) {
		    case 'A':
		    case 'a':
			if (who == AGENT)
			    continue;
		        /* else fall... */
		    case 'L':
		    case 'l':
		        p -> party_location = AGENT;
			break;

		    case 'P':
		    case 'p':
		    case 'M':
		    case 'm':
		        p -> party_location = MANAGER;
			break;

		    default:
			(void) sprintf (PY_pepy,
					"%s: malformed location \"%s\"",
					file, vec[0]);
			goto you_lose;
		}

		if ((p -> party_Identity = text2oid (cp = vec[1]))
		        == NULLOID) {
unknown_oid: ;
		    (void) sprintf (PY_pepy, "%s: unknown OID \"%s\"",
				    file, cp);
		    goto you_lose;
		}

		if ((p -> party_TDomain = text2oid (cp = vec[2])) == NULLOID)
		    goto unknown_oid;

		if (get_string (vec[3], &p -> party_TAddress,
				&p -> party_TAddress_len) == NOTOK)
		    goto you_lose;

		if (sscanf (cp = vec[4], "%d", &p -> party_MaxMessageSize) != 1
		        || p -> party_MaxMessageSize < MAXSNMP
		        || p -> party_MaxMessageSize > BIGSNMP) {
		    (void) sprintf (PY_pepy,
				    "%s: malformed MaxMessageSize \"%s\"",
				    file, cp);
		    goto you_lose;
		}

		if ((p -> party_AuthProtocol = text2oid (cp = vec[5]))
		        == NULLOID)
		    goto unknown_oid;

		if (sscanf (cp = vec[6], "%lu", &p -> party_AuthStart) != 1) {
		    (void) sprintf (PY_pepy, "%s: malformed AuthClock \"%s\"",
				    file, cp);
		    goto you_lose;
		}

		if (sscanf (cp = vec[7], "%ld", &p -> party_AuthLifetime) != 1
		        || p -> party_AuthLifetime < 0) {
		    (void) sprintf (PY_pepy,
				    "%s: malformed AuthLifetime \"%s\"",
				    file, cp);
		    goto you_lose;
		}

		if ((p -> party_PrivProtocol = text2oid (cp = vec[8]))
		        == NULLOID)
		    goto unknown_oid;
		break;

	    default:
		(void) sprintf (PY_pepy, "%s: malformed line: \"%s\"", file,
				line);
		goto you_lose;
	}

	*pp = p, pp = &p -> party_next;
	p = NULL;
    }
    if (p)
	free ((char *) p);

    (void) fclose (fp);

    return OK;
}

/*  */

static int  get_private_data (file, who, rflag)
char   *file;
int	who,
	rflag;
{
    register char *cp;
    char    buffer[BUFSIZ],
	    line[BUFSIZ],
	   *vec[NVEC + NSLACK + 1];
    FILE   *fp;

    if ((fp = fopen (file, "r")) == NULL) {
	(void) sprintf (PY_pepy, "unable to read %s: %s", file,
			sys_errname (errno));
	return NOTOK;
    }
    if (!rflag && check_file_perms (file, fp) == NOTOK) {
	(void) fclose (fp);
	return NOTOK;
    }

    while (fgets (buffer, sizeof buffer, fp)) {
	int	nvec;
	register struct snmp_party *p;
	register struct row *r;
	OID	oid;

again: ;
	if (cp = index (buffer, '\n'))
	    *cp = NULL;
	else {
line_too_long: ;
	    (void) sprintf (PY_pepy, "%s: line too long (%d character max)",
			    file, sizeof buffer - 1);
	    goto you_lose;
	}
	if (cp > buffer && *--cp == '\\') {
	    int	    i = sizeof buffer - (cp - buffer);

	    if (i <= 0)
		goto line_too_long;
	    if (fgets (cp, i, fp) == NULL) {
		(void) sprintf (PY_pepy, "%s: eof after continuation line",
				file);
		goto you_lose;
	    }
	    goto again;
	}
	if (*buffer == '#')
	    continue;
	(void) strcpy (line, buffer);

	bzero ((char *) vec, sizeof vec);
	switch (nvec = str2vec (buffer, vec)) {
	    case 0:
	        break;

	    case 4:
	    case 3:
	    case 2:
		if ((oid = text2oid (cp = vec[0])) == NULL) {
		    (void) sprintf (PY_pepy, "%s: unknown OID \"%s\"", file,
				    cp);
you_lose: ;
		    (void) fclose (fp);
		    return NOTOK;
		}
		p = lookup_party_aux (oid, EITHER, 0);
		oid_free (oid);
		if (p == NULL) {
		    if (who == AGENT)
			break;

		    (void) sprintf (PY_pepy, "%s: unknown party \"%s\"", file,
				    cp);
		    goto you_lose;
		}

		cp = vec[1];
		for (r = rows; r -> r_name; r++)
		    if (lexequ (r -> r_name, cp) == 0) {
			p -> party_status = r -> r_value;
			break;
		    }
		if (!r -> r_name || r -> r_value >= ROW_CREATEGO) {
		    (void) sprintf (PY_pepy,
				    "%s: malformed party status \"%s\"",
				    file, cp);
		    goto you_lose;
		}

		if (nvec > 2) {
		    if (get_string (vec[2], &p -> party_AuthPrivate,
				    &p -> party_AuthPrivate_len)
			    == NOTOK)
			goto you_lose;

		    if (nvec > 3
			    && get_string (vec[3],
					   &p -> party_PrivPrivate,
					   &p -> party_PrivPrivate_len)
				    == NOTOK)
			goto you_lose;
		}
		break;

	    default:
		(void) sprintf (PY_pepy, "%s: malformed line: \"%s\"", file,
				line);
		goto you_lose;
	}
    }

    (void) fclose (fp);

    return OK;
}

/*  */

static int  get_context_data (file, rflag)
char   *file;
int	rflag;
{
    register char *cp;
    char    buffer[BUFSIZ],
	    line[BUFSIZ],
	   *vec[NVEC + NSLACK + 1];
    register struct snmp_context *sc;
    struct snmp_context **pc;
    FILE   *fp;

    if ((fp = fopen (file, "r")) == NULL) {
	(void) sprintf (PY_pepy, "unable to read %s: %s", file,
			sys_errname (errno));
	return NOTOK;
    }
    if (!rflag && check_file_perms (file, fp) == NOTOK) {
	(void) fclose (fp);
	return NOTOK;
    }

    pc = &contextHead, sc = NULL;
    while (fgets (buffer, sizeof buffer, fp)) {
	register struct row *r;

again: ;
	if (cp = index (buffer, '\n'))
	    *cp = NULL;
	else {
line_too_long: ;
	    (void) sprintf (PY_pepy, "%s: line too long (%d character max)",
			    file, sizeof buffer - 1);
	    goto you_lose;
	}
	if (cp > buffer && *--cp == '\\') {
	    int	    i = sizeof buffer - (cp - buffer);

	    if (i <= 0)
		goto line_too_long;
	    if (fgets (cp, i, fp) == NULL) {
		(void) sprintf (PY_pepy, "%s: eof after continuation line",
				file);
		goto you_lose;
	    }
	    goto again;
	}
	if (*buffer == '#')
	    continue;
	(void) strcpy (line, buffer);

	if (sc == NULL) {
	    static int	i = 0;

	    if ((sc = (struct snmp_context *) calloc (1, sizeof *sc))
		    == NULL) {
no_mem: ;
		(void) sprintf (PY_pepy, "out of memory");
you_lose: ;
		(void) fclose (fp);
		return NOTOK;
	    }

	    sc -> sc_index = ++i;
	    sc -> sc_status = ROW_NOTREADY;
	    sc -> sc_storage = STORAGE_NONVOLATILE;
	}

	bzero ((char *) vec, sizeof vec);
	switch (str2vec (buffer, vec)) {
	    case 0:
	        continue;

	    case 8:
		switch (*vec[0]) {
		    case 'A':
		    case 'a':
		        sc -> sc_location = AGENT;
			break;

		    case 'M':
		    case 'm':
		        sc -> sc_location = MANAGER;
			break;

		    default:
			(void) sprintf (PY_pepy,
					"%s: malformed location \"%s\"",
					file, vec[0]);
			goto you_lose;
		}

		if ((sc -> sc_identity = text2oid (cp = vec[1])) == NULLOID) {
unknown_oid: ;
		    (void) sprintf (PY_pepy, "%s: unknown OID \"%s\"",
				    file, cp);
		    goto you_lose;
		}

		if (get_string (vec[2], &sc -> sc_name,
				&sc -> sc_name_len) == NOTOK)
		    goto you_lose;

		cp = vec[3];
		for (r = rows; r -> r_name; r++)
		    if (lexequ (r -> r_name, cp) == 0) {
			sc -> sc_status = r -> r_value;
			break;
		    }
		if (!r -> r_name || r -> r_value >= ROW_CREATEGO) {
		    (void) sprintf (PY_pepy,
				    "%s: malformed context status \"%s\"",
				    file, cp);
		    goto you_lose;
		}

		if (lexequ (cp = vec[4], "view") == 0) {
		    if (sscanf (cp = vec[5], "%d", &sc -> sc_view) != 1
		            || sc -> sc_view <= 0) {
			(void) sprintf (PY_pepy,
					"%s: malformed view index \"%s\"",
					file, cp);
			goto you_lose;
		    }

		    if (get_string (vec[6], &sc -> sc_entity,
				    &sc -> sc_entity_len) == NOTOK)
			goto you_lose;

		    if ((sc -> sc_time = text2oid (cp = vec[7])) == NULLOID)
			goto unknown_oid;

		    if ((sc -> sc_dst_oid = oid_cpy (nullSpecific)) == NULLOID
			    || (sc -> sc_src_oid = oid_cpy (nullSpecific))
				    == NULLOID
			    || (sc -> sc_ctx = oid_cpy (nullSpecific))
				    == NULLOID)
			goto no_mem;
		}
		else
		    if (lexequ (cp, "proxy") == 0) {
			OID	oid;

			if ((sc -> sc_dst_oid = oid = text2oid (cp = vec[5]))
			        == NULLOID)
			    goto unknown_oid;
			if ((sc -> sc_dst = lookup_party_aux (oid, MANAGER, 0))
				    == NULL) {
			    if (strcmp (cp, "0.0"))
				ADVISE (LLOG_EXCEPTIONS, NULLCP,
					"%s: unknown party \"%s\"\n",
					file, cp);
			}

			if ((sc -> sc_src_oid = oid = text2oid (cp = vec[6]))
			        == NULLOID)
			    goto unknown_oid;
			if ((sc -> sc_src = lookup_party_aux (oid, AGENT, 0))
				    == NULL) {
			    if (strcmp (cp, "0.0"))
				ADVISE (LLOG_EXCEPTIONS, NULLCP,
					"%s: unknown party \"%s\"\n",
					file, cp);
			}

			if ((sc -> sc_ctx = text2oid (cp = vec[7])) == NULLOID)
			    goto unknown_oid;

			if ((sc -> sc_time = oid_cpy (nullSpecific))
			        == NULLOID)
			    goto no_mem;
		    }
		    else {
			(void) sprintf (PY_pepy,
					"%s: malformed context type \"%s\"",
					file, cp);
			goto you_lose;
		    }
		break;

	    default:
		(void) sprintf (PY_pepy, "%s: malformed line: \"%s\"", file,
				line);
		goto you_lose;
	}

	*pc = sc, pc = &sc -> sc_next;
	sc = NULL;
    }
    if (sc)
	free ((char *) sc);

    (void) fclose (fp);

    return OK;
}

/*  */

static struct perm {
    char   *p_name;
    int	    p_value;
}	perms[] = {
    "get",	    PARTY_PRIV_GET,
    "get-next",     PARTY_PRIV_NEXT,
    "response",	    PARTY_PRIV_RESP,
    "set",	    PARTY_PRIV_SET,
    "trap",	    PARTY_PRIV_TRAP,
    "get-bulk",	    PARTY_PRIV_BULK,
    "inform",	    PARTY_PRIV_INFO,
    "notify",	    PARTY_PRIV_NTFY,

    NULL
};


/* ARGSUSED */

static int  get_access_data (file, who, rflag)
char   *file;
int	who,
	rflag;
{
    register char *cp;
    char    buffer[BUFSIZ],
	    line[BUFSIZ],
	   *vec[NVEC + NSLACK + 1];
    struct snmp_access **ap;
    FILE   *fp;

    if ((fp = fopen (file, "r")) == NULL) {
	(void) sprintf (PY_pepy, "unable to read %s: %s", file,
			sys_errname (errno));
	return NOTOK;
    }
    if (!rflag && check_file_perms (file, fp) == NOTOK) {
	(void) fclose (fp);
	return NOTOK;
    }

    ap = &accessHead;
    while (fgets (buffer, sizeof buffer, fp)) {
	int	nvec;
	char  **vp;
	register struct snmp_access *sa;
	register struct row *r;
	OID	oid;

again: ;
	if (cp = index (buffer, '\n'))
	    *cp = NULL;
	else {
line_too_long: ;
	    (void) sprintf (PY_pepy, "%s: line too long (%d character max)",
			    file, sizeof buffer - 1);
	    goto you_lose;
	}
	if (cp > buffer && *--cp == '\\') {
	    int	    i = sizeof buffer - (cp - buffer);

	    if (i <= 0)
		goto line_too_long;
	    if (fgets (cp, i, fp) == NULL) {
		(void) sprintf (PY_pepy, "%s: eof after continuation line",
				file);
		goto you_lose;
	    }
	    goto again;
	}
	if (*buffer == '#')
	    continue;
	(void) strcpy (line, buffer);

	bzero ((char *) vec, sizeof vec);
	if ((nvec = str2vec (buffer, vec)) < 4) {
	    if (nvec == 0)
		continue;

	    (void) sprintf (PY_pepy, "%s: malformed line: \"%s\"", file,
			    line);
	    goto you_lose;
	}

	if ((sa = (struct snmp_access *) calloc (1, sizeof *sa)) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory");
	    goto you_lose;
	}
	*ap = sa, ap = &sa -> sa_next;

	if ((oid = text2oid (cp = vec[0])) == NULL) {
unknown_oid: ;
	    (void) sprintf (PY_pepy, "%s: unknown OID \"%s\"", file, cp);

you_lose: ;
	    (void) fclose (fp);
	    return NOTOK;
	}
	if ((sa -> sa_target = lookup_party_aux (oid, EITHER, 0)) == NULL) {
	    ADVISE (LLOG_EXCEPTIONS, NULLCP, "%s: unknown party \"%s\"\n",
		    file, cp);
	}
	else
	    sa -> sa_target_index = sa -> sa_target -> party_index;
	oid_free (oid);

	if ((oid = text2oid (cp = vec[1])) == NULL)
	    goto unknown_oid;
	if ((sa -> sa_subject = lookup_party_aux (oid, EITHER, 0)) == NULL) {
	    ADVISE (LLOG_EXCEPTIONS, NULLCP, "%s: unknown party \"%s\"\n",
		    file, cp);
	}
	else
	    sa -> sa_subject_index = sa -> sa_subject -> party_index;
	oid_free (oid);

	if ((oid = text2oid (cp = vec[2])) == NULL)
	    goto unknown_oid;
	if ((sa -> sa_resources = lookup_context_aux (oid, 0)) == NULL) {
	    ADVISE (LLOG_EXCEPTIONS, NULLCP, "%s: unknown context \"%s\"\n",
		    file, cp);
	}
	else
	    sa -> sa_resources_index = sa -> sa_resources -> sc_index;

	cp = vec[3];
	for (r = rows; r -> r_name; r++)
	    if (lexequ (r -> r_name, cp) == 0) {
		sa -> sa_status = r -> r_value;
		break;
	    }
	if (!r -> r_name || r -> r_value >= ROW_CREATEGO) {
	    (void) sprintf (PY_pepy, "%s: malformed access status \"%s\"",
			    file, cp);
	    goto you_lose;
	}
	sa -> sa_storage = STORAGE_NONVOLATILE;

	for (vp = vec + 4; cp = *vp; vp++) {
	    register struct perm *p;

	    for (p = perms; p -> p_name; p++)
		if (lexequ (cp, p -> p_name) == 0) {
		    sa -> sa_privileges |= p -> p_value;
		    break;
		}
	    if (!p -> p_name) {
		(void) sprintf (PY_pepy,
				"%s: malformed access permission \"%s\"",
				file, cp);
		goto you_lose;
	    }
	}
	sa -> sa_new_privileges = sa -> sa_privileges;
    }

    (void) fclose (fp);

    return OK;
}

/*  */

static int  get_view_data (file, rflag)
char   *file;
int	rflag;
{
    register char *cp;
    char    buffer[BUFSIZ],
	    line[BUFSIZ],
	   *vec[NVEC + NSLACK + 1];
    struct snmp_view **vp;
    FILE   *fp;

    if ((fp = fopen (file, "r")) == NULL) {
	(void) sprintf (PY_pepy, "unable to read %s: %s", file,
			sys_errname (errno));
	return NOTOK;
    }
    if (!rflag && check_file_perms (file, fp) == NOTOK) {
	(void) fclose (fp);
	return NOTOK;
    }

    vp = &viewHead;
    while (fgets (buffer, sizeof buffer, fp)) {
	int	hasmask = 0;
	register struct snmp_view *vu;
	register struct row *r;
	static int family = 0;

again: ;
	if (cp = index (buffer, '\n'))
	    *cp = NULL;
	else {
line_too_long: ;
	    (void) sprintf (PY_pepy, "%s: line too long (%d character max)",
			    file, sizeof buffer - 1);
	    goto you_lose;
	}
	if (cp > buffer && *--cp == '\\') {
	    int	    i = sizeof buffer - (cp - buffer);

	    if (i <= 0)
		goto line_too_long;
	    if (fgets (cp, i, fp) == NULL) {
		(void) sprintf (PY_pepy, "%s: eof after continuation line",
				file);
		goto you_lose;
	    }
	    goto again;
	}
	if (*buffer == '#')
	    continue;
	(void) strcpy (line, buffer);

	bzero ((char *) vec, sizeof vec);
	switch (str2vec (buffer, vec)) {
	    case 0:
	        continue;

	    case 5:
		hasmask = 1;
		/* and fall... */
	    case 4:
		if ((vu = (struct snmp_view *) calloc (1, sizeof *vu))
		        == NULL) {
out_of_memory: ;
		    (void) sprintf (PY_pepy, "out of memory");
		    goto you_lose;
		}
		*vp = vu, vp = &vu -> vu_next;

		if (sscanf (cp = vec[0], "%d", &vu -> vu_index) != 1
		            || vu -> vu_index <= 0) {
			(void) sprintf (PY_pepy,
					"%s: malformed view index \"%s\"",
					file, cp);
you_lose: ;
			(void) fclose (fp);
			return NOTOK;
		    }

		cp = vec[1];
		for (r = rows; r -> r_name; r++)
		    if (lexequ (r -> r_name, cp) == 0) {
			vu -> vu_status = r -> r_value;
			break;
		    }
		if (!r -> r_name || r -> r_value >= ROW_CREATEGO) {
		    (void) sprintf (PY_pepy,
				    "%s: malformed view status \"%s\"",
				    file, cp);
		    goto you_lose;
		}
		vu -> vu_storage = STORAGE_NONVOLATILE;

		if (lexequ (cp = vec[2], "excluded") == 0)
		    vu -> vu_type = VIEW_EXCLUDED;
		else
		    if (lexequ (cp, "included") == 0)
			vu -> vu_type = VIEW_INCLUDED;
		    else {
			(void) sprintf (PY_pepy,
					"%s: malformed view type \"%s\"",
					file, cp);
			goto you_lose;
		    }

		if ((vu -> vu_subtree = text2oid (cp = vec[3])) == NULL) {
		    (void) sprintf (PY_pepy, "%s: unknown OID \"%s\"", file,
				    cp);
		    goto you_lose;
		}

		if (hasmask) {
		    int	    len;
		    u_char *m;

		    if (get_string (vec[4], &m, &len) == NOTOK)
			goto you_lose;
		    vu -> vu_mask = str2qb ((char *) m, len, 1);
		    free ((char *) m);
		    if (vu -> vu_mask == NULL)
			goto out_of_memory;
		}
		break;

	    default:
		(void) sprintf (PY_pepy, "%s: malformed line: \"%s\"", file,
				line);
		goto you_lose;
	}
    }

    (void) fclose (fp);

    return OK;
}

/*  */

static int  get_string (s, base, cc)
char   *s;
u_char **base;
int    *cc;
{
    char   *ss = s;

    if (lexnequ (s, "0x", sizeof "0x" - 1) == 0) {
	int	hex,
		len;
	char   *fmt;
	register u_char *p;

	s += 2;
	len = (strlen (s) >> 1) + 2;	/* for safety... */
	if ((*base = (u_char *) malloc ((unsigned) len)) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory");
	    return NOTOK;
	}
	p = *base;

	if (index (s, '.'))
	    hex = 0, fmt = "%d";
	else
	    hex = 1, fmt = "%x";
	while (*s) {
	    int	    i;

	    if (sscanf (s, fmt, &i) != 1) {
oops: ;
		free ((char *) *base);
		*base = NULL;
		(void) sprintf (PY_pepy, "malformed string \"%s\"", ss);
		return NOTOK;
	    }
	    *p++ = i & 0xff;

	    if (hex)
		while (isxdigit ((u_char) *s))
		    s++;
	    else
		while (isdigit ((u_char) *s))
		    s++;
	    switch (*s) {
		case '.':
		    if (hex)
			goto oops;
		    s++;
		    break;

		case ':':
		    if (!hex)
			goto oops;
		    s++;
		    break;

		case NULL:
		    break;

		default:
		    goto oops;
	    }
	}

	*cc = p - *base;
    }
    else {
	if ((*base = (u_char *) strdup (s)) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory");
	    return NOTOK;
	}
	*cc = strlen (s);
    }

    return OK;
}

/*  */

static int  sp_compar (a, b)
struct snmp_party **a, **b;
{
    int	    i = oid_cmp ((*a) -> party_Identity, (*b) -> party_Identity);

    if (i == 0) {
	if (PY_pepy[0])
	    (void) sprintf (PY_pepy + strlen (PY_pepy), ", ");
	else
	    (void) sprintf (PY_pepy, "duplicate party entries: ");
	(void) sprintf (PY_pepy + strlen (PY_pepy), "\"%s\"",
			oid2ode ((*a) -> party_Identity));
    }    

    return i;
}


static int  sc_compar (a, b)
struct snmp_context **a, **b;
{
    int	    i = oid_cmp ((*a) -> sc_identity, (*b) -> sc_identity);

    if (i == 0) {
	if (PY_pepy[0])
	    (void) sprintf (PY_pepy + strlen (PY_pepy), ", ");
	else
	    (void) sprintf (PY_pepy, "duplicate context entries: ");
	(void) sprintf (PY_pepy + strlen (PY_pepy), "\"%s\"",
			oid2ode ((*a) -> sc_identity));
    }    

    return i;
}


static int  sa_compar (a, b)
struct snmp_access **a, **b;
{
    int	    i = oid_cmp (&((*a) -> sa_instance), &((*b) -> sa_instance));

    if (i == 0) {
	if (PY_pepy[0])
	    (void) sprintf (PY_pepy + strlen (PY_pepy), ", ");
	else
	    (void) sprintf (PY_pepy, "duplicate access entries: ");
	(void) sprintf (PY_pepy + strlen (PY_pepy),
			" target \"%d\"/subject \"%d\"/resources \"%s\"",
			(*a) -> sa_target_index, (*a) -> sa_subject_index,
			(*a) -> sa_resources_index);
    }    

    return i;
}


static int  vu_compar (a, b)
struct snmp_view **a, **b;
{
    int	    i = oid_cmp (&((*a) -> vu_instance), &((*b) -> vu_instance));

    if (i == 0) {
	if (PY_pepy[0])
	    (void) sprintf (PY_pepy + strlen (PY_pepy), ", ");
	else
	    (void) sprintf (PY_pepy, "duplicate view entries: ");
	(void) sprintf (PY_pepy + strlen (PY_pepy),
			" view \"%d\"/subtree \"%s\"",
			(*a) -> vu_index, oid2ode ((*a) -> vu_subtree));
    }

    return i;
}

/*  */

static int  check_party_configurations (who)
int	who;
{
    int	    i;
    u_long  now = party_time_now ();
    register struct snmp_party *p;
    register struct snmp_context *sc;
    register struct snmp_access *sa;
    register struct snmp_view *vu;

    i = 0;
    for (p = partyHead; p; p = p -> party_next) {
	if (debug) {
	    fprintf (stderr, "party %s:\n", oid2ode (p -> party_Identity));
	    fprintf (stderr, "\tindex=%d status=%d(%s)\n", p -> party_index,
		     p -> party_status, rows[p -> party_status - 1].r_name);
	    fprintf (stderr, "\ttransport address=%s\n", party_address (p));
	    fprintf (stderr, "\tmaxsize=%d location=%d\n",
		     p -> party_MaxMessageSize, p -> party_location);
	    fprintf (stderr, "\tauth protocol=%s start=%lu lifetime=%ld\n",
		     oid2ode (p -> party_AuthProtocol),
		     p -> party_AuthStart ? now - p -> party_AuthStart : 0,
		     p -> party_AuthLifetime);
	    fprintf (stderr, "\t    public=%s\n",
		     u_char2str (p -> party_AuthPublic,
				 p -> party_AuthPublic_len));
	    fprintf (stderr, "\t    private=%s\n",
		     u_char2str (p -> party_AuthPrivate,
				 p -> party_AuthPrivate_len));
	    fprintf (stderr, "\tpriv protocol=%s\n",
		     oid2ode (p -> party_PrivProtocol));
	    fprintf (stderr, "\t    public=%s\n",
		     u_char2str (p -> party_PrivPublic,
				 p -> party_PrivPublic_len));
	    fprintf (stderr, "\t    private=%s\n",
		     u_char2str (p -> party_PrivPrivate,
				 p -> party_PrivPrivate_len));
	}

	i++;

	p -> party_tDomain =
	        oid_cmp (p -> party_TDomain, snmpUDPDomain) == 0
		    ? SNMPUDPDOMAIN
		    : oid_cmp (p -> party_TDomain, rfc1157Domain) == 0
			? RFC1157DOMAIN : BADOMAIN;
	if (p -> party_tDomain == SNMPUDPDOMAIN
	        || p -> party_tDomain == RFC1157DOMAIN) {
	    if (p -> party_TAddress_len != 6) {
invalid_addr: ;
		(void) sprintf (PY_pepy, 
				"party %s has invalid address (%s)",
				oid2ode (p -> party_Identity),
				u_char2str (p -> party_TAddress,
					    p -> party_TAddress_len));
		return NOTOK;
	    }
	    if ((p -> party_TAddress[0] != 0
			|| p -> party_TAddress[1] != 0
			|| p -> party_TAddress[2] != 0
			|| p -> party_TAddress[3] != 0)
		    && (p -> party_TAddress[4] == 0
			    && p -> party_TAddress[5] == 0))
		goto invalid_addr;
	}

	p -> party_authProt =
		oid_cmp (p -> party_AuthProtocol, noAuth) == 0
		    ? NOAUTH
		    : oid_cmp (p -> party_AuthProtocol, md5Auth) == 0
			  ? MD5AUTHPROT
		          : oid_cmp (p -> party_AuthProtocol,
				     rfc1157noAuth) == 0
					 ? RFC1157NOAUTH : BADAUTH;
	switch (p -> party_authProt) {
	    case NOAUTH:
	        if (p -> party_AuthPrivate) {
		    if (debug)
			fprintf (stderr,
				 "party %s doesn't need an authentication key",
				 oid2ode (p -> party_Identity));
		}
		break;
	    
	    case MD5AUTHPROT:
		if (p -> party_AuthStart == 0) {
		    p -> party_AuthStart = now;
		    party_changed = 1;
		}
	        if (p -> party_AuthPrivate == NULL) {
		    if (debug)
			fprintf (stderr,
				 "party %s requires an authentication key for MD5\n",
				 oid2ode (p -> party_Identity));
		    p -> party_status = ROW_NOTREADY;
		    break;
		}
	        if (p -> party_AuthPrivate_len
		        != AUTH_KEY_LENGTH) {
		    (void) sprintf (PY_pepy,
				    "authentication key for MD5 party %s is wrong length %d (expecting %d)",
				    oid2ode (p -> party_Identity),
				    p -> party_AuthPrivate_len,
				    AUTH_KEY_LENGTH);
		    return NOTOK;
		}
		break;

	    case RFC1157NOAUTH:
	        if (p -> party_AuthPrivate == NULL) {
		    if (debug)
			fprintf (stderr,
				 "party %s requires an authentication key as community name\n",
				 oid2ode (p -> party_Identity));
		    p -> party_status = ROW_NOTREADY;
		    break;
		}
		break;

	    case BADAUTH:
		(void) sprintf (PY_pepy,
				"party %s uses an unsupported authentication algorithm",
				oid2ode (p -> party_Identity));
		return NOTOK;

	    default:
		break;
	}

	p -> party_privProt =
		oid_cmp (p -> party_PrivProtocol, noPriv) == 0
		    ? NOPRIV
		    : oid_cmp (p -> party_PrivProtocol, desPriv) == 0
			  ? DESPRIVPROT : BADPRIV;
	switch (p -> party_privProt) {
	    case NOPRIV:
	        if (p -> party_PrivPrivate) {
		    if (debug)
			fprintf (stderr,
				 "party %s doesn't need a privacy key\n",
				 oid2ode (p -> party_Identity));
		}
		break;
	    
	    case DESPRIVPROT:
	        if (p -> party_PrivPrivate == NULL) {
		    fprintf (stderr,
			     "party %s requires a privacy key for DES\n",
			     oid2ode (p -> party_Identity));
		    p -> party_status = ROW_NOTREADY;
		    break;
		}
	        if (p -> party_PrivPrivate_len
		        != PRIV_KEY_LENGTH * 2) {
		    (void) sprintf (PY_pepy,
				    "privacy key for DES is wrong length %d (expecting %d)",
				    oid2ode (p -> party_Identity),
				    p -> party_PrivPrivate_len,
				    PRIV_KEY_LENGTH * 2);
		    return NOTOK;
		}
		break;

	    case BADPRIV:
		(void) sprintf (PY_pepy,
				"party %s uses an unsupported privacy algorithm",
				oid2ode (p -> party_Identity));
		return NOTOK;

	    default:
		break;
	}

	if (p -> party_tDomain == RFC1157DOMAIN) {
	    if (p -> party_authProt != RFC1157NOAUTH) {
		(void) sprintf (PY_pepy,
				"party %s must use rfc1157noAuth for rfc1157Domain",
				oid2ode (p -> party_Identity));
		return NOTOK;
	    }

	    if (p -> party_privProt != NOPRIV) {
		(void) sprintf (PY_pepy,
				"party %s must use noPriv for rfc1157Domain",
				oid2ode (p -> party_Identity));
		return NOTOK;
	    }
	}

	/* so party_ready() doesn't surprise us... */
	p -> party_clone |= PARTY_CLONED | PARTY_AUTHPRIV | PARTY_PRIVPRIV;
	p -> party_old_clone = p -> party_clone;
    }

    if (i <= 0) {
	(void) sprintf (PY_pepy, "no parties defined");
	return NOTOK;
    }
    if (who == AGENT && i > 1) {
	register struct snmp_party **base,
				   **spe,
				   **spp;

	if ((base = (struct snmp_party **)
			    malloc ((unsigned) (i * sizeof *base))) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory trying to sort %d parties",
			    i);
	    return NOTOK;
	}

	spe = base;
	for (p = partyHead; p; p = p -> party_next)
	    *spe++ = p;

	PY_pepy[0] = NULL;
	qsort ((char *) base, i, sizeof *base, (IFP) sp_compar);

	spp = base;
	p = partyHead = *spp++;

	while (spp < spe) {
	    p -> party_next = *spp;
	    p = *spp++;
	}
	p -> party_next = NULL;

	free ((char *) base);

	if (PY_pepy[0])
	    return NOTOK;
    }

    i = 0;
    for (sc = contextHead; sc; sc = sc -> sc_next) {
	if (debug) {
	    fprintf (stderr, "context %s (%s):\n",
		     oid2ode (sc -> sc_identity),
		     u_char2str (sc -> sc_name, sc -> sc_name_len));
	    fprintf (stderr, "\tindex=%d status=%d(%s)\n", sc -> sc_index,
		     sc -> sc_status, rows[sc -> sc_status - 1].r_name);
	    fprintf (stderr, "\tlocation=%d\n", sc -> sc_location);
	    if (sc -> sc_view > 0)
		fprintf (stderr, "\tview: %d entity=%s time=%s\n",
			 sc -> sc_view,
			 u_char2str (sc -> sc_entity, sc -> sc_entity_len),
			 oid2ode (sc -> sc_time));
	    else {
		fprintf (stderr, "\tproxy: dst=%s\n",
			 sc -> sc_dst
			    ? oid2ode (sc -> sc_dst -> party_Identity)
			    : "none");
		fprintf (stderr, "\t       src=%s\n",
			 sc -> sc_src
			    ? oid2ode (sc -> sc_src -> party_Identity)
			    : "none");
		fprintf (stderr, "\t       ctx=%s\n",
			 sc -> sc_ctx ? oid2ode (sc -> sc_ctx) : "none");
	    }
	}

	i++;

	if (sc -> sc_view == 0)
	    if (p = sc -> sc_dst) {
		p -> party_flags |= PARTY_PROXYD;
		if (who == MANAGER)
		    continue;

		if (p -> party_status != ROW_ACTIVE)
		    sc -> sc_status = ROW_NOTREADY;
		else
		    if (p -> party_tDomain == SNMPUDPDOMAIN) {
			if (p = sc -> sc_src) {
			    p -> party_flags |= PARTY_PROXYS;
			    if (p -> party_status != ROW_ACTIVE
				    || !sc -> sc_ctx)
				sc -> sc_status = ROW_NOTREADY;
			}
			else
			    sc -> sc_status = ROW_NOTREADY;
		    }
	    }
	    else
		if (who != MANAGER)
		    sc -> sc_status = ROW_NOTREADY;
    }

    if (i <= 0) {
	(void) sprintf (PY_pepy, "no contexts defined");
	return NOTOK;
    }
    if (who == AGENT && i > 1) {
	register struct snmp_context **base,
                                     **sce,
                                     **scp;

	if ((base = (struct snmp_context **)
			    malloc ((unsigned) (i * sizeof *base))) == NULL) {
	    (void) sprintf (PY_pepy,
			    "out of memory trying to sort %d contexts", i);
	    return NOTOK;
	}

	sce = base;
	for (sc = contextHead; sc; sc = sc -> sc_next)
	    *sce++ = sc;

        PY_pepy[0] = NULL;
        qsort ((char *) base, i, sizeof *base, (IFP) sc_compar);

        scp = base;
        sc = contextHead = *scp++;

        while (scp < sce) {
            sc -> sc_next = *scp;
            sc = *scp++;
        }
        sc -> sc_next = NULL;

        free ((char *) base);

        if (PY_pepy[0])
            return NOTOK;
    }

    i = 0;
    for (sa = accessHead; sa; sa = sa -> sa_next) {
	register OID oid = &sa -> sa_instance;

	if (!sa -> sa_target || !sa -> sa_subject || !sa -> sa_resources) {
	    (void) sprintf (PY_pepy, "incomplete access control entry...");
	    return NOTOK;
	}

	i++;
	oid -> oid_nelem = 3;
	if ((oid -> oid_elements = (unsigned int *)
			calloc ((unsigned int) oid -> oid_nelem,
				sizeof *oid -> oid_elements)) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory");
	    return NOTOK;
	}
	oid -> oid_elements[0] = sa -> sa_target_index;
	oid -> oid_elements[1] = sa -> sa_subject_index;
	oid -> oid_elements[2] = sa -> sa_resources_index;

	if (sa -> sa_target -> party_tDomain
	        != sa -> sa_subject -> party_tDomain) {
	    (void) sprintf (PY_pepy,
			    "transport domain mismatch for access control entry %s",
			    sprintoid (oid));
	    return NOTOK;
	}
	if (sa -> sa_target -> party_authProt
	        != sa -> sa_subject -> party_authProt) {
	    (void) sprintf (PY_pepy,
			    "authentication protocol mismatch for access control entry %s",
			    sprintoid (oid));
	    return NOTOK;
	}
	if (sa -> sa_target -> party_privProt
	        != sa -> sa_subject -> party_privProt) {
	    (void) sprintf (PY_pepy,
			    "privacy protocol mismatch for access control entry %s",
			    sprintoid (oid));
	    return NOTOK;
	}
    }

    if (i <= 0) {
	(void) sprintf (PY_pepy, "no accesses defined");
	return NOTOK;
    }
    if (who == AGENT && i > 1) {
	register struct snmp_access **base,
				    **ape,
				    **app;

	if ((base = (struct snmp_access **)
			    malloc ((unsigned) (i * sizeof *base))) == NULL) {
	    (void) sprintf (PY_pepy,
			    "out of memory trying to sort %d accesses", i);
	    return NOTOK;
	}

	ape = base;
	for (sa = accessHead; sa; sa = sa -> sa_next)
	    *ape++ = sa;

	PY_pepy[0] = NULL;
	qsort ((char *) base, i, sizeof *base, (IFP) sa_compar);

	app = base;
	sa = accessHead = *app++;

	while (app < ape) {
	    sa -> sa_next = *app;
	    sa = *app++;
	}
	sa -> sa_next = NULL;

	free ((char *) base);

	if (PY_pepy[0])
	    return NOTOK;
    }
    if (debug)
	for (sa = accessHead; sa; sa = sa -> sa_next) {
	    fprintf (stderr, "access %s:\n", sprintoid (&sa -> sa_instance));
	    fprintf (stderr, "\tprivs=0x%x status=%d(%s)\n",
		     sa -> sa_privileges, sa -> sa_status,
		     rows[sa -> sa_status - 1].r_name);
	}

    if (debug)
	fprintf (stderr, "party access:\n");
    for (p = partyHead; p; p = p -> party_next) {
	for (sa = accessHead; sa; sa = sa -> sa_next)
	    if (sa -> sa_status == ROW_ACTIVE && sa -> sa_target == p) {
		p -> party_access = sa;
		if (debug) {
		    fprintf (stderr, "    party %s:\n",
			     oid2ode (p -> party_Identity));
		    fprintf (stderr, "\taccess %s\n",
			     sprintoid (&sa -> sa_instance));
		    for (sa = sa -> sa_next; sa; sa = sa -> sa_next)
			if (sa -> sa_target != p)
			    break;
		        else
			    if (sa -> sa_status == ROW_ACTIVE)
				fprintf (stderr, "\taccess %s\n",
					 sprintoid (&sa -> sa_instance));
		}
		break;
	    }
    }

    if (who == MANAGER)
	return OK;

    i = 0;
    for (vu = viewHead; vu; vu = vu -> vu_next) {
	int	j;
	register OID oid = &vu -> vu_instance;

	i++;
	oid -> oid_nelem = 1 + (j = vu -> vu_subtree -> oid_nelem);
	if ((oid -> oid_elements = (unsigned int *)
			calloc ((unsigned int) oid -> oid_nelem,
				sizeof *oid -> oid_elements)) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory");
	    return NOTOK;
	}
	oid -> oid_elements[0] = vu -> vu_index;
	bcopy ((char *) vu -> vu_subtree -> oid_elements,
	       (char *) (oid -> oid_elements + 1),
	       j * sizeof *oid -> oid_elements);
    }

    if (i <= 0) {
	(void) sprintf (PY_pepy, "no views defined");
	return NOTOK;
    }
    if (i > 1) {
	register struct snmp_view **base,
				  **vpe,
				  **vpp;

	if ((base = (struct snmp_view **)
			    malloc ((unsigned) (i * sizeof *base))) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory trying to sort %d views",
			    i);
	    return NOTOK;
	}

	vpe = base;
	for (vu = viewHead; vu; vu = vu -> vu_next)
	    *vpe++ = vu;

	PY_pepy[0] = NULL;
	qsort ((char *) base, i, sizeof *base, (IFP) vu_compar);

	vpp = base;
	vu = viewHead = *vpp++;

	while (vpp < vpe) {
	    vu -> vu_next = *vpp;
	    vu = *vpp++;
	}
	vu -> vu_next = NULL;

	free ((char *) base);

	if (PY_pepy[0])
	    return NOTOK;
    }
    if (debug)
	for (vu = viewHead; vu; vu = vu -> vu_next) {
	    fprintf (stderr, "view %s:\n", sprintoid (&vu -> vu_instance));
	    fprintf (stderr, "\tstatus=%d(%s)\n", vu -> vu_status,
		     rows[vu -> vu_status - 1].r_name);
	    fprintf (stderr, "\tsubtree=%s\n\ttype=%d\n",
		     oid2ode (vu -> vu_subtree), vu -> vu_type);
	    if (vu -> vu_mask)
		fprintf (stderr, "\tmask=%s\n",
			 u_char2str ((u_char *) vu -> vu_mask -> qb_forw 
							      -> qb_data,
				     vu -> vu_mask -> qb_len));
	}

    return export_views ();
}

/*  */

static int  export_views ()
{
    int	    i;
    register struct snmp_context *sc;
    register struct snmp_view *vu;
    struct snmp_view *viewTail;

    for (vu = viewHead; vu; vu = vu -> vu_next) {
	vu -> vu_refcnt = 0;

	for (sc = contextHead; sc; sc = sc -> sc_next)
	    if (sc -> sc_view == vu -> vu_index)
		vu -> vu_refcnt++;
    }

    view_granular = 0;
    for (vu = viewHead; vu; vu = vu -> vu_next)
	if (vu -> vu_status == ROW_ACTIVE && vu -> vu_refcnt > 0) {
	    OID	    tree = vu -> vu_subtree;
	    OT	    ot;

	    for (ot = anchor; ot; ot = ot -> ot_next)
		if (ot -> ot_getfnx) {
		    OID	    object = ot -> ot_name;

		    if (object -> oid_nelem < tree -> oid_nelem
			    && bcmp ((char *) object -> oid_elements,
				     (char *) tree -> oid_elements,
				     object -> oid_nelem
				         * sizeof *object -> oid_elements)
				    == 0) {
			view_granular = 1;
			break;
		    }
		}

	    if (view_granular)
		break;
	}

    i = 0;
    for (vu = viewHead; vu; vu = vu -> vu_next)
	i++;

    if (i > 1) {
	register struct snmp_view **base,
				  **vpe,
				  **vpp;

	if ((base = (struct snmp_view **)
			    malloc ((unsigned) (i * sizeof *base))) == NULL) {
	    (void) sprintf (PY_pepy, "out of memory trying to sort %d views",
			    i);
	    return NOTOK;
	}

	vpe = base;
	for (vu = viewHead; vu; vu = vu -> vu_next)
	    *vpe++ = vu;

	vpp = &viewTail;
	while (vpe > base) {
	    vu = *vpp = *--vpe;
	    vpp = &vu -> vu_prev;
	}
	*vpp = NULL;

	free ((char *) base);
    }
    else
	viewTail = viewHead;
    
    if (debug) {
	fprintf (stderr, "views in reverse...\n");
	for (vu = viewTail; vu; vu = vu -> vu_prev)
	    fprintf (stderr, "\tview %s:\n", sprintoid (&vu -> vu_instance));
    }

    for (sc = contextHead; sc; sc = sc -> sc_next)
	if (sc -> sc_view > 0) {
	    sc -> sc_vu = NULL;

	    for (vu = viewTail; vu; vu = vu -> vu_prev)
		if (vu -> vu_status == ROW_ACTIVE
		        && vu -> vu_index == sc -> sc_view) {
		    sc -> sc_vu = vu;
		    if (debug) {
			fprintf (stderr, "    context %s (%s):\n",
				 oid2ode (sc -> sc_identity),
				 u_char2str (sc -> sc_name,
					     sc -> sc_name_len));
			fprintf (stderr, "\tview %s\n",
				 sprintoid (&vu -> vu_instance));
			for (vu = vu -> vu_prev; vu; vu = vu -> vu_prev)
			    if (vu -> vu_index != sc -> sc_view)
				break;
			    else
				if (vu -> vu_status == ROW_ACTIVE)
				    fprintf (stderr, "\tview %s\n",
					     sprintoid (&vu -> vu_instance));
		    }
		    break;
		}

	    if (!sc -> sc_vu)
		sc -> sc_status = ROW_NOTREADY;
	    else
		if (sc -> sc_status == ROW_NOTREADY)
		    sc -> sc_status = ROW_NOTINSVC;
	}

    return OK;
}

/*  */

/*
 * function returns the approximate number of seconds that have elapsed since
 * midnight, January 1, 1992.
 */

u_long	party_time_now ()
{
  struct timeval now;
  static long	jan1_1992 = 8000L * 24 * 60 * 60;

  (void) gettimeofday (&now, (struct timezone *) 0);
  if (now.tv_sec == 0x7fffffff)
      return ((u_long) 0xffffffff);
  return ((u_long) (now.tv_sec - jan1_1992));
}

/*  */    

char   *party_address (p)
register struct snmp_party *p;
{
    static char buffer[BUFSIZ];

    if ((p -> party_tDomain != SNMPUDPDOMAIN
	        && p -> party_tDomain != RFC1157DOMAIN)
	    || p -> party_TAddress_len != 6) {
	(void) sprintf (buffer, "%s/%s", oid2ode (p -> party_TDomain),
			u_char2str (p -> party_TAddress,
				    p -> party_TAddress_len));

	return buffer;
    }

    (void) sprintf (buffer, "%d.%d.%d.%d/%d",
		    p -> party_TAddress[0] & 0xff,
		    p -> party_TAddress[1] & 0xff,
		    p -> party_TAddress[2] & 0xff,
		    p -> party_TAddress[3] & 0xff,
		    (p -> party_TAddress[4] & 0xff) << 8
		        | (p -> party_TAddress[5] & 0xff));
    if (strcmp (buffer, "0.0.0.0/0") == 0)
	buffer[0] = NULL;

    return buffer;
}

/*  */

struct snmp_party *lookup_party_aux (oid, who, valid)
register OID oid;
int	who,
    	valid;
{
    register struct snmp_party *p;

    for (p = partyHead; p; p = p -> party_next)
	if (oid_cmp (oid, p -> party_Identity) == 0) {
	    if (valid && p -> party_status != ROW_ACTIVE)
		break;

	    if (who == EITHER || who == p -> party_location)
		return p;

	    break;
	}

  return NULLPARTY;
}


struct snmp_party *index_party_aux (idx, who, valid)
int	idx,
	who,
    	valid;
{
    register struct snmp_party *p;

    for (p = partyHead; p; p = p -> party_next)
	if (p -> party_index == idx) {
	    if (valid && p -> party_status != ROW_ACTIVE)
		break;

	    if (who == EITHER || who == p -> party_location)
		return p;

	    break;
	}

  return NULLPARTY;
}

/*  */

struct snmp_context *lookup_context_aux (oid, valid)
register OID oid;
int	valid;
{
    register struct snmp_context *sc;

    for (sc = contextHead; sc; sc = sc -> sc_next)
	if (sc -> sc_identity && oid_cmp (oid, sc -> sc_identity) == 0) {
	    if (valid && sc -> sc_status != ROW_ACTIVE)
		break;

	    return sc;
	}

  return NULL;
}


struct snmp_context *index_context_aux (idx, valid)
int	idx,
	valid;
{
    register struct snmp_context *sc;

    for (sc = contextHead; sc; sc = sc -> sc_next)
	if (sc -> sc_index == idx) {
	    if (valid && sc -> sc_status != ROW_ACTIVE)
		break;

	    return sc;
	}

  return NULL;
}

/*  */

char   *u_char2str (base, cc)
u_char *base;
int	cc;
{
    register char   *bp;
    register u_char *dp;
    u_char     *ep;
    static int  len;
    static char *buffer = NULL;

    if (base == NULL) {
	if (buffer)
	    free (buffer), buffer = NULL;

	return NULL;
    }

    if (cc <= 0)
	return "\"\"";
    ep = (dp = base) + cc;
    if (lexnequ ((char *) base, "0x", sizeof "0x" - 1))
	for (ep = (dp = base) + cc; dp < ep; dp++)
	    if (!isprint (*dp) && *dp != ' ')
		break;
    if (dp < ep)
	cc *= 3;
    cc += 3;

    if (buffer && len < cc)
	free (buffer), buffer = NULL;
    if (buffer == NULL && (buffer = malloc ((unsigned) (len = cc))) == NULL)
	return NULL;

    if (dp < ep) {
	char   *s = "0x";

	bp = buffer;
	for (dp = base; dp < ep; dp++) {
	    (void) sprintf (bp, "%s%02x", s, *dp & 0xff);
	    bp += strlen (bp);
	    s = ":";
	}
	*bp = NULL;
    }
    else {
	cc -= 3;
	(void) sprintf (buffer, "\"%*.*s\"", cc, cc, base);
    }

    return buffer;
}

/*  */

int	sync_party (prefix, who, force)
char   *prefix;
int	who,
	force;
{
    int	    mask,
	    result,
	    uid,
	    gid;
    char    cate[BUFSIZ],
	    date[BUFSIZ],
	    file[BUFSIZ];
    register struct snmp_party *p, **pp;
    register struct snmp_context *sc, **pc;
    register struct snmp_access *sa, **pa;
    register struct snmp_view *vu, **pu;
    struct stat st;
    FILE   *fp;

    if (force)
	party_changed = context_changed = access_changed = view_changed = 1;

    if (party_changed || context_changed || access_changed || view_changed) {
	long	now;
	register struct tm *tm;
	struct UTCtime uts;
	register struct UTCtime *ut = &uts;

	(void) time (&now);
	if (tm = gmtime (&now))
	    tm2ut (tm, ut);
	else
	    bzero ((char *) ut, sizeof *ut);
	(void) strcpy (date, gent2str (ut));
	(void) strcpy (cate, ctime (&now));

	mask = umask (0077);
	uid = getuid (), gid = getgid ();

	if (debug)
	    party_changed = context_changed = access_changed = view_changed =1;
    }

    for (pp = &partyHead; p = *pp; )
	 if (p -> party_status == ROW_DESTROY) {
	     *pp = p -> party_next;

	     free_party_aux (p, 0);

	     party_changed = 1;
	 }
        else
	    pp = &p -> party_next;

    for (pc = &contextHead; sc = *pc; )
	 if (sc -> sc_status == ROW_DESTROY) {
	     *pc = sc -> sc_next;

	     free_context_aux (sc, 0);

	     context_changed = 1;
	 }
        else
	    pc = &sc -> sc_next;

    for (pa = &accessHead; sa = *pa; )
	 if (sa -> sa_status == ROW_DESTROY) {
	     *pa = sa -> sa_next;

	     free_access_aux (sa, 0);

	     access_changed = 1;
	}
        else
	    pa = &sa -> sa_next;

    for (pu = &viewHead; vu = *pu; )
	 if (vu -> vu_status == ROW_DESTROY) {
	     *pu = vu -> vu_next;

	     free_view_aux (vu, 0);

	     view_changed = 1;
	 }
        else
	    pu = &vu -> vu_next;

    result = OK;
    if (party_changed) {
	context_changed = access_changed = 1;

	if (who == AGENT) {
	    for (p = partyHead; p; p = p -> party_next)
		p -> party_flags &= ~(PARTY_PROXYD | PARTY_PROXYS);

	    for (sc = contextHead; sc; sc = sc -> sc_next)
		if (sc -> sc_view == 0) {
		    sc -> sc_src = lookup_party (sc -> sc_src_oid, AGENT);
		    if (sc -> sc_dst = p = lookup_party (sc -> sc_dst_oid,
							MANAGER)) {
			p -> party_flags |= PARTY_PROXYD;
			if (p -> party_status != ROW_ACTIVE)
			    sc -> sc_status = ROW_NOTREADY;
			else
			    if (p -> party_tDomain == SNMPUDPDOMAIN) {
				if (p = sc -> sc_src) {
				    p -> party_flags |= PARTY_PROXYS;
				    if (p -> party_status != ROW_ACTIVE
					    || !sc -> sc_ctx)
					sc -> sc_status = ROW_NOTREADY;
				    else
					if (sc -> sc_status == ROW_NOTREADY)
					    sc -> sc_status = ROW_NOTINSVC;
				}
				else
				    sc -> sc_status = ROW_NOTREADY;
			    }
			    else
				if (sc -> sc_status == ROW_NOTREADY)
				    sc -> sc_status = ROW_NOTINSVC;
		    }
		    else
			sc -> sc_status = ROW_NOTREADY;
		}
	}

	(void) sprintf (file, "%s%s", prefix, PUBLIC);
	if (stat (file, &st) == NOTOK)
	    st.st_mode = 0600, st.st_uid = uid, st.st_gid = gid;

	(void) sprintf (file, "%s%s.tmp", prefix, PUBLIC);
	if (fp = fopen (file, "w")) {
	    if (fchmod (fileno (fp), (int) st.st_mode) == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed", "fchmod of %s (%d) to 0%o",
			 file, fileno (fp), 0600);
	    }
	    if (fchown (fileno (fp), (int) st.st_uid, (int) st.st_gid)
		    == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed",
			 "fchmod of %s (%d) to %d,%d",
			 file, fileno (fp), (int) st.st_uid, (int) st.st_gid);
	    }

	    fprintf (fp, "# %spublic - created %s", prefix, cate);

	    for (p = partyHead; p; p = p -> party_next) {
		if (force && p -> party_status != ROW_ACTIVE)
		    continue;
		if (p -> party_storage == STORAGE_VOLATILE)
		    continue;

		fprintf (fp, "\n%c %s \\\n",
			   p ->  party_location == AGENT ? 'L' 
		         : (p -> party_flags & PARTY_PROXYD) ? 'P' : 'M',
			 sprintoid (p ->  party_Identity));
		fprintf (fp, "\t%s %s %d \\\n", oid2ode (p ->  party_TDomain),
			 u_char2str (p ->  party_TAddress,
				     p ->  party_TAddress_len),
			 p -> party_MaxMessageSize);
		fprintf (fp, "\t%s %lu %ld \\\n",
			 oid2ode (p ->  party_AuthProtocol),
			 p -> party_AuthStart, p -> party_AuthLifetime);
		fprintf (fp, "\t%s", oid2ode (p -> party_PrivProtocol));
		if (p -> party_AuthPublic)
		    fprintf (fp, " \\\n\t%s",
			     u_char2str (p -> party_AuthPublic,
					 p -> party_AuthPublic_len));
		if (p -> party_PrivPublic) {
		    if (!p -> party_AuthPublic)
			fprintf (fp, " \\\n\t\"\"");
		    fprintf (fp, " \\\n\t%s",
			     u_char2str (p -> party_PrivPublic,
					 p -> party_PrivPublic_len));
		}
		fprintf (fp, "\n");
	    }

	    if (fflush (fp) == EOF) {
		(void) sprintf (PY_pepy, "error writing %s: %s", file,
				sys_errname (errno));
		result = NOTOK;
	    }
	    (void) fclose (fp);
	}
	else {
	    (void) sprintf (PY_pepy, "unable to write %s: %s", file,
			    sys_errname (errno));
	    result = NOTOK;
	}

	if (result == NOTOK)
	    goto do_context;

	(void) sprintf (file, "%s%s", prefix, PRIVATE);
	if (stat (file, &st) == NOTOK)
	    st.st_mode = 0600, st.st_uid = uid, st.st_gid = gid;

	(void) sprintf (file, "%s%s.tmp", prefix, PRIVATE);
	if (fp = fopen (file, "w")) {
	    if (fchmod (fileno (fp), (int) st.st_mode) == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed", "fchmod of %s (%d) to 0%o",
			 file, fileno (fp), 0600);
	    }
	    if (fchown (fileno (fp), (int) st.st_uid, (int) st.st_gid)
		    == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed",
			 "fchmod of %s (%d) to %d,%d",
			 file, fileno (fp), (int) st.st_uid, (int) st.st_gid);
	    }

	    fprintf (fp, "# %sprivate - created %s", prefix, cate);

	    for (p = partyHead; p; p = p -> party_next) {
		if (force && p -> party_status != ROW_ACTIVE)
		    continue;
		if (p -> party_storage == STORAGE_VOLATILE)
		    continue;

		fprintf (fp, "\n%s\t%s",
			 sprintoid (p ->  party_Identity),
			 rows[p -> party_status - 1].r_name);
		if (p -> party_AuthPrivate)
		    fprintf (fp, "\t%s",
			     u_char2str (p -> party_AuthPrivate,
					 p -> party_AuthPrivate_len));
		if (p -> party_PrivPrivate) {
		    if (!p -> party_AuthPrivate)
			fprintf (fp, "\t\"\"");
		    fprintf (fp, " %s",
			     u_char2str (p -> party_PrivPrivate,
					 p -> party_PrivPrivate_len));
		}
		fprintf (fp, "\n");
	    }

	    if (fflush (fp) == EOF) {
		(void) sprintf (PY_pepy, "error writing %s: %s", file,
				sys_errname (errno));
		result = NOTOK;
	    }
	    (void) fclose (fp);
	}
	else {
	    (void) sprintf (PY_pepy, "unable to write %s: %s", file,
			    sys_errname (errno));
	    result = NOTOK;
	}
    }

do_context: ;
    if (context_changed) {
	access_changed = view_changed = 1;

	if (result == NOTOK)
	    goto do_access;

	(void) sprintf (file, "%s%s", prefix, CONTEXT);
	if (stat (file, &st) == NOTOK)
	    st.st_mode = 0600, st.st_uid = uid, st.st_gid = gid;

	(void) sprintf (file, "%s%s.tmp", prefix, CONTEXT);
	if (fp = fopen (file, "w")) {
	    if (fchmod (fileno (fp), (int) st.st_mode) == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed", "fchmod of %s (%d) to 0%o",
			 file, fileno (fp), 0600);
	    }
	    if (fchown (fileno (fp), (int) st.st_uid, (int) st.st_gid)
		    == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed",
			 "fchmod of %s (%d) to %d,%d",
			 file, fileno (fp), (int) st.st_uid, (int) st.st_gid);
	    }

	    fprintf (fp, "# %scontext - created %s", prefix, cate);

	    for (sc = contextHead; sc; sc = sc -> sc_next) {
		if (force && sc -> sc_status != ROW_ACTIVE)
		    continue;
		if (sc -> sc_storage == STORAGE_VOLATILE)
		    continue;

		fprintf (fp, "\n%c %s %s %s \\\n",
			 sc ->  sc_location == AGENT ? 'A' : 'M',
			 sprintoid (sc -> sc_identity),
			 u_char2str (sc -> sc_name, sc -> sc_name_len),
			 rows[sc -> sc_status - 1].r_name);
		if (sc -> sc_view > 0)
		    fprintf (fp, "\tview %d %s %s",
			     sc -> sc_view,
			     u_char2str (sc -> sc_entity, sc -> sc_entity_len),
			     oid2ode (sc -> sc_time));
		else {
		    fprintf (fp, "\tproxy %s \\\n",
			     sprintoid (sc -> sc_dst_oid));
		    fprintf (fp, "\t%s \\\n",
			     sprintoid (sc -> sc_src_oid));
		    fprintf (fp, "\t%s",
			     sprintoid (sc -> sc_ctx));
		}
		fprintf (fp, "\n");
	    }

	    if (fflush (fp) == EOF) {
		(void) sprintf (PY_pepy, "error writing %s: %s", file,
				sys_errname (errno));
		result = NOTOK;
	    }
	    (void) fclose (fp);
	}
	else {
	    (void) sprintf (PY_pepy, "unable to write %s: %s", file,
			    sys_errname (errno));
	    result = NOTOK;
	}
    }

do_access: ;
    if (who == MANAGER)
	view_changed = 0;
    if (access_changed) {
	for (pa = &accessHead; sa = *pa; ) {
	    sa -> sa_target = index_party_aux (sa -> sa_target_index, EITHER,
					       0);
	    sa -> sa_subject = index_party_aux (sa -> sa_subject_index, EITHER,
						0);
	    sa -> sa_resources = index_context_aux (sa -> sa_resources_index,
						    0);

	    if (!sa -> sa_target || !sa -> sa_subject || !sa -> sa_resources) {
		*pa = sa -> sa_next;

		free_access_aux (sa, 0);
	    }
	    else
		pa = &sa -> sa_next;
	}

	if (debug)
	    fprintf (stderr, "party access:\n");
	for (p = partyHead; p; p = p -> party_next) {
	    p -> party_access = NULL;

	    for (sa = accessHead; sa; sa = sa -> sa_next)
		if (p == sa -> sa_target) {
		    p -> party_access = sa;
		    if (debug) {
			fprintf (stderr, "    party %s:\n",
				 oid2ode (p -> party_Identity));
			fprintf (stderr, "\taccess %s\n",
				 sprintoid (&sa -> sa_instance));
			for (sa = sa -> sa_next; sa; sa = sa -> sa_next)
			    if (sa -> sa_target != p)
				break;
			    else
				if (sa -> sa_status == ROW_ACTIVE)
				    fprintf (stderr, "\taccess %s\n",
					     sprintoid (&sa -> sa_instance));
		    }
		    break;
		}
	}

	if (result == NOTOK)
	    goto do_view;

	(void) sprintf (file, "%s%s", prefix, ACCESS);
	if (stat (file, &st) == NOTOK)
	    st.st_mode = 0600, st.st_uid = uid, st.st_gid = gid;

	(void) sprintf (file, "%s%s.tmp", prefix, ACCESS);
	if (fp = fopen (file, "w")) {
	    if (fchmod (fileno (fp), (int) st.st_mode) == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed", "fchmod of %s (%d) to 0%o",
			 file, fileno (fp), 0600);
	    }
	    if (fchown (fileno (fp), (int) st.st_uid, (int) st.st_gid)
		    == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed",
			 "fchmod of %s (%d) to %d,%d",
			 file, fileno (fp), (int) st.st_uid, (int) st.st_gid);
	    }

	    fprintf (fp, "# %saccess - created %s", prefix, cate);

	    for (sa = accessHead; sa; sa = sa -> sa_next) {
		register struct perm *pm;

		if (force && sa -> sa_status != ROW_ACTIVE)
		    continue;
		if (sa -> sa_storage == STORAGE_VOLATILE)
		    continue;

		fprintf (fp, "\n%s \\\n",
			 sprintoid (sa -> sa_target -> party_Identity));
		fprintf (fp, "\t%s \\\n",
			 sprintoid (sa -> sa_subject -> party_Identity));
		fprintf (fp, "\t%s \\\n",
			 sprintoid (sa -> sa_resources -> sc_identity));
		fprintf (fp, "\t%s", rows[sa -> sa_status - 1].r_name);
		for (pm = perms; pm -> p_name; pm++)
		    if (sa -> sa_privileges & pm -> p_value)
			fprintf (fp, " %s", pm -> p_name);
		if (!sa -> sa_next
		        || sa -> sa_next -> sa_target_index
		    		!= sa -> sa_target_index)
		    fprintf (fp, "\n");
	    }

	    fprintf (fp, "\n");
	    if (fflush (fp) == EOF) {
		(void) sprintf (PY_pepy, "error writing %s: %s", file,
				sys_errname (errno));
		result = NOTOK;
	    }
	    (void) fclose (fp);
	}
	else {
	    (void) sprintf (PY_pepy, "unable to write %s: %s", file,
			    sys_errname (errno));
	    result = NOTOK;
	}
    }
    
do_view: ;
    if (view_changed) {
	if (export_views () == NOTOK) {
	    ADVISE (LLOG_EXCEPTIONS, NULLCP, "export_views: %s", PY_pepy);
	}

	if (result == NOTOK)
	    goto wrap_up;

	(void) sprintf (file, "%s%s", prefix, VIEW);
	if (stat (file, &st) == NOTOK)
	    st.st_mode = 0600, st.st_uid = uid, st.st_gid = gid;

	(void) sprintf (file, "%s%s.tmp", prefix, VIEW);
	if (fp = fopen (file, "w")) {
	    if (fchmod (fileno (fp), (int) st.st_mode) == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed", "fchmod of %s (%d) to 0%o",
			 file, fileno (fp), 0600);
	    }
	    if (fchown (fileno (fp), (int) st.st_uid, (int) st.st_gid)
		    == NOTOK) {
		ADVISE  (LLOG_EXCEPTIONS, "failed",
			 "fchmod of %s (%d) to %d,%d",
			 file, fileno (fp), (int) st.st_uid, (int) st.st_gid);
	    }

	    fprintf (fp, "# %sview - created %s", prefix, cate);

	    for (vu = viewHead; vu; vu = vu -> vu_next) {
		if (force && vu -> vu_status != ROW_ACTIVE)
		    continue;
		if (vu -> vu_storage == STORAGE_VOLATILE)
		    continue;

		fprintf (fp, "\n%d\t%s\t%s\t%s", vu -> vu_index,
			 rows[vu -> vu_status - 1].r_name,
			 vu -> vu_type == VIEW_INCLUDED ? "included"
			 				: "excluded",
			 oid2ode (vu -> vu_subtree));
		if (vu -> vu_mask)
		    fprintf (fp, "\t%s", u_char2str ((u_char *) vu -> vu_mask
							           -> qb_forw
							           -> qb_data,
						     vu -> vu_mask -> qb_len));
		fprintf (fp, "\n");
	    }

	    fprintf (fp, "\n");
	    if (fflush (fp) == EOF) {
		(void) sprintf (PY_pepy, "error writing %s: %s", file,
				sys_errname (errno));
		result = NOTOK;
	    }
	    (void) fclose (fp);
	}
	else {
	    (void) sprintf (PY_pepy, "unable to write %s: %s", file,
			    sys_errname (errno));
	    result = NOTOK;
	}
    }

wrap_up: ;
    if (party_changed || context_changed || access_changed || view_changed)
	(void) umask (mask);

    if (result == NOTOK) {
	if (party_changed) {
	    (void) sprintf (file, "%s%s.tmp", prefix, PUBLIC);
	    (void) unlink (file);
	    (void) sprintf (file, "%s%s.tmp", prefix, PRIVATE);
	    (void) unlink (file);
	}
	
	if (context_changed) {
	    (void) sprintf (file, "%s%s.tmp", prefix, CONTEXT);
	    (void) unlink (file);
	}

	if (access_changed) {
	    (void) sprintf (file, "%s%s.tmp", prefix, ACCESS);
	    (void) unlink (file);
	}
	
	if (view_changed) {
	    (void) sprintf (file, "%s%s.tmp", prefix, VIEW);
	    (void) unlink (file);
	}
    }
    else {
	char    back[BUFSIZ];

	if (party_changed) {
	    (void) sprintf (file, "%s%s", prefix, PUBLIC);
	    (void) sprintf (back, "%s%s.%s", prefix, PUBLIC,
			    debug ? date : "old");
	    if (rename (file, back) == NOTOK) {
		ADVISE (LLOG_EXCEPTIONS, back, "unable to rename %s to", file);
	    }
	    (void) sprintf (back, "%s%s.tmp", prefix, PUBLIC);
	    if (rename (back, file) == NOTOK) {
		(void) sprintf (PY_pepy, "unable to rename %s to %s: %s", back,
				file, sys_errname (errno));
		result = NOTOK;
	    }

	    (void) sprintf (file, "%s%s", prefix, PRIVATE);
	    (void) sprintf (back, "%s%s.%s", prefix, PRIVATE,
			    debug ? date : "old");
	    if (rename (file, back) == NOTOK) {
		ADVISE (LLOG_EXCEPTIONS, back, "unable to rename %s to", file);
	    }
	    (void) sprintf (back, "%s%s.tmp", prefix, PRIVATE);
	    if (rename (back, file) == NOTOK) {
		(void) sprintf (PY_pepy, "unable to rename %s to %s: %s", back,
				file, sys_errname (errno));
		result = NOTOK;
	    }
	}

	if (context_changed) {
	    (void) sprintf (file, "%s%s", prefix, CONTEXT);
	    (void) sprintf (back, "%s%s.%s", prefix, CONTEXT,
			    debug ? date : "old");
	    if (rename (file, back) == NOTOK) {
		ADVISE (LLOG_EXCEPTIONS, back, "unable to rename %s to", file);
	    }
	    (void) sprintf (back, "%s%s.tmp", prefix, CONTEXT);
	    if (rename (back, file) == NOTOK) {
		(void) sprintf (PY_pepy, "unable to rename %s to %s: %s", back,
				file, sys_errname (errno));
		result = NOTOK;
	    }
	}

	if (access_changed) {
	    (void) sprintf (file, "%s%s", prefix, ACCESS);
	    (void) sprintf (back, "%s%s.%s", prefix, ACCESS,
			    debug ? date : "old");
	    if (rename (file, back) == NOTOK) {
		ADVISE (LLOG_EXCEPTIONS, back, "unable to rename %s to", file);
	    }
	    (void) sprintf (back, "%s%s.tmp", prefix, ACCESS);
	    if (rename (back, file) == NOTOK) {
		(void) sprintf (PY_pepy, "unable to rename %s to %s: %s", back,
				file, sys_errname (errno));
		result = NOTOK;
	    }
	}

	if (view_changed) {
	    (void) sprintf (file, "%s%s", prefix, VIEW);
	    (void) sprintf (back, "%s%s.%s", prefix, VIEW,
			    debug ? date : "old");
	    if (rename (file, back) == NOTOK) {
		ADVISE (LLOG_EXCEPTIONS, back, "unable to rename %s to", file);
	    }
	    (void) sprintf (back, "%s%s.tmp", prefix, VIEW);
	    if (rename (back, file) == NOTOK) {
		(void) sprintf (PY_pepy, "unable to rename %s to %s: %s", back,
				file, sys_errname (errno));
		result = NOTOK;
	    }
	}	    
    }

    party_changed = access_changed = view_changed = 0;

    return result;
}

/*  */

void	exor (a, alen, b, blen)
register u_char *a;
int	alen;
register u_char *b;
register int	 blen;
{
    if (alen < blen)
	blen = alen;

    while (blen-- > 0) {
	*a = (*a & 0xff) ^ (*b++ & 0xff);
	a++;
    }
}


