/* party.h - SNMPv2 party databases */

/* 
 * $Header$
 *
 *
 * $Log$
 */

/*
 *			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#ifndef	_module_SNMP_defined_
#ifndef	PEPYPATH
#include <isode/pepsy/SNMP-types.h>
#else
#include "SNMP-types.h"
#endif
#endif

/*  */

struct snmp_party {
    struct snmp_party *party_next;	/* next in linked list */


/* identity information */

    OID		party_Identity;		/* identity */
    int		party_index;		/* index */


/* transport information */

    OID		party_TDomain;		/* transport domain */
    int		party_tDomain;
#define	BADOMAIN	(-1)
#define	SNMPUDPDOMAIN	0
#define	RFC1157DOMAIN	1

    u_char     *party_TAddress;		/*   .. address */
    int		party_TAddress_len;	/*   .. .. */

    int		party_MaxMessageSize;	/* max cotes party will accept */
#define	MAXSNMP		  484
#define	BIGSNMP		65507

    int		party_location;
#define AGENT	1
#define MANAGER	2
#define EITHER	3


/* authentication information */

    OID		party_AuthProtocol;	/* authentication protocol */
    int		party_authProt;
#define	BADAUTH	      (-1)
#define NOAUTH		0
#define MD5AUTHPROT	1
#define	RFC1157NOAUTH	2

    u_long	party_AuthStart;	/* party's view of time */

					/* authentication keys... */
    u_char     *party_AuthPrivate;		/* private */
    int		party_AuthPrivate_len;		/* .. */
    u_char     *party_AuthPublic;		/* public */
    int		party_AuthPublic_len;		/*   .. */

    long	party_AuthLifetime;	/* party's lifetime */


/* privacy information */

    OID		party_PrivProtocol;	/* privacy protocol */
    int		party_privProt;
#define	BADPRIV	      (-1)
#define NOPRIV		0
#define DESPRIVPROT	1

					/* privacy keys... */
    u_char     *party_PrivPrivate;		/* private */
    int		party_PrivPrivate_len;		/*   .. */
    u_char     *party_PrivPublic;		/* public */
    int		party_PrivPublic_len;		/*   .. */


/* miscellaneous information */

    int		party_status;
    int		party_storage;

    u_long	party_traps;		/* snmpTrapNumbers */

    int		party_testmode;


/* internal information */

    u_long	party_flags;
#define	PARTY_NULL	0x00
#define	PARTY_PROXYD	0x01		/* this party is a proxy destination */
#define	PARTY_PROXYS	0x02		/* this party is a proxy source */

    struct snmp_access *party_access;	/* fast-pointer into access table */

    int		party_clone;
#define	PARTY_CLONING	0x0001		/* cloning in process */
#define	PARTY_CLONED	0x0002		/* cloning completed successfully */
#define	PARTY_AUTHPROT	0x0004		/* done once, on first set */
#define	PARTY_AUTHPUB	0x0008		/*   .. */
#define	PARTY_AUTHLIFE	0x0010		/*   .. */
#define	PARTY_PRIVPROT	0x0020		/*   .. */
#define	PARTY_PRIVPUB	0x0040		/*   .. */
#define	PARTY_AUTHPRIV	0x0080		/* done at least once, at any time */
#define	PARTY_PRIVPRIV	0x0100		/*   .. */
    struct snmp_party *party_template;

    int		party_fresh;
    int		party_new_MaxMessageSize;
    int		party_new_Local;
    long	party_new_AuthStart;
    long	party_old_AuthStart;
    u_char     *party_new_AuthPrivate;
    int		party_new_AuthPrivate_len;
    u_char     *party_new_AuthPublic;
    int		party_new_AuthPublic_len;
    u_char     *party_new_PrivPublic;
    int		party_new_PrivPublic_len;
    u_char     *party_new_PrivPrivate;
    int		party_new_PrivPrivate_len;
    int		party_new_status;
    int		party_new_storage;
    int		party_new_testmode;
    int		party_old_clone;

    struct snmp_access *party_clockzero;
    struct snmp_access *party_clocksync;
    u_long	party_timemin;
    u_long	party_timeout;
    u_long	party_timemax;
};
#define NULLPARTY ((struct snmp_party *) 0)


extern	struct snmp_party *partyHead;

/*  */

struct snmp_context {
    struct snmp_context *sc_next;	/* next in linked-list */

    u_char     *sc_name;		/* textual name */
    int		sc_name_len;		/*   .. */

    OID		sc_identity;		/* identity */
    int		sc_index;		/* index */

    int		sc_location;

    int		sc_view;		/* if a local context, then > 0 */
    struct snmp_view   *sc_vu;		/* associated view */
    u_char     *sc_entity;		/* localEntity */
    int		sc_entity_len;		/*   .. */
    OID		sc_time;		/* localTime */

					/* if a remote context */
    struct snmp_party  *sc_dst;		/* proxy destination */
    OID		sc_dst_oid;
    struct snmp_party  *sc_src;		/* proxy source */
    OID		sc_src_oid;
    OID		sc_ctx;			/* proxy context */

    int		sc_status;
    int		sc_storage;

    int		sc_fresh;
    u_char     *sc_new_name;
    int		sc_new_name_len;
    int		sc_new_location;
    int		sc_new_view;
    u_char     *sc_new_entity;
    int		sc_new_entity_len;
    OID		sc_new_time;
    OID		sc_new_dst_oid;
    OID		sc_new_src_oid;
    OID		sc_new_ctx;
    int		sc_new_status;
    int		sc_new_storage;
};


extern	struct snmp_context *contextHead;

/*  */

struct snmp_access {
    struct snmp_access *sa_next;	/* next in linked-list */

    struct OIDentifier	sa_instance;	/* instance-identifier */

    struct snmp_party  *sa_target;	/* principals */
    int		sa_target_index;

    struct snmp_party  *sa_subject;	/*   .. */
    int		sa_subject_index;

    struct snmp_context *sa_resources;  /* context */
    int		sa_resources_index;

    int	    sa_privileges;		/* SNMP operations allowed */
#define	PARTY_PRIV_NONE	0x00
#define	PARTY_PRIV_GET	0x01
#define	PARTY_PRIV_NEXT	0x02
#define	PARTY_PRIV_RESP	0x04
#define	PARTY_PRIV_SET	0x08
#define	PARTY_PRIV_TRAP	0x10
#define	PARTY_PRIV_BULK	0x20
#define	PARTY_PRIV_INFO	0x40
#define	PARTY_PRIV_NTFY	0x80

    int		sa_status;
    int		sa_storage;

    int		sa_fresh;
    int		sa_new_privileges;
    int		sa_new_status;
    int		sa_new_storage;
};


extern	struct snmp_access *accessHead;

/*  */

struct snmp_view {
    struct snmp_view *vu_next;		/* next in linked-list */
    struct snmp_view *vu_prev;		/* prev in linked-list */

    struct OIDentifier	vu_instance;	/* instance-identifier */

    int		vu_index;		/* index */

    OID		vu_subtree;		/* subtree */
    struct qbuf *vu_mask;		/* mask */

    int		vu_type;
#define	VIEW_INCLUDED	1
#define	VIEW_EXCLUDED	2

    int		vu_status;
    int		vu_storage;

    int		vu_refcnt;

    int		vu_fresh;
    int		vu_new_type;
    struct qbuf *vu_new_mask;
    int		vu_new_status;
    int		vu_new_storage;
};


extern struct snmp_view *viewHead;

/*  */

extern	OID	noAuth;
extern	OID	md5Auth;
extern	OID	rfc1157noAuth;

extern	OID	noPriv;
extern	OID	desPriv;

extern	OID	currentTime;

extern	OID	snmpUDPDomain;
extern	OID	rfc1157Domain;

extern	int	party_changed;
extern	int	context_changed;
extern	int	access_changed;
extern	int	view_changed;
extern	int	view_granular;


int	init_party (), fin_party (), sync_party ();
void	free_party_aux (), free_context_aux (), free_access_aux (),
    	free_view_aux ();
u_long	party_time_now ();
char   *party_address ();
struct snmp_party *lookup_party_aux (), *index_party_aux ();
struct snmp_context *lookup_context_aux (), *index_context_aux ();
char   *u_char2str ();
void	exor ();

#define	lookup_party(o,w)	lookup_party_aux ((o), (w), 1)
#define	index_party(i,w)	index_party_aux ((i), (w), 1)
#define	lookup_context(o)	lookup_context_aux ((o), 1)
#define	index_context(i)	index_context_aux ((i), 1)

#define	free_party(p)		free_party_aux ((p), 1)
#define	free_context(c)		free_context_aux ((c), 1)
#define	free_access(a)		free_access_aux ((a), 1)
#define	free_view(v)		free_view_aux ((v), 1)

/*  */

#define AUTH_KEY_LENGTH 16

#define	AUTH_UNKNOWN	 0
#define	AUTH_MISMATCH	 1
#define	AUTH_WRONGAUTH	 2
#define	AUTH_LIFETIME	 3
#define	AUTH_DIGEST	 4

int	authenticate_incoming ();
struct qbuf *generate_outgoing_auth ();
PE	pe2oct2pe ();

/*  */

#define PRIV_KEY_LENGTH	 8

int	do_encryption ();
struct qbuf *do_decryption ();
int	do_decryption_fixup ();

/*  */

#define	partyIdentity		 0
#define	partyIndex		 1
#define	partyTDomain		 2
#define	partyTAddress		 3
#define	partyMaxMessageSize	 4
#define	partyLocal		 5
#define	partyAuthProtocol	 6
#define	partyAuthClock		 7
#define	partyAuthPrivate	 8
#define	partyAuthPublic		 9
#define	partyAuthLifetime	10
#define	partyPrivProtocol	11
#define	partyPrivPrivate	12
#define	partyPrivPublic		13
#define partyCloneFrom		14
#define partyStorageType	15
#define	partyStatus 		16
#define	snmpTrapNumbers		17
#define	partyTestMode		18

#define	contextIdentity		40
#define	contextName		41
#define	contextIndex		42
#define	contextLocal		43
#define	contextViewIndex	44
#define	contextLocalEntity	45
#define	contextLocalTime	46
#define	contextProxyDstParty	47
#define	contextProxySrcParty	48
#define	contextProxyContext	49
#define	contextStorageType	50
#define	contextStatus		51

#define	aclTarget		60
#define	aclSubject		61
#define	aclResources		62
#define	aclPrivileges		63
#define	aclStorageType		64
#define	aclStatus		65
