/* smux.c - SNMP proxy subroutines (the name is historical) */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/* 
 * $Header$
 *
 *
 * $Log$
 */

/*
 *			  ISODE 8.0 NOTICE
 *
 *   Acquisition, use, and distribution of this module and related
 *   materials are subject to the restrictions of a license agreement.
 *   Consult the Preface in the User's Manual for the full terms of
 *   this agreement.
 *
 *
 *			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#include <stdio.h>
#include <errno.h>
#include "objects.h"
#include "smux.h"
#include "party.h"
#include "snmp-g.h"
#include "dgram.h"
#include "internet.h"

/*  */

static	int	debug = 0;

static	int	snmp_fd = NOTOK;
static	struct type_SNMP_NetworkAddress *snmp_address = NULL;
static	struct snmp_party *agentParty;

static	int	snmp_udp = NOTOK;
static	OID	snmp_enterprise = NULLOID;
static	PE	snmp_pe = NULLPE;
static	PS	snmp_ps = NULLPS;
static	struct type_SNMP_Snmp1157Msg *snmp_msg = NULL;
static	int	snmp_size = 0;
static  struct type_SNMP_VarBindList *snmp_vp = NULL;

static	void	snmp_free ();


extern	int	errno;

/*  */

int	snmp_init (descr, objectID, aflag, pflag, dflag, zflag, bflag)
char   *descr,
       *objectID,
       *aflag,
       *pflag;
int	dflag,
	zflag,
	bflag;
{
    struct sockaddr_in in_socket;
    struct sockaddr_in *isock = &in_socket;
    OID	    oid;
    static int	once_only = 0;

    if (!aflag) {
	(void) sprintf (PY_pepy, "must specify agent (aflag)");
	return NOTOK;
    }
    debug = dflag;
    
    (void) snmp_close ();
    if (once_only == 0) {
	if (descr || objectID) {
	    if (!objectID)
		snmp_enterprise = NULLOID;
	    else
		if ((snmp_enterprise = text2oid (objectID)) == NULLOID) {
		    (void) sprintf (PY_pepy, "%s: unknown object", objectID);
		    return NOTOK;
		}

	    init_system (descr, objectID, "64");
	    if (init_snmp (0, zflag) == NOTOK)
		return NOTOK;
	}
	else
	    (void) gettimeofday (&my_boottime, (struct timezone *) 0);

	if (init_party (pflag ? pflag : "party", MANAGER, 0, NULLVP) == NOTOK)
	    return NOTOK;

	once_only = 1;
    }

    if ((oid = text2oid (aflag)) == NULL) {
	(void) sprintf (PY_pepy, "unknown OID: \"%s\"", aflag);
	return NOTOK;
    }

    agentParty = lookup_party (oid, EITHER);

    oid_free (oid);

    if (agentParty == NULL) {
	(void) sprintf (PY_pepy, "no such party: %s", aflag);
	return NOTOK;
    }
    if (agentParty -> party_tDomain != RFC1157DOMAIN) {
	(void) sprintf (PY_pepy,
			"party %s not reachable via rfc1157Domain",
			oid2ode (agentParty -> party_Identity));
	return NOTOK;
    }
    if (bflag > 0) {
	if (bflag > MAXDGRAM)
	    bflag = MAXDGRAM;
	else
	    if (bflag < MAXSNMP)
		bflag = MAXSNMP;
	agentParty -> party_MaxMessageSize = bflag;
    }

    if (snmp_address)
	qb_free (snmp_address);
    if ((snmp_address = str2qb ((char *) agentParty -> party_TAddress,
				4, 1)) == NULL) {
	(void) sprintf (PY_pepy, "out of memory");
	return NOTOK;
    }

    bzero ((char *) isock, sizeof *isock);
    isock -> sin_family = AF_INET;
    bcopy ((char *) agentParty -> party_TAddress, (char *) &isock -> sin_addr,
	   sizeof isock -> sin_addr);
    bcopy ((char *) (agentParty -> party_TAddress + sizeof isock -> sin_addr),
	   (char *) &isock -> sin_port, sizeof isock -> sin_port);

    if ((snmp_fd = start_udp_server (isock, 0, 0, 0)) == NOTOK) {
	(void) sprintf (PY_pepy, "party %s at ",
			oid2ode (agentParty -> party_Identity));
	(void) sprintf (PY_pepy + strlen (PY_pepy), "%s: %s",
			party_address (agentParty), sys_errname (errno));
	return NOTOK;
    }
    if (debug)
	fprintf (stderr, "listening for party %s at %s %d\n",
		 oid2ode (agentParty -> party_Identity),
		 party_address (agentParty),
		 agentParty -> party_MaxMessageSize);

    return snmp_fd;
}

/*  */

int	snmp_wait (event, secs)
struct type_SNMP_PDUs **event;
int	secs;
{
    int	    maxsize,
    	    result;
    char    source[BUFSIZ];
    fd_set  mask;
    struct sockaddr_in in_socket;
    register struct sockaddr_in *isock = &in_socket;
    struct type_SNMP_PDU *parm;
    struct type_SNMP_VarBindList *vp;

    snmp_free ();

    if (snmp_fd == NOTOK) {
	(void) sprintf (PY_pepy, "SNMP not opened");
	return DONE;
    }
    *event = NULL;

    FD_ZERO (&mask);
    FD_SET (snmp_fd, &mask);

    result = DONE;
    if (xselect (snmp_fd + 1, &mask, NULLFD, NULLFD, secs) == OK) {
	if (errno == EINTR || errno == EWOULDBLOCK)
	    return OK;

	(void) sprintf (PY_pepy, "xselect: %s", sys_errname (errno));
	goto out;
    }

    if (!FD_ISSET (snmp_fd, &mask))
	return OK;

    if ((snmp_udp = join_udp_client (snmp_fd, isock)) == NOTOK) {
	if (errno == EINTR || errno == EWOULDBLOCK)
	    return OK;
	
	(void) sprintf (PY_pepy, "join_udp_client: %s", sys_errname (errno));
	goto out;
    }

    (void) sprintf (source, "%s/%d", inet_ntoa (isock -> sin_addr),
		    (int) ntohs (isock -> sin_port));

    if ((snmp_ps = ps_alloc (dg_open)) == NULLPS
	    || dg_setup (snmp_ps, snmp_udp, agentParty -> party_MaxMessageSize,
			 read_udp_socket, write_udp_socket, check_udp_socket)
		    == NOTOK) {
	if (snmp_ps == NULLPS)
	    (void) sprintf (PY_pepy, "ps_alloc(dg_open): failed");
	else
	    (void) sprintf (PY_pepy, "dg_setup: %s",
			    ps_error (snmp_ps -> ps_errno));
	goto out;
    }
    snmpstat.s_packets++;
    
    result = NOTOK;
    snmp_size = snmp_ps -> ps_byteno;
    if ((snmp_pe = ps2pe (snmp_ps)) == NULLPE) {
	(void) sprintf (PY_pepy, "ps2pe: %s (%s)",
			ps_error (snmp_ps -> ps_errno), source);
	snmpstat.s_encodingerrs++;
	goto out;
    }
    snmp_size = snmp_ps -> ps_byteno - snmp_size;

    if (decode_SNMP_Snmp1157Msg (snmp_pe, 1, NULLIP, NULLVP, &snmp_msg)
	    == NOTOK) {
	if (snmp_msg)
	    free_SNMP_Snmp1157Msg (snmp_msg), snmp_msg = NULL;
	snmpstat.s_encodingerrs++;
	goto out;
    }
    if (debug)
	(void) print_SNMP_Snmp1157Msg (snmp_pe, 1, NULLIP, NULLVP, NULLCP);

    if (snmp_msg -> community -> qb_forw -> qb_forw != snmp_msg -> community
	    || agentParty -> party_AuthPrivate_len
		        != snmp_msg -> community -> qb_len
	    || bcmp ((char *) agentParty -> party_AuthPrivate,
		     snmp_msg -> community -> qb_forw -> qb_data,
		     agentParty -> party_AuthPrivate_len)) {
	(void) sprintf (PY_pepy, "invalid community (%s)", source);
	snmpstat.s_badcommunitynames++;
	if (*((integer *) snmpstat.s_enableauthentraps) == TRAPS_ENABLED)
	    (void) snmp_trap (NULLOID,
			      int_SNMP_generic__trap_authenticationFailure, 0,
			      (struct type_SNMP_VarBindList *) 0);
	goto out;
    }
    if (snmp_msg -> version != int_SNMP_version_version__1) {
	(void) sprintf (PY_pepy, "bad version number %d (%s)",
			snmp_msg -> version, source);
	snmpstat.s_badversions++;
	goto out;
    }

    switch (snmp_msg -> data -> offset) {
	case type_SNMP_PDUs_get__request:
	    snmpstat.s_ingetrequests++;
	    goto ok_request;

	case type_SNMP_PDUs_get__next__request:
	    snmpstat.s_ingetnexts++;
	    goto ok_request;

	case type_SNMP_PDUs_set__request:
	    snmpstat.s_insetrequests++;
ok_request: ;
	    parm = (*event = snmp_msg -> data) -> un.response;
	    parm -> error__status = parm -> error__index = 0;
	    maxsize = snmp_size;
	    for (vp = parm -> variable__bindings; vp; vp = vp -> next)
		maxsize -= ps_get_abs_oid (vp -> VarBind -> name)
		    		+ ps_get_abs (vp -> VarBind -> value)
				+ 2;
	    if (maxsize > agentParty -> party_MaxMessageSize) {
		(void) sprintf (PY_pepy,
				"dropOperation: %d (%s) actual=%d minimal=%d MTU=%d",
				snmp_msg -> data -> offset, source, snmp_size,
				maxsize, agentParty -> party_MaxMessageSize);
		snmpstat.s_silentdrops++;
		break;
	    }
	    {
		register struct type_SNMP_VarBindList **opp;

		if (snmp_vp)
		    free_SNMP_VarBindList (snmp_vp), snmp_vp = NULL;
		opp = &snmp_vp;

		for (vp = parm -> variable__bindings; vp; vp = vp -> next) {
		    register struct type_SNMP_VarBindList *bind;
		    register struct type_SNMP_VarBind *v;

		    if ((bind = (struct type_SNMP_VarBindList *)
					calloc (1, sizeof *bind)) == NULL) {
no_mem: ;
			(void) sprintf (PY_pepy, "out of memory");

			parm -> error__status = int_SNMP_error__status_genErr;
			if (snmp_vp)
			    free_SNMP_VarBindList (snmp_vp), snmp_vp = NULL;
			if (snmp_response () == DONE)
			    result = DONE;
			goto out;
		    }
		    *opp = bind, opp = &bind -> next;

		    if ((v = (struct type_SNMP_VarBind *)
						calloc (1, sizeof *v)) == NULL)
			goto no_mem;
		    bind -> VarBind = v;

		    if ((v -> name = oid_cpy (vp -> VarBind -> name))
			    == NULLOID)
			goto no_mem;
		    (v -> value = vp -> VarBind -> value) -> pe_refcnt++;
		}
	    }
	    return OK;

	case type_SNMP_PDUs_response:
	    snmpstat.s_ingetresponses++;
	    switch (snmp_msg -> data ->  un.response -> error__status) {
	        case int_SNMP_error__status_tooBig:
		    snmpstat.s_intoobigs++;
		    break;

		case int_SNMP_error__status_readOnly:
		    snmpstat.s_inreadonlys++;
		    break;

		case int_SNMP_error__status_noSuchName:
		    snmpstat.s_innosuchnames++;
		    break;

		case int_SNMP_error__status_badValue:
		    snmpstat.s_inbadvalues++;
		    break;

		case int_SNMP_error__status_genErr:
		    snmpstat.s_ingenerrs++;
		    break;
	    }
	    goto unexpected;

	case type_SNMP_PDUs_trap:
	    snmpstat.s_intraps++;
unexpected: ;
	    (void) sprintf (PY_pepy, "unexpectedOperation: %d (%s)",
			   snmp_msg -> data -> offset, source);
	    break;

	default:
	    (void) sprintf (PY_pepy, "badOperation %d (%s)",
			   snmp_msg -> data -> offset, source);
	    break;
    }

out: ;
    snmp_free ();
    if (result == DONE) {
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;
	qb_free (snmp_address), snmp_address = NULL;
    }

    return result;
}

/*  */

int	snmp_response ()
{
    int	    offset,
	    result;
    struct type_SNMP_Response__PDU *parm;
    struct type_SNMP_VarBindList *vp;
    PE	    pe = NULL;

    if (snmp_udp == NOTOK) {
	(void) sprintf (PY_pepy, "SNMP not expecting response");
	return NOTOK;
    }

    offset = snmp_msg -> data -> offset;
    snmp_msg -> data -> offset = type_SNMP_PDUs_response;
    parm = snmp_msg -> data ->  un.response;
    switch (parm -> error__status) {
        case int_SNMP_error__status_authorizationError:
	    if (!parm -> variable__bindings) {
		parm -> error__status = int_SNMP_error__status_noError;
		break;
	    }
	    parm -> error__index = 1;
	    /* and fall... */
	case int_SNMP_error__status_noAccess:
	case int_SNMP_error__status_noCreation:
	case int_SNMP_error__status_notWritable:
	case int_SNMP_error__status_inconsistentName:
	    parm -> error__status = int_SNMP_error__status_noSuchName;
	    break;

        case int_SNMP_error__status_wrongType:
	case int_SNMP_error__status_wrongLength:
	case int_SNMP_error__status_wrongEncoding:
	case int_SNMP_error__status_wrongValue:
	case int_SNMP_error__status_inconsistentValue:
	case int_SNMP_error__status_resourceUnavailable:
	    parm -> error__status = int_SNMP_error__status_badValue;
	    break;

	case int_SNMP_error__status_commitFailed:
	case int_SNMP_error__status_undoFailed:
	    parm -> error__status = int_SNMP_error__status_genErr;
	    break;
    }

    switch (parm -> error__status) {
	case int_SNMP_error__status_noError:
	    if (offset == type_SNMP_PDUs_get__next__request) {
		for (vp = snmp_vp; vp; vp = vp -> next)
		    snmp_size -= ps_get_abs_oid (vp -> VarBind -> name);
		for (vp = parm -> variable__bindings; vp; vp = vp -> next)
		    snmp_size += ps_get_abs_oid (vp -> VarBind -> name);
	    }
	    for (vp = snmp_vp; vp; vp = vp -> next)
		snmp_size -= ps_get_abs (vp -> VarBind -> value);
	    for (vp = parm -> variable__bindings; vp; vp = vp -> next)
		snmp_size += ps_get_abs (vp -> VarBind -> value);
	    if (snmp_size < agentParty -> party_MaxMessageSize) {
		free_SNMP_VarBindList (snmp_vp), snmp_vp = NULL;
		break;
	    }
	    parm -> error__status = int_SNMP_error__status_tooBig;
	    /* and fall... */

	default:
	    if (snmp_vp) {
		free_SNMP_VarBindList (parm -> variable__bindings);
		parm -> variable__bindings = snmp_vp, snmp_vp = NULL;
	    }
	    break;
    }

    result = NOTOK;
    if (encode_SNMP_Snmp1157Msg (&pe, 1, 0, NULLCP, snmp_msg) == NOTOK)
	goto out;
    if (debug)
	(void) print_SNMP_Snmp1157Msg (pe, 1, NULLIP, NULLVP, NULLCP);

    if (pe2ps (snmp_ps, pe) == NOTOK) {
	(void) sprintf (PY_pepy, "pe2ps: %s", ps_error (snmp_ps -> ps_errno));
/*
	result = DONE;
 */
    }
    else {
	result = OK;
	snmpstat.s_outpkts++;
	snmpstat.s_outgetresponses++;
	switch (snmp_msg -> data ->  un.response -> error__status) {
	    case int_SNMP_error__status_tooBig:
	        snmpstat.s_outtoobigs++;
		break;

	    case int_SNMP_error__status_readOnly:
		break;

	    case int_SNMP_error__status_noSuchName:
		snmpstat.s_outnosuchnames++;
		break;

	    case int_SNMP_error__status_badValue:
		snmpstat.s_outbadvalues++;
		break;

	    case int_SNMP_error__status_genErr:
		snmpstat.s_outgenerrs++;
		break;
	}
    }

out: ;
    if (pe)
	pe_free (pe);

/* BEGIN MOBY PEPSY HACK... */
    vp = snmp_msg -> data -> un.get__request -> variable__bindings;
    free_SNMP_VarBindList (vp);
    snmp_msg -> data -> un.get__request -> variable__bindings = NULL;
/* END MOBY PEPSY HACK... */
    snmp_free ();
    if (result == DONE) {
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;
	qb_free (snmp_address), snmp_address = NULL;
    }

    return result;
}

/*  */

int	snmp_trap (enterprise, generic, specific, bindings)
OID	enterprise;
int	generic,
	specific;
struct type_SNMP_VarBindList *bindings;
{
    int	    didone,
	    result;
    PE	    pe = NULL;
    PS	    ps = NULL;
    register struct snmp_context *sc;
    struct timeval now;
    struct type_SNMP_Snmp1157Msg msgs;
    register struct type_SNMP_Snmp1157Msg *msg = &msgs;
    struct type_SNMP_PDUs pdus;
    struct type_SNMP_Trap__PDU traps;
    register struct type_SNMP_Trap__PDU *trap = &traps;
    struct type_SNMP_TimeTicks snmp_stamp;

    if (snmp_fd == NOTOK) {
	(void) sprintf (PY_pepy, "SNMP not opened");
	return DONE;
    }

    bzero ((char *) msg, sizeof *msg);
    msg -> version = int_SNMP_version_version__1;
    if ((msg -> community = str2qb ((char *) agentParty ->
						    party_AuthPrivate,
				    agentParty -> party_AuthPrivate_len,
				    1))  == NULL) {
	(void) sprintf (PY_pepy, "out of memory");
	return DONE;
    }
    msg -> data = &pdus;

    bzero ((char *) &pdus, sizeof pdus);
    pdus.offset = type_SNMP_PDUs_trap;
    pdus.un.trap = trap;

    bzero ((char *) trap, sizeof *trap);
    trap -> enterprise = enterprise ? enterprise 
		  : snmp_enterprise ? snmp_enterprise
		  : agentParty -> party_Identity;
    trap -> agent__addr = snmp_address;
    if ((trap -> generic__trap = generic) == int_SNMP_generic__trap_warmStart)
	(void) gettimeofday (&my_boottime, (struct timezone *) 0);
    trap -> specific__trap = specific;
    trap -> time__stamp = &snmp_stamp;
    (void) gettimeofday (&now, (struct timezone *) 0);
    trap -> time__stamp -> parm = (now.tv_sec - my_boottime.tv_sec) * 100
					+ ((now.tv_usec - my_boottime.tv_usec)
					   			      / 10000);
    trap -> variable__bindings = bindings;

    result = NOTOK;
    if (encode_SNMP_Snmp1157Msg (&pe, 1, 0, NULLCP, msg) == NOTOK)
	goto out;
    if (debug)
	(void) print_SNMP_Snmp1157Msg (pe, 1, NULLIP, NULLVP, NULLCP);

    didone = 0;
    result = DONE;
    for (sc = contextHead; sc; sc = sc -> sc_next)
	if (sc -> sc_status == ROW_ACTIVE
		&& sc -> sc_view == 0
	        && sc -> sc_dst == agentParty) {
	    register struct snmp_party *p;
	    register struct snmp_access *sa;

	    for (sa = accessHead; sa; sa = sa -> sa_next)
		if (sa -> sa_status == ROW_ACTIVE
		        && sa -> sa_resources == sc
		        && (sa -> sa_privileges & PARTY_PRIV_NTFY)
			&& (p = sa -> sa_subject) -> party_tDomain
			            == SNMPUDPDOMAIN
			&& p -> party_TAddress_len == 6
		        && (p -> party_TAddress[0] != 0
				|| p -> party_TAddress[1] != 0
				|| p -> party_TAddress[2] != 0
				|| p -> party_TAddress[3] != 0)
			&& (p -> party_TAddress[4] != 0
				|| p -> party_TAddress[5] != 0)) {
		    struct sockaddr_in in_socket;
		    struct sockaddr_in *isock = &in_socket;

		    bzero ((char *) isock, sizeof *isock);
		    isock -> sin_family = AF_INET;
		    bcopy ((char *) p -> party_TAddress,
			   (char *) &isock -> sin_addr,
			   sizeof isock -> sin_addr);
		    bcopy ((char *) (p -> party_TAddress
				     	    + sizeof isock -> sin_addr),
			   (char *) &isock -> sin_port,
			   sizeof isock -> sin_port);

		    if ((ps = ps_alloc (dg_open)) == NULLPS
			    || dg_setup (ps, snmp_fd,
					 p -> party_MaxMessageSize,
					 read_udp_socket, write_udp_socket,
					 check_udp_socket) == NOTOK) {
			if (ps == NULLPS)
			    (void) sprintf (PY_pepy,
					    "ps_alloc(dg_open): failed");
			else
			    (void) sprintf (PY_pepy, "dg_setup: %s",
					    ps_error (ps -> ps_errno));
			goto out;
		    }

		    if (hack_dgram_socket (snmp_fd, (struct sockaddr *) isock)
			    == NOTOK) {
			(void) sprintf (PY_pepy, "hack_dgram_socket failed");
			goto out;
		    }

		    if (pe2ps (ps, pe) == NOTOK) {
			(void) sprintf (PY_pepy, "pe2ps: %s",
					ps_error (snmp_ps -> ps_errno));
			goto out;
		    }
		    snmpstat.s_outpkts++;
		    snmpstat.s_outtraps++;

		    (void) hack_dgram_socket (snmp_fd,
					      (struct sockaddr *) NULL);

		    (void) ps_free (ps), ps = NULL;

		    didone = 1;
		}
	}
    if (didone)
	result = OK;
    else {
	(void) sprintf (PY_pepy, "unable to find any trap sinks");
	result = NOTOK;
    }

out: ;
    if (pe)
	pe_free (pe);
    if (ps)
	ps_free (ps);
    if (result == DONE) {
	snmp_free ();
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;
	qb_free (snmp_address), snmp_address = NULL;
    }
    if (msg -> community)
	qb_free (msg -> community);

    return result;

}

/*  */

int	snmp_close ()
{
    snmp_free ();
    if (snmp_fd != NOTOK) {
	(void) close_udp_socket (snmp_fd), snmp_fd = NOTOK;
	qb_free (snmp_address), snmp_address = NULL;
    }

    return OK;
}

/*  */

static void  snmp_free ()
{
    if (snmp_udp != NOTOK)
	(void) close_udp_socket (snmp_udp), snmp_udp = NOTOK;
    if (snmp_ps)
	ps_free (snmp_ps), snmp_ps = NULL;
    if (snmp_pe)
	pe_free (snmp_pe), snmp_pe = NULL;
    if (snmp_msg)
	free_SNMP_Snmp1157Msg (snmp_msg), snmp_msg = NULL;
    if (snmp_vp)
	free_SNMP_VarBindList (snmp_vp), snmp_vp = NULL;
}
