-- snmp.py - SNMPv2 definitions
    
-- $Header$
--
-- Originally contributed by NYSERNet Inc.  This work was partially supported
-- by the U.S. Defense Advanced Research Projects Agency and the Rome Air
-- Development Center of the U.S. Air Force Systems Command under contract
-- number F30602-88-C-0016.
--
--
-- $Log$

--
--			  ISODE 8.0 NOTICE
--
--    Acquisition, use, and distribution of this module and related
--    materials are subject to the restrictions of a license agreement.
--    Consult the Preface in the User's Manual for the full terms of
--    this agreement.
--
--
--			4BSD/ISODE SNMPv2 NOTICE
--
--    Acquisition, use, and distribution of this module and related
--    materials are subject to the restrictions of a license agreement.
--    Consult the Preface in the User's Manual for the full terms of
--    this agreement.
--
--


SNMP DEFINITIONS ::=

BEGIN

-- top-level message

SnmpPrivMsg ::=
    [1]
	IMPLICIT SEQUENCE {
	    privDst
		OBJECT IDENTIFIER,
	    privData[1]
		IMPLICIT OCTET STRING
	}

SnmpAuthMsg ::=
    [1]
	IMPLICIT SEQUENCE {
	    authInfo
		AuthInformation,
	    authData
		SnmpMgmtCom
	}

AuthInformation ::= 
	CHOICE {
		OCTET STRING,

	    [2]
		IMPLICIT SEQUENCE {
		    authDigest
			OCTET STRING,
		    authDstTimestamp
			[APPLICATION 7] IMPLICIT INTEGER,
		    authSrcTimestamp
			[APPLICATION 7] IMPLICIT INTEGER
		}
	}


SnmpMgmtCom ::=
    [2]
	IMPLICIT SEQUENCE {
	    dstParty
		OBJECT IDENTIFIER,
            srcParty
		OBJECT IDENTIFIER,
            context
		OBJECT IDENTIFIER,
            pdu
                PDUs
	}


-- protocol data units

PDUs ::=
        CHOICE {
            get-request
                GetRequest-PDU,

            get-next-request
                GetNextRequest-PDU,

	    get-bulk-request
		GetBulkRequest-PDU,

            response
                Response-PDU,

            set-request
                SetRequest-PDU,

	    inform-request
		InformRequest-PDU,

            snmpV2-trap
                SNMPv2-Trap-PDU,

            trap
                Trap-PDU,

            xxx
                XXX-PDU
        }

GetRequest-PDU ::=
    [0]
        IMPLICIT PDU

GetNextRequest-PDU ::=
    [1]
        IMPLICIT PDU

GetBulkRequest-PDU ::=
    [5]
        IMPLICIT -- BulkPDU -- PDU

Response-PDU ::=
    [2]
        IMPLICIT PDU

SetRequest-PDU ::=
    [3]
        IMPLICIT PDU

InformRequest-PDU ::=
    [6]
        IMPLICIT PDU

SNMPv2-Trap-PDU ::=
    [7]
        IMPLICIT PDU

XXX-PDU ::=
    [42]
        IMPLICIT PDU

max-bindings
    INTEGER ::= 2147483647

PDU ::=
        SEQUENCE {
            request-id
                -- Integer32 -- INTEGER,

            error-status		-- sometimes ignored
                INTEGER {
                    noError(0),
                    tooBig(1),
                    noSuchName(2),	-- for proxy compatibility
                    badValue(3),	-- for proxy compatibility
                    readOnly(4),	-- for proxy compatibility
                    genErr(5),
		    noAccess(6),
                    wrongType(7),
                    wrongLength(8),
		    wrongEncoding(9),
		    wrongValue(10),
		    noCreation(11),
		    inconsistentValue(12),
		    resourceUnavailable(13),
		    commitFailed(14),
		    undoFailed(15),
		    authorizationError(16),
		    notWritable(17),
		    inconsistentName(18)
                },

            error-index			-- sometimes ignored
                INTEGER (0..max-bindings),

            variable-bindings		-- values are sometimes ignored
                VarBindList
        }

VarBind ::=
        SEQUENCE {
            name
                ObjectName,

--	CHOICE {
--	    value
--		ObjectSyntax,
--
--	    unSpecified          -- -- in retrieval requests
--		NULL,
--
--                               -- -- exceptions
--	    noSuchObject[0]
--		IMPLICIT NULL,
--
--	    noSuchInstance[1]
--		IMPLICIT NULL,
--
--	    endOfMibView[2]
--		IMPLICIT NULL
--	}
	    value
		ObjectSyntax
        }

VarBindList ::=
    SEQUENCE --* (SIZE (0..max-bindings)) *-- OF
	VarBind


-- auxiliary types

ObjectName ::=
    OBJECT IDENTIFIER
ObjectSyntax ::=
    ANY


-- trap logging

DisplayString ::=
    OCTET STRING

Audit ::=
	SEQUENCE {
	    source
		DisplayString,

	    dateAndTime
		GeneralizedTime,

	    sizeOfEncodingWhichFollows
		INTEGER
	}


-- snmptcl hack

SnmpV1orV2 ::=
	CHOICE {
	    snmpV1
		Snmp1157Msg,

	    snmpV2
		SnmpPrivMsg
	}


-- rfc1157 proxy

Snmp1157Msg ::=
        SEQUENCE {
            version
                INTEGER { version-1(0) },

            community
                OCTET STRING,

            data
                PDUs
        }


-- old traps

NetworkAddress ::=
    CHOICE {
	internet
	    IpAddress
    }

IpAddress ::=
    [APPLICATION 0]            -- in network-byte order
        IMPLICIT OCTET STRING (SIZE (4))
    
TimeTicks ::=
    [APPLICATION 3]
	IMPLICIT INTEGER

Trap-PDU ::=
    [4]
        IMPLICIT SEQUENCE {
            enterprise			-- type of object generating
                OBJECT IDENTIFIER,	-- trap, see sysObjectID

            agent-addr			-- address of object generating trap
                NetworkAddress,

            generic-trap	        -- generic trap type
                INTEGER {
                    coldStart(0),
                    warmStart(1),
                    linkDown(2),
                    linkUp(3),
                    authenticationFailure(4),
                    egpNeighborLoss(5),
                    enterpriseSpecific(6)
                },

            specific-trap		-- specific code, present even
                INTEGER,		-- if generic-trap is not
					-- enterpriseSpecific

            time-stamp			-- time elapsed between the last
                TimeTicks,		-- (re)initalization of the network
					-- entity and the generation of the
					-- trap

            variable-bindings		-- "interesting" information
                VarBindList
        }


-- miscellany for syntax.c

NsapAddress ::=
    [APPLICATION 5]
	OCTET STRING (SIZE (1..21))


END


%{	/* this is necessary 'cause we have UInteger32 in AuthInformation */

PE	ulong2prim ();

PE	num2prim (i, class, id)
integer	i;
PElementClass class;
PElementID	id;
{
    register integer mask,
                    sign,
                    n;
    register PElementData dp;
    register PE	    pe;

    if (class == PE_CLASS_APPL && id == 7)
	return ulong2prim ((u_long) i, class, id);

    if ((pe = pe_alloc (class, PE_FORM_PRIM, id)) == NULLPE)
	return NULLPE;

    sign = i >= 0 ? i : i ^ (-1);
    mask = 0x1ff << (((n = sizeof i) - 1) * 8 - 1);
    while (n > 1 && (sign & mask) == 0)
	mask >>= 8, n--;

    if ((pe -> pe_prim = PEDalloc (n)) == NULLPED) {
	pe_free (pe);
	return NULLPE;
    }

    for (dp = pe -> pe_prim + (pe -> pe_len = n); n-- > 0; i >>= 8)
	*--dp = i & 0xff;

    return pe;
}


u_long	prim2ulong ();

integer	prim2num (pe)
register PE	pe;
{
    register integer    i;
    register PElementData dp,
			  ep;

    if (pe -> pe_form != PE_FORM_PRIM || pe -> pe_prim == NULLPED)
	return pe_seterr (pe, PE_ERR_PRIM, NOTOK);

    if (pe -> pe_class == PE_CLASS_APPL && pe -> pe_id == 7)
	return prim2ulong (pe);

    if (pe -> pe_len > sizeof (i))
	return pe_seterr (pe, PE_ERR_OVER, NOTOK);

    pe -> pe_errno = PE_ERR_NONE;/* in case integer is NOTOK-valued */
    dp = pe -> pe_prim;
    if (pe -> pe_len > 1 && ((*dp == 0 && ((*(dp+1)) & 0x80) == 0) ||
	(*dp == 0xff && ((*(dp + 1)) & 0x80) == 0x80)))
	return pe_seterr (pe, PE_ERR_SYNTAX, NOTOK);
    i = ((*dp) & 0x80) ? (-1) : 0;
    for (ep = dp + pe -> pe_len; dp < ep;)
	i = (i << 8) | (*dp++ & 0xff);

    return i;
}
%}
