/* syntax.c - SMI syntax handling */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/* 
 * $Header$
 *
 * Originally contributed by NYSERNet Inc.  This work was partially supported
 * by the U.S. Defense Advanced Research Projects Agency and the Rome Air
 * Development Center of the U.S. Air Force Systems Command under contract
 * number F30602-88-C-0016.
 *
 *
 * $Log$
 */

/*
 *			  ISODE 8.0 NOTICE
 *
 *   Acquisition, use, and distribution of this module and related
 *   materials are subject to the restrictions of a license agreement.
 *   Consult the Preface in the User's Manual for the full terms of
 *   this agreement.
 *
 *
 *			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#include <ctype.h>
#include <stdio.h>
#include "SNMP-types.h"
#include "objects.h"
#include "tailor.h"

#include "internet.h"
#include "clns.h"

/*    DATA */

#define	MAXSYN	500

static object_syntax syntaxes[MAXSYN + 1];
static OS    synlast = syntaxes;

/*    INTEGER */

static int  integer_encode (x, pe)
integer	*x;
PE     *pe;
{
    if ((*pe = int2prim (*x)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  integer_decode (x, pe)
integer **x;
PE	pe;
{
    integer	i;

    if (pe -> pe_class != PE_CLASS_UNIV || pe -> pe_id != PE_PRIM_INT) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }

    if ((i = prim2num (pe)) == NOTOK && pe -> pe_errno != PE_ERR_NONE) {
	pe -> pe_errno = int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }

    if ((*x = (integer *) malloc (sizeof **x)) == NULL) {
	pe -> pe_errno = int_SNMP_error__status_genErr;
	return NOTOK;
    }
    **x = i;

    return OK;
}


static int  integer_free (x)
integer *x;
{
    free ((char *) x);
}


/* ARGSUSED */

static int  integer_parse (x, s, os)
integer **x;
char   *s;
OS	os;
{
    long    l;
    register struct enum_syntax *es;

    if (!isdigit (*s))
	for (es = os -> os_enum; es; es = es -> es_next)
	    if (lexequ (es -> es_name, s) == 0) {
		l = es -> es_value;
		goto do_assign;
	    }

    if (sscanf (s, "%ld", &l) != 1)
	return NOTOK;

do_assign: ;
    if ((*x = (integer *) malloc (sizeof **x)) == NULL)
	return NOTOK;
    **x = (integer) l;

    return OK;
}


/* ARGSUSED */

static int  integer_print (x, os)
integer *x;
OS	os;
{
    int	    i = *x;
    register struct enum_syntax *es;

    for (es = os -> os_enum; es; es = es -> es_next)
	if (es -> es_value == i) {
	    printf ("%s(%d)", es -> es_name, i);
	    return;
	}

    printf ("%d", *x);
}


static	add_integer ()
{
/* SNMP SMI & SNMPv2 SMI */
    (void) add_syntax ("INTEGER", integer_encode, integer_decode, integer_free,
		       integer_parse, integer_print);

/* SNMPv2 SMI */
    (void) add_syntax ("Integer32", integer_encode, integer_decode,
		       integer_free, integer_parse, integer_print);
}

/*    OCTET STRING */

static int  string_encode (x, pe)
struct qbuf *x;
PE     *pe;
{
    if ((*pe = qb2prim_aux (x, PE_CLASS_UNIV, PE_PRIM_OCTS, 0)) == NULLPE)
	return NOTOK;

    return OK;
}


int	_agentP = 0;

static int  string_decode (x, pe)
struct qbuf **x;
PE	pe;
{
    int	    result;
    char    lastc;
    register char  *cp,
		   *ep;
    register struct qbuf *qb;
    struct qbuf *pb;

    if ((result = string_decode_aux (x, pe)) == NOTOK || !_agentP)
	return result;
    pb = *x;

    lastc = NULL;
    for (qb = pb -> qb_forw; qb != pb; qb = qb -> qb_forw)
	for (ep = (cp = qb -> qb_data) + qb -> qb_len; cp < ep; cp++) {
	    if ((lastc == '\r' && *cp != '\n' && *cp != '\0')
		    || !isascii ((u_char) *cp)) {
not_nvt: ;
		pe -> pe_errno = int_SNMP_error__status_wrongValue;
		return NOTOK;
	    }

	    lastc = *cp;
	}
    if (lastc == '\r')
	goto not_nvt;

    return OK;
}


static int  string_decode_aux (x, pe)
struct qbuf **x;
PE	pe;
{
    struct qbuf *qb;

    if (pe -> pe_class != PE_CLASS_UNIV || pe -> pe_id != PE_PRIM_OCTS) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }

    if (pe -> pe_form != PE_FORM_PRIM) {
	pe -> pe_errno = int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }

    if ((qb = prim2qb (pe)) == NULL) {
	pe -> pe_errno = pe -> pe_errno == PE_ERR_NMEM
				? int_SNMP_error__status_genErr
				: int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }
    *x = qb;

    return OK;
}


/* ARGSUSED */

static int  string_parse (x, s, os)
struct qbuf **x;
char   *s;
OS	os;
{
    struct qbuf *qb;

    if (lexnequ (s, "0x", sizeof "0x" - 1) == 0) {
	int	hex,
		len;
	char   *fmt;
	register u_char *p;

	s += 2;
	len = (strlen (s) >> 1) + 2;	/* for safety... */
	if ((qb = str2qb (NULLCP, len, 1)) == NULL)
	    return NOTOK;
	p = (u_char *) qb -> qb_forw -> qb_data;

	if (index (s, '.'))
	    hex = 0, fmt = "%d";
	else
	    hex = 1, fmt = "%x";
	while (*s) {
	    int	    i;

	    if (sscanf (s, fmt, &i) != 1) {
oops: ;
		qb_free (qb);
		return NOTOK;
	    }
	    *p++ = i & 0xff;

	    if (hex)
		while (isxdigit ((u_char) *s))
		    s++;
	    else
		while (isdigit ((u_char) *s))
		    s++;
	    switch (*s) {
		case '.':
		    if (hex)
			goto oops;
		    s++;
		    break;

		case ':':
		    if (!hex)
			goto oops;
		    s++;
		    break;

		case NULL:
		    break;

		default:
		    goto oops;
	    }
	}

	qb -> qb_len = qb -> qb_forw -> qb_len =
				    p - (u_char *) qb -> qb_forw -> qb_data;
    }
    else
	if ((qb = str2qb (s, strlen (s), 1)) == NULL)
	    return NOTOK;

    *x = qb;

    return OK;    
}


/* ARGSUSED */

static int  string_print (x, os)
struct qbuf *x;
OS	os;
{
    register char *cp,
		  *ep;
    char   *p;
    register struct qbuf *qb;

    p = "0x";
    for (qb = x -> qb_forw; qb != x; qb = qb -> qb_forw)
	for (ep = (cp = qb -> qb_data) + qb -> qb_len; cp < ep; cp++) {
	    printf ("%s%02x", p, *cp & 0xff);
	    p = ":";
	}
}


/* ARGSUSED */

static int  string_display (x, os)
struct qbuf *x;
OS	os;
{
    register struct qbuf *qb;

    printf ("\"");
    for (qb = x -> qb_forw; qb != x; qb = qb -> qb_forw)
	printf ("%*.*s", qb -> qb_len, qb -> qb_len, qb -> qb_data);
    printf ("\"");
}


static	add_string ()
{
/* SNMP SMI */
    (void) add_syntax ("OctetString", string_encode, string_decode_aux,
		       qb_free, string_parse, string_print);

/* SNMPv2 TC & MIB-II */
    (void) add_syntax ("DisplayString", string_encode, string_decode, qb_free,
		       string_parse, string_display);
}

/*    OBJECT IDENTIFIER */

static int  object_encode (x, pe)
OID	x;
PE     *pe;
{
    if ((*pe = oid2prim (x)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  object_decode (x, pe)
OID    *x;
PE	pe;
{
    OID    oid;

    if (pe -> pe_class != PE_CLASS_UNIV || pe -> pe_id != PE_PRIM_OID) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }

    if ((oid  = prim2oid (pe)) == NULLOID) {
	pe -> pe_errno = pe -> pe_errno == PE_ERR_NMEM
				? int_SNMP_error__status_genErr
				: int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }

    if ((*x = oid_cpy (oid)) == NULLOID) {
	pe -> pe_errno = int_SNMP_error__status_genErr;
	return NOTOK;
    }

    return OK;
}


/* ARGSUSED */

static int  object_parse (x, s, os)
OID    *x;
char   *s;
OS	os;
{
    OID	    oid = text2oid (s);

    if (oid == NULL)
	return NOTOK;
    *x = oid;

    return OK;    
}


/* ARGSUSED */

static int  object_print (x, os)
OID	x;
OS	os;
{
    char  *cp,
	   ode[BUFSIZ];

    (void) strcpy (ode, oid2ode (x));
    printf ("%s", ode);
    if (strcmp (ode, cp = sprintoid (x)))
	printf (" (%s)", cp);
}


static	add_object ()
{
/* SNMP SMI & SNMPv2 SMI */
    (void) add_syntax ("ObjectID", object_encode, object_decode, oid_free,
		       object_parse, object_print);
}

/*    BIT STRING */

static int  bits_encode (x, pe)
PE	x;
PE     *pe;
{
    if ((x = bit2prim (x)) == NULLPE || (*pe = pe_cpy (x)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  bits_decode (x, pe)
PE     *x;
PE	pe;
{
    PE	    p;

    if (pe -> pe_class != PE_CLASS_UNIV || pe -> pe_id != PE_PRIM_BITS) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }

    if (pe -> pe_form != PE_FORM_PRIM || (p = prim2bit (pe)) == NULLPE) {
	pe -> pe_errno = int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }

    if ((*x = pe_cpy (p)) == NULLPE) {
	pe -> pe_errno = int_SNMP_error__status_genErr;
	return NOTOK;
    }

    return OK;
}


/* ARGSUSED */

static int  bits_parse (x, s, os)
PE	*x;
char   *s;
OS	os;
{
    int	    i;
    register char *cp;
    PE	    p;

    if (*s == '\'')
	s++;
    for (cp = s; *cp; cp++)
	if (*cp != '1'  && *cp != '0') {
	    if (*cp == '\''
		    && (cp[1] == 'B' || cp[1] == 'b' )
		    && cp[2] == NULL)
		break;
	    return NOTOK;
	}

    if ((p = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_BITS)) == NULL)
	return NOTOK;
    if ((i = cp - s) > 0 && bit_off (p, i - 1) == NOTOK) {
losing: ;
	pe_free (p);
	return NOTOK;
    }
    for (cp = s; isdigit ((u_char) *cp); cp++)
	if (*cp == '1' && bit_on (p, cp - s) == NOTOK)
	    goto losing;

    *x = p;

    return OK;    
}


/* ARGSUSED */

static int  bits_print (x, os)
PE	x;
OS	os;
{
    register int    i;

    printf ("'");
    for (i = 0; i < x -> pe_nbits; i++)
	printf ("%c", bit_test (x, i) == 1 ? '1' : '0');
    printf ("'B");
}


static	add_bits ()
{
/* SNMPv2 SMI */
    (void) add_syntax ("BitString", bits_encode, bits_decode, pe_free,
		       bits_parse, bits_print);
}

/*    NULL */

#ifdef	notdef
/* ARGSUSED */

static int  null_encode (x, pe)
char   *x;
PE     *pe;
{
    if ((*pe = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM, PE_PRIM_NULL)) == NULLPE)
	return NOTOK;

    return OK;
}


/* ARGSUSED */

static int  null_decode (x, pe)
char  **x;
PE	pe;
{
    if (pe -> pe_class != PE_CLASS_UNIV || pe -> pe_id != PE_PRIM_NULL) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }

    if (pe -> pe_form != PE_FORM_PRIM) {
	pe -> pe_errno = int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }

    if ((*x = (char *) calloc (1, sizeof **x)) == NULL) {
	pe -> pe_errno = int_SNMP_error__status_genErr;
	return NOTOK;
    }

    return OK;
}


static int  null_free (x)
char *x;
{
    free ((char *) x);
}


/* ARGSUSED */

static int  null_parse (x, s, os)
char  **x;
char   *s;
OS	os;
{
    if (lexequ (s, "NULL"))
	return NOTOK;

    if ((*x = (char *) calloc (1, sizeof **x)) == NULL)
	return NOTOK;

    return OK;    
}


/* ARGSUSED */

static int  null_print (x, os)
char   *x;
OS	os;
{
    printf ("NULL");
}


static	add_null ()
{
/* SNMP SMI */
    (void) add_syntax ("NULL", null_encode, null_decode, null_free, null_parse,
		       null_print);
}
#endif

/*    IpAddress */

static int  ipaddr_encode (x, pe)
struct sockaddr_in *x;
PE     *pe;
{
    if ((*pe = str2prim ((char *) &x -> sin_addr, 4, PE_CLASS_APPL, 0))
	    == NULLPE)
	return NOTOK;

    return OK;
}


static int  ipaddr_decode (x, pe)
struct sockaddr_in **x;
PE	pe;
{
    struct type_SNMP_IpAddress *ip;
    struct qbuf *qb;
    struct sockaddr_in *isock;

    if (pe -> pe_class != PE_CLASS_APPL || pe -> pe_id != 0) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }
    if (pe -> pe_form != PE_FORM_PRIM) {
	pe -> pe_errno = int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }
    if (pe -> pe_len != 4) {
	pe -> pe_errno = int_SNMP_error__status_wrongLength;
	return NOTOK;
    }

    if ((ip = prim2qb (pe)) == NULL) {
	pe -> pe_errno = pe -> pe_errno == PE_ERR_NMEM
				? int_SNMP_error__status_genErr
				: int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }
    if (qb_pullup (ip) == NOTOK
	    || (isock = (struct sockaddr_in *) calloc (1, sizeof *isock))
		    == NULL) {
	free_SNMP_IpAddress (ip);
	pe -> pe_errno = int_SNMP_error__status_genErr;
	return NOTOK;
    }
    if ((qb = ip -> qb_forw) -> qb_len != 4) {
	free ((char *) isock);
	free_SNMP_IpAddress (ip);
	pe -> pe_errno = int_SNMP_error__status_wrongLength;
	return NOTOK;
    }
    isock -> sin_family = AF_INET;
    bcopy (qb -> qb_data,
	   (char *) &isock -> sin_addr,
	   sizeof isock -> sin_addr);

    *x = isock;

    free_SNMP_IpAddress (ip);
    return OK;
}


static int  ipaddr_free (x)
struct sockaddr_in *x;
{
    free ((char *) x);
}


/* ARGSUSED */

static int  ipaddr_parse (x, s, os)
struct sockaddr_in **x;
char   *s;
OS	os;
{
    register struct hostent *hp = gethostbystring (s);
    register struct sockaddr_in *isock;

    if (hp == NULL)
	return NOTOK;

    if ((isock = (struct sockaddr_in *) calloc (1, sizeof *isock)) == NULL)
	return NOTOK;
    isock -> sin_family = AF_INET;
    inaddr_copy (hp, isock);
    *x = isock;

    return OK;    
}


/* ARGSUSED */

static int  ipaddr_print (x, os)
struct sockaddr_in *x;
OS	os;
{
    printf ("%s", inet_ntoa (x -> sin_addr));
}


static	add_ipaddr ()
{
/* SNMP SMI & SNMPv2 SMI */
    (void) add_syntax ("IpAddress", ipaddr_encode, ipaddr_decode, ipaddr_free,
		       ipaddr_parse, ipaddr_print);
}

/*    NetworkAddress */

/* good enough for now (and probably forever)... */

static	add_netaddr ()
{
/* SNMP SMI */
    (void) add_syntax ("NetworkAddress", ipaddr_encode, ipaddr_decode,
		       ipaddr_free, ipaddr_parse, ipaddr_print);
}

/*    UNSIGNED LONGs */

int	_prim2ulong (pe, c64, n)
register PE	pe;
counter64      *c64;
int	n;
{
    register u_long   i;
    register PElementData dp,
			  ep;

    if (pe -> pe_form != PE_FORM_PRIM || (dp = pe -> pe_prim) == NULLPED)
	return pe_seterr (pe, PE_ERR_PRIM, NOTOK);
    if (pe -> pe_len > n + 1)
	return pe_seterr (pe, PE_ERR_OVER, NOTOK);
    if (pe -> pe_len == n + 1 && (*dp & 0x7f))
	return pe_seterr (pe, PE_ERR_OVER, NOTOK);
    if (*dp & 0x80)
	return pe_seterr (pe, PE_ERR_SIGNED, NOTOK);

    ep = dp + pe -> pe_len;
    if (pe -> pe_len > sizeof i) {
	ep -= sizeof i;

	i = 0L;
	while (dp < ep)
	    i = (i << 8) | (*dp++ & 0xff);
	c64 -> hiword = i;

	ep = dp + sizeof i;
    }
    else
	c64 -> hiword = 0L;
    
    i = 0L;
    while (dp < ep)
	i = (i << 8) | (*dp++ & 0xff);
    c64 -> loword = i;

    return OK;
}


PE	_ulong2prim (c64, class, id)
counter64      *c64;
PElementClass	class;
PElementID	id;
{
    register int    n;
    int	    extend;
    register u_long mask;
    u_long  i;
    register PElementData dp;
    register PE	    pe;

    if ((pe = pe_alloc (class, PE_FORM_PRIM, id)) == NULLPE)
	return NULLPE;

    if ((i = c64 -> hiword) == 0)
	i = c64 -> loword;
    mask = 0xff << (((n = sizeof i) - 1) * 8);
    while (n > 1 && (i & mask) == 0)
	mask >>= 8, n--;
    extend = (i & (0x80 << ((n - 1) * 8))) ? 1 : 0;
    if (c64 -> hiword)
	n += sizeof i;

    if ((pe -> pe_prim = PEDalloc (n + extend)) == NULLPED) {
	pe_free (pe);
	return NULLPE;
    }

    dp = pe -> pe_prim + (pe -> pe_len = n + extend);
    i = c64 -> loword;
    if (c64 -> hiword) {
	register int	m;

	for (m = sizeof i; m-- > 0; i >>= 8)
	    *--dp = i & 0xff;
	i = c64 -> hiword, n -= sizeof i;
    }
    for (; n-- > 0; i >>= 8)
	*--dp = i & 0xff;
    if (extend)
	*--dp = 0x00;

    return pe;
}


u_long	prim2ulong (pe)
register PE	pe;
{
    counter64	c64;

    if (_prim2ulong (pe, &c64, sizeof (u_long)) == NOTOK)
	return 0;

    pe -> pe_errno = PE_ERR_NONE;	/* in case result is ZERO-valued */
    return c64.loword;
}


PE	ulong2prim (i, class, id)
register u_long i;
PElementClass	class;
PElementID	id;
{
    counter64	c64;

    c64.hiword = 0;
    c64.loword = i;

    return _ulong2prim (&c64, class, id);
}

/*    Counter */

static int  counter_encode (x, pe)
u_long	*x;
PE     *pe;
{
    if ((*pe = ulong2prim (*x, PE_CLASS_APPL, 1)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  counter_decode (x, pe)
u_long **x;
PE	pe;
{
    return _counter_decode (x, pe, 1);
}


int	_counter_decode (x, pe, id)
u_long **x;
PE	pe;
int	id;
{
    u_long	i;

    if (pe -> pe_class != PE_CLASS_APPL || pe -> pe_id != id) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }

    if ((i = prim2ulong (pe)) == 0 && pe -> pe_errno != PE_ERR_NONE) {
	pe -> pe_errno = int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }

    if ((*x = (u_long *) malloc (sizeof **x)) == NULL) {
	pe -> pe_errno = int_SNMP_error__status_genErr;
	return NOTOK;
    }
    **x = i;

    return OK;
}


static int  counter_free (x)
u_long *x;
{
    free ((char *) x);
}


/* ARGSUSED */

static int  counter_parse (x, s, os)
u_long **x;
char   *s;
OS	os;
{
    u_long  i;

    if (sscanf (s, "%U", &i) != 1)
	return NOTOK;
    if ((*x = (u_long *) malloc (sizeof **x)) == NULL)
	return NOTOK;
    **x = i;

    return OK;    
}


/* ARGSUSED */

static int  counter_print (x, os)
u_long *x;
OS	os;
{
    printf ("%U", *x);
}


static	add_counter ()
{
/* SNMP SMI */
    (void) add_syntax ("Counter", counter_encode, counter_decode, counter_free,
		       counter_parse, counter_print);

/* SNMPv2 SMI */
    (void) add_syntax ("Counter32", counter_encode, counter_decode,
		       counter_free, counter_parse, counter_print);
}

/*    Counter64 */

static int  counter64_encode (x, pe)
counter64 *x;
PE     *pe;
{
    if ((*pe = _ulong2prim (x, PE_CLASS_APPL, 6)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  counter64_decode (x, pe)
counter64 **x;
PE	pe;
{
    counter64	c64;

    if (pe -> pe_class != PE_CLASS_APPL || pe -> pe_id != 6) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }

    if (_prim2ulong (pe, &c64, 2 * sizeof (u_long)) == NOTOK) {
	pe -> pe_errno = int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }

    if ((*x = (counter64 *) malloc (sizeof **x)) == NULL) {
	pe -> pe_errno = int_SNMP_error__status_genErr;
	return NOTOK;
    }
    **x = c64;

    return OK;
}


static int  counter64_free (x)
counter64 *x;
{
    free ((char *) x);
}


/* ARGSUSED */

static int  counter64_parse (x, s, os)
counter64 **x;
char   *s;
OS	os;
{
    char   *p;
    counter64 c64;
    double    d;

    if (((d = strtod (s, &p)) == 0.0 && s == p) || d < 0.0)
	return NOTOK;
    if (d > 4294967295.0) {
	register int	i;
	double	e = d;

	for (i = sizeof c64.loword * 8; i > 0; i--)
	    d /= 2.0;
	c64.hiword = d;
	for (i = sizeof c64.loword * 8; i > 0; i--)
	    d *= 2.0;
	c64.loword = e - d;	
    }
    else
	c64.hiword = 0L, c64.loword = d;

    if ((*x = (counter64 *) malloc (sizeof **x)) == NULL)
	return NOTOK;
    **x = c64;

    return OK;
}


/* ARGSUSED */

static int  counter64_print (x, os)
counter64 *x;
OS	os;
{
    register int    i;
    int	    j,
	    carry;
    char   *cp,
	   *cpl,
	    last5[6],
    	    string[sizeof "18446744073709551615"];
    double    d;

    if (x -> hiword == 0) {
	printf ("%U", x -> loword);
	return;
    }

    d = x -> hiword;
    for (i = sizeof x -> loword * 8; i > 0; i--)
        d *= 2.0;

    if (x -> hiword <= 0x1fffff) { 
        d += x->loword;
	printf ("%.f", d);
        return;
    }

    d += (x -> loword & 0xfffff000);
    (void) sprintf (string, "%.f", d);
    (void) sprintf (last5, "%5.5d", x -> loword & 0xfff);
    for (i = carry = 0, cp = string + strlen (string) - 1, cpl = last5 + 4;
	     i < 5;
	     i++, cp--, cpl--) {
        j = carry + (*cp - '0') + (*cpl - '0');
        if (j > 9) {
            j -= 10;
            carry = 1;
        }
        else
            carry = 0;
        *cp = j + '0';
    }

    printf ("%s", string);
}


static	add_counter64 ()
{
/* SNMPv2 SMI */
    (void) add_syntax ("Counter64", counter64_encode, counter64_decode,
		       counter64_free, counter64_parse, counter64_print);
}

/*    Gauge */

static int  gauge_encode (x, pe)
u_long	*x;
PE     *pe;
{
    if ((*pe = ulong2prim (*x, PE_CLASS_APPL, 2)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  gauge_decode (x, pe)
u_long **x;
PE	pe;
{
    return _counter_decode (x, pe, 2);
}


static int  uint_encode (x, pe)
u_long	*x;
PE     *pe;
{
    if ((*pe = ulong2prim (*x, PE_CLASS_APPL, 7)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  uint_decode (x, pe)
u_long **x;
PE	pe;
{
    return _counter_decode (x, pe, 7);
}


static	add_gauge ()
{
/* SNMP SMI */
    (void) add_syntax ("Gauge", gauge_encode, gauge_decode, counter_free,
		       counter_parse, counter_print);

/* SNMPv2 SMI */
    (void) add_syntax ("Gauge32", gauge_encode, gauge_decode, counter_free,
		       counter_parse, counter_print);
    (void) add_syntax ("UInteger32", uint_encode, uint_decode, counter_free,
		       counter_parse, counter_print);
}

/*    TimeTicks */

static int  timeticks_encode (x, pe)
u_long	*x;
PE     *pe;
{
    if ((*pe = ulong2prim (*x, PE_CLASS_APPL, 3)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  timeticks_decode (x, pe)
u_long **x;
PE	pe;
{
    return _counter_decode (x, pe, 3);
}


/* ARGSUSED */

static int  timeticks_print (x, os)
u_long *x;
OS	os;
{
    u_long  d,
	    h,
	    m,
	    s,
	    ds;

    ds = *x;
    s = ds / 100, ds = ds % 100;
    m = s / 60, s = s % 60;
    h = m / 60, m = m % 60;
    d = h / 24, h = h % 24;

    if (d > 0)
	printf ("%d day%s, ", d, d != 1 ? "s" : "");
    if (d > 0 || h > 0)
	printf ("%d hour%s, ", h, h != 1 ? "s" : "");
    if (d > 0 || h > 0 || m > 0)
	printf ("%d minute%s, ", m, m != 1 ? "s" : "");
    printf ("%d", s);
    if (ds > 0)
	printf (".%02d", ds);
    printf (" second%s (%U timeticks)", s != 1 || ds > 0 ? "s" : "", *x);
}


static	add_timeticks ()
{
/* SNMP SMI & SNMPv2 SMI */
    (void) add_syntax ("TimeTicks", timeticks_encode, timeticks_decode,
		       counter_free, counter_parse, timeticks_print);
}

/*    NsapAddress */

static int  nsapaddr_encode (x, pe)
struct sockaddr_iso *x;
PE     *pe;
{
    char    buffer[sizeof x -> siso_data + 1];

    buffer[0] = x -> siso_nlen & 0xff;
    bcopy (x -> siso_data, buffer + 1, (int) x -> siso_nlen);

    if ((*pe = str2prim (buffer, (int) (x -> siso_nlen + 1), PE_CLASS_APPL,
			 5)) == NULLPE)
	return NOTOK;

    return OK;
}


static int  nsapaddr_decode (x, pe)
struct sockaddr_iso **x;
PE	pe;
{
    int	    len;
    struct type_SNMP_NsapAddress *nsap;
    struct qbuf *qb;
    struct sockaddr_iso *isock;

    if (pe -> pe_class != PE_CLASS_APPL || pe -> pe_id != 5) {
	pe -> pe_errno = int_SNMP_error__status_wrongType;
	return NOTOK;
    }
    if (pe -> pe_form != PE_FORM_PRIM) {
	pe -> pe_errno = int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }
    if ((len = pe -> pe_len) < 1 || len == 2 || len == 3 || len > 21) {
	pe -> pe_errno = int_SNMP_error__status_wrongLength;
	return NOTOK;
    }

    if ((nsap = prim2qb (pe)) == NULL) {
	pe -> pe_errno = pe -> pe_errno == PE_ERR_NMEM
				? int_SNMP_error__status_genErr
				: int_SNMP_error__status_wrongEncoding;
	return NOTOK;
    }
    if (qb_pullup (nsap) == NOTOK
	    || (isock = (struct sockaddr_iso *) calloc (1, sizeof *isock))
		    == NULL) {
	free_SNMP_NsapAddress (nsap);
	return NOTOK;
    }
    qb = nsap -> qb_forw;
    isock -> siso_family = AF_ISO;
    if ((len = qb -> qb_data[0] & 0xff) >= qb -> qb_len)
	len = qb -> qb_len - 1;
    bcopy (qb -> qb_data + 1, isock -> siso_data,
	   (int) (isock -> siso_nlen = len));

    *x = isock;

    free_SNMP_NsapAddress (nsap);
    return OK;
}


static int  nsapaddr_free (x)
struct sockaddr_iso *x;
{
    free ((char *) x);
}


/* ARGSUSED */

static int  nsapaddr_parse (x, s, os)
struct sockaddr_iso **x;
char   *s;
OS	os;
{
    int	    len;
    register struct sockaddr_iso *isock;

    if ((isock = (struct sockaddr_iso *) calloc (1, sizeof *isock)) == NULL)
	return NOTOK;
    isock -> siso_family = AF_ISO;

    if (lexnequ (s, "0x", sizeof "0x" - 1) == 0) {
	struct qbuf *qb = NULL;

	if (string_parse (&qb, s, NULLOS) == NOTOK
	        || qb -> qb_len > sizeof isock -> siso_data) {
	    if (qb)
		qb_free (qb);
losing: ;
	    free ((char *) isock);
	    return NOTOK;
	}
	
	bcopy (qb -> qb_forw -> qb_data, isock -> siso_data,
	       (int) (isock -> siso_nlen = qb -> qb_len));

	qb_free (qb);
    }
    else {
	if (strncmp (s, "NS+", len = sizeof "NS+" - 1) == 0)
	    s += len;
	if (((len = strlen (s)) & 01) || len > 2 * sizeof isock -> siso_data)
	    goto losing;
	isock -> siso_nlen = implode ((u_char *) isock -> siso_data, s, len);
    }

    *x = isock;

    return OK;    
}


/* ARGSUSED */

static int  nsapaddr_print (x, os)
struct sockaddr_iso *x;
OS	os;
{
#ifdef	notdef
    register char *cp,
		  *ep;
    char   *p;

    p = "";
    for (ep = (cp = x -> siso_data) + x -> siso_nlen; cp < ep; cp++) {
	printf ("%s%02x", p, *cp & 0xff);
	p = ":";
    }
#else
    char    buffer[sizeof x -> siso_data * 2 + 1];

    buffer[explode (buffer, (u_char *) x -> siso_data, (int) x -> siso_nlen)] =
	NULL;
    printf ("NS+%s", buffer);
#endif
}


static	add_nsapaddr ()
{
/* SNMPv2 SMI */
    (void) add_syntax ("NsapAddress", nsapaddr_encode, nsapaddr_decode,
		       nsapaddr_free, nsapaddr_parse, nsapaddr_print);
}

/*    Textual Conventions */

static int  text_print (x, os)
struct qbuf *x;
OS	os;
{
    register char *cp,
		  *dp,
		  *ep;
    register struct qbuf *qb;

    if (qb_pullup (x) == NOTOK) {
	string_print (x, os);
	return;
    }

    dp = (char *) os -> os_display;
    qb = x -> qb_forw;
    ep = (cp = qb -> qb_data) + qb -> qb_len;
    while (*dp && cp < ep) {
	int	octets,
		repeat;
	char	fmt,
		sep1,
		sep2,
	       *ap = dp;

again: ;
	if (*dp == '*')
	    repeat = *cp++ & 0xff, dp++;
	else
	    repeat = 1;

	if (isdigit (*dp)) {
	    octets = 0;
	    do
		octets = octets * 10 + (*dp++ - '0');
	    while (isdigit (*dp));
	}
	else
	    octets = 1;

	switch (fmt = *dp++) {
	    case 'x':
	    case 'd':
	    case 'o':
	    case 'a':
	        break;

	    default:
		fmt = 'x';
		break;
	}

	if (!isdigit (*dp) && *dp != '*') {
	    sep1 = *dp++;
	    if (*dp && !isdigit (*dp) && *dp != '*')
		sep2 = *dp++;
	    else
		sep2 = NULL;
	}
	else
	    sep1 = sep2 = NULL;

	while (repeat-- > 0) {
	    int	    i;

	    if ((i = ep - cp) <= 0)
		return;
	    if (i > octets)
		i = octets;

	    if (fmt == 'a') {
		printf ("%*.*s", i, i, cp);
		cp += i;
	    }
	    else {
		unsigned long l;
		char s[4];

		l = 0;
		while (i-- > 0) {
		    l <<= 8;
		    l |= *cp++ & 0xff;
		}
		(void) sprintf (s, "%%l%c", fmt);
		printf (s, l);
	    }

	    if (cp >= ep)
		return;

	    if (sep1)
		printf ("%c", repeat > 0 || sep2 == NULL ? sep1 : sep2);
	}
	if (*dp == NULL && cp < ep) {
	    dp = ap;
	    goto again;
	}
    }
}


int	add_textual (name, syntax, print)
char   *name,
       *syntax,
       *print;
{
    int	    misc;
    char   *cp;
    register OS	    os;

    if (text2syn (name))
	return DONE;

    if ((os = text2syn (syntax)) == NULL)
	return NOTOK;
    misc = os -> os_misc;

    cp = strdup (name);
    (void) add_syntax (cp, os -> os_encode, os -> os_decode, os -> os_free,
		       os -> os_parse, os -> os_print);
    if (os = text2syn (cp)) {
	os -> os_textual = strdup (syntax);
	if (*print) {
	    os -> os_display = strdup (print);

	    if (os -> os_print == string_print)
		os -> os_print = text_print;
	}

	os -> os_misc = strcmp (cp, "TestAndIncr") ? misc : -1;

	return OK;
    }

    free (cp);
    return NOTOK;
}

/*  */

int	readsyntax ()
{
    static int once_only = 0;

    if (once_only)
	return;
    once_only = 1;

    add_integer ();
    add_string ();
    add_object ();
    add_bits ();
#ifdef	notdef
    add_null ();
#endif
    add_ipaddr ();
    add_netaddr ();
    add_counter ();
    add_counter64 ();
    add_gauge ();
    add_timeticks ();

    add_nsapaddr ();
}

/*  */

int	add_syntax (name, f_encode, f_decode, f_free, f_parse, f_print)
char   *name;
IFP	f_encode,
    	f_decode,
    	f_free,
    	f_parse,
	f_print;
{
    int	    i;
    register OS	    os;

    if ((i = (os = synlast) - syntaxes) >= MAXSYN) {
	if (debug)
	    fprintf (stderr, "ran out of space for syntax %s\n", name);
	return NOTOK;
    }
    synlast++;

    bzero ((char *) os, sizeof *os);
    os -> os_name = name;
    os -> os_encode = f_encode;
    os -> os_decode = f_decode;
    os -> os_free = f_free;
    os -> os_parse = f_parse;
    os -> os_print = f_print;

    return i;
}

/*  */

OS	text2syn (name)
char   *name;
{
    register OS	    os;

    for (os = syntaxes; os < synlast; os++)
	if (strcmp (os -> os_name, name) == 0)
	    return os;

    return NULLOS;
}

/*  */

flsyntax (first, last)
OS     *first,
       *last;
{
    if (first)
	*first = syntaxes;
    if (last)
	*last = synlast - 1;
}

/*  */

int	add_enum (os, name, value)
OS	os;
char   *name;
int	value;
{
    struct enum_syntax *es;

    if ((es = (struct enum_syntax *) calloc (1, sizeof *es)) == NULL
	    || (es -> es_name = strdup (name)) == NULL) {
	if (es)
	    free ((char *) es);
	return NOTOK;
    }
    es -> es_value = value;

    es -> es_next = os -> os_enum, os -> os_enum = es;
    return OK;
}
