#!/usr/local/bin/snmptcl -f

if {![info exists snmptcl_loaded(browsing.tcl)]} {
    if {[file readable browsing.tcl]} {
	source			browsing.tcl
    } else {
	source			$snmptcl_library/browsing.tcl
    }
}

if {[info exists snmptcl_loaded(mib2-browse.tcl)]} {
    return
}
set	snmptcl_loaded(mib2-browse.tcl)	1


#
# turn on "Applications" menu
#

.title.apps	configure	-state normal

.title.apps.menu \
		add cascade	-label "MIB-II browse" \
				-menu [set menu .title.apps.menu.mib2-browse]
menu		$menu


#
# system group
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "System group for $snmptcl_context" system {
{ sysDescr.0			sysDescr 		}
{ sysObjectID.0			sysObjectID 		}
{ sysUpTime.0			sysUpTime 		}
{ sysContact.0			sysContact 		}
{ sysName.0			sysName 		}
{ sysLocation.0			sysLocation 		}
{ sysServices.0			sysServices 		}
{ snmpEnableAuthenTraps.0	snmpEnableAuthenTraps	}
{ snmpV2EnableAuthenTraps.0	snmpV2EnableAuthenTraps	}
{ snmpORLastChange.0		snmpORLastChange	}
{ unixNetstat.0			unixNetstat		}
} 120} \
				-label "System group"

set	snmptcl_print_object(sysServices)	snmptcl_print_sysServices

proc	snmptcl_print_sysServices {var data}	{
    upvar	$data	info

    set		value		[lindex $info 4]
    set	    	privs		""
    foreach	priv		{ {physical 1} {datalink 2} {internet 4}
				  {end-to-end 8} {session 16} {presentation 32}
				  {application 64}} {
	if {[lindex $priv 1] & $value} {
	    lappend	privs	[lindex $priv 0]
	}
    }

    return	"[join $privs {, }] ($value)"
}

set	snmptcl_print_object(snmpORLastChange)	snmptcl_print_TimeStamp


$menu		add cascade	-label "Set System scalars" \
				-menu [set system $menu.system]
menu		$system
$system		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set sysContact for $snmptcl_context" set sysContact} \
				-label sysContact
$system		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set sysName for $snmptcl_context" set sysName} \
				-label sysName
$system		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set sysLocation for $snmptcl_context" set sysLocation} \
				-label sysLocation
$system		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set snmpEnableAuthenTraps for $snmptcl_context" set snmpEnableAuthenTraps} \
				-label snmpEnableAuthenTraps
$system		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set snmpV2EnableAuthenTraps for $snmptcl_context" set snmpV2EnableAuthenTraps} \
				-label snmpV2EnableAuthenTraps
$system		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set snmpSetSerialNo for $snmptcl_context" set snmpSetSerialNo} \
				-label snmpSetSerialNo
$system		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set agentAction for $snmptcl_context" set agentAction coldStart} \
				-label agentAction


#
# SNMPv2 object resources
#

$menu		add command	-command {snmptcl_proc_browsetable $snmptcl_agent 1 "SNMPv2 object resources for $snmptcl_context" snmpv2or {
{ snmpORID			ObjectID		}
{ snmpORDescr			Description		}
} 300 {snmptcl_browsetable_command "#" {}}} \
				-label "SNMPv2 object resources"

$menu		add separator


#
# Interfaces table
#

$menu		add command	-command {snmptcl_proc_browsetable $snmptcl_agent 1 "Interfaces table for $snmptcl_context" iftable {
{ ifDescr			Name			}
{ ifType			Type			}
{ ifMtu				MTU			}
{ ifSpeed			Speed			}
{ ifPhysAddress			"Physical Address"	}
{ ifAdminStatus			Admin			}
{ ifOperStatus			Oper			}
{ ifLastChange			Change			}
{ ifInOctets			IOcts			}
{ ifInUcastPkts			IUPkts			}
{ ifInNUcastPkts		INUPkts			}
{ ifInDiscards			IDiscs			}
{ ifInErrors			IErrs			}
{ ifInUnknownProtos	 	UnkProts		}
{ ifOutOctets			OOcts			}
{ ifOutUcastPkts	 	OUPkts			}
{ ifOutNUcastPkts	 	ONUPkts			}
{ ifOutDiscards			OutDiscs		}
{ ifOutErrors			OErrs			}
{ ifOutQLen			Queue			}
{ ifSpecific			c.f.			}
} 60 {snmptcl_browsetable_command "#" {}}} \
				-label "Interfaces table"

set	snmptcl_print_object(ifSpeed)	snmptcl_print_ifSpeed

proc	snmptcl_print_ifSpeed	{var data} {
    upvar	$data	info

    set		value		[lindex $info 4]
    set		i		0
    if {[string first . $value] >= 0} {
	incr	i
	set	value		[expr int($value/1024)]
    }
    for {} {$value > 999} {set value [expr $value>>10]} {
	incr	i
    }
    switch $i {
	0
	    -
	1
	    -
	2
	    -
	3 {
	    return [format "%s%s" $value [lindex [list "" "K" "M" "G"] $i]]
	}

	default {
	    return "huge"
	}
    }
}

$menu		add cascade	-label "Graph Interface table" \
				-menu [set iftable $menu.iftable]

menu		$iftable
$iftable	add command	-command {snmptcl_proc_browsetable $snmptcl_agent -1 "Unicast Packets for $snmptcl_context" unipkts {
ifInUcastPkts
ifOutUcastPkts
} 10 {snmptcl_graph_command "" ""}} \
				-label "packets/second"

$iftable	add command	-command {snmptcl_proc_browsetable $snmptcl_agent -1 "Unicast Errors for $snmptcl_context" unierrs {
ifInErrors
ifOutErrors
} 10 {snmptcl_graph_command "" ""}} \
				-label "errors/second"



$menu		add separator


#
# IP scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "IP scalars for $snmptcl_context" ip {
{ ipForwarding.0		ipForwarding		}
{ ipDefaultTTL.0		ipDefaultTTL		}
{ ipInReceives.0		ipInReceives		}
{ ipInHdrErrors.0		ipInHdrErrors		}
{ ipInAddrErrors.0		ipInAddrErrors		}
{ ipForwDatagrams.0		ipForwDatagrams		}
{ ipInUnknownProtos.0		ipInUnknownProtos	}
{ ipInDiscards.0		ipInDiscards		}
{ ipInDelivers.0		ipInDelivers		}
{ ipOutRequests.0		ipOutRequests		}
{ ipOutDiscards.0		ipOutDiscards		}
{ ipOutNoRoutes.0		ipOutNoRoutes		}
{ ipReasmTimeout.0		ipReasmTimeout		}
{ ipReasmReqds.0		ipReasmReqds		}
{ ipReasmOKs.0			ipReasmOKs		}
{ ipReasmFails.0		ipReasmFails		}
{ ipFragOKs.0			ipFragOKs		}
{ ipFragFails.0			ipFragFails		}
{ ipFragCreates.0		ipFragCreates		}
{ ipRoutingDiscards.0		ipRoutingDiscards	}
} 60} \
				-label "IP scalars"

$menu		add cascade	-label "Set IP scalars" \
				-menu [set ip $menu.ip]
menu		$ip
$ip		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set ipForwarding for $snmptcl_context" set ipForwarding} \
				-label ipForwarding
$ip		add command	-command {snmptcl_proc_setscalar $snmptcl_agent "set ipDefaultTTL for $snmptcl_context" set ipDefaultTTL} \
				-label ipDefaultTTL


#
# IP mappings
#

$menu		add command	-command {snmptcl_proc_browsetable $snmptcl_agent 1 "IP mappings for $snmptcl_context" ipmap {
{ ipNetToMediaPhysAddress	"Physical Address"	}
{ ipNetToMediaType		Mapping			}
} 60 {snmptcl_browsetable_command "if#   IP Address" snmptcl_print_ipmapping}} \
				-label "IP mappings"

proc	snmptcl_print_ipmapping	{inst}	{
    if {[llength $inst] != 2} {
	return		$inst
    }

    return	[format "%3s %-15s" [lindex $inst 0] [lindex $inst 1]]
}


#
# IP routes
#

$menu		add command	-command {snmptcl_proc_browsetable $snmptcl_agent 1 "IP routes for $snmptcl_context" iproute {
{ ipRouteMask			Mask			}
{ ipRouteIfIndex		"if#"			}
{ ipRouteNextHop		Gateway			}
{ ipRouteType			Type			}
{ ipRouteProto			Proto			}
{ ipRouteAge			Age			}
{ ipRouteMetric1		M1			}
{ ipRouteMetric2		M2			}
{ ipRouteMetric3		M3			}
{ ipRouteMetric4		M4			}
{ ipRouteMetric5		M5			}
{ ipRouteInfo			c.f.			}
{ unixIpRouteFlags		Flags			}
{ unixIpRouteRefCnt		Refcnt			}
{ unixIpRouteUses		Uses			}
} 120 {snmptcl_browsetable_command destination {}}} \
				-label "IP routes"

$menu		add separator


#
# ICMP input scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "ICMP input scalars for $snmptcl_context" icmpin {
{ icmpInMsgs.0			icmpInMsgs		}
{ icmpInErrors.0		icmpInErrors		}
{ icmpInDestUnreachs.0		icmpInDestUnreachs	}
{ icmpInTimeExcds.0		icmpInTimeExcds		}
{ icmpInParmProbs.0		icmpInParmProbs		}
{ icmpInSrcQuenchs.0		icmpInSrcQuenchs	}
{ icmpInRedirects.0		icmpInRedirects		}
{ icmpInEchos.0			icmpInEchos		}
{ icmpInEchoReps.0		icmpInEchoReps		}
{ icmpInTimestamps.0		icmpInTimestamps	}
{ icmpInTimestampReps.0		icmpInTimestampReps	}
{ icmpInAddrMasks.0		icmpInAddrMasks		}
{ icmpInAddrMaskReps.0		icmpInAddrMaskReps	}
} 60} \
				-label "ICMP input scalars"

#
# ICMP output scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "ICMP output scalars for $snmptcl_context" icmpout {
{ icmpOutMsgs.0			icmpOutMsgs		}
{ icmpOutErrors.0		icmpOutErrors		}
{ icmpOutDestUnreachs.0		icmpOutDestUnreachs	}
{ icmpOutTimeExcds.0		icmpOutTimeExcds	}
{ icmpOutParmProbs.0		icmpOutParmProbs	}
{ icmpOutSrcQuenchs.0		icmpOutSrcQuenchs	}
{ icmpOutRedirects.0		icmpOutRedirects	}
{ icmpOutEchos.0		icmpOutEchos		}
{ icmpOutEchoReps.0		icmpOutEchoReps		}
{ icmpOutTimestamps.0		icmpOutTimestamps	}
{ icmpOutTimestampReps.0	icmpOutTimestampReps	}
{ icmpOutAddrMasks.0		icmpOutAddrMasks	}
{ icmpOutAddrMaskReps.0		icmpOutAddrMaskReps	}
} 60} \
				-label "ICMP output scalars"

$menu		add separator


#
# TCP scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "TCP scalars for $snmptcl_context" tcp {
{ tcpRtoAlgorithm.0		tcpRtoAlgorithm		}
{ tcpRtoMin.0			tcpRtoMin		}
{ tcpRtoMax.0			tcpRtoMax		}
{ tcpMaxConn.0			tcpMaxConn		}
{ tcpActiveOpens.0		tcpActiveOpens		}
{ tcpPassiveOpens.0		tcpPassiveOpens		}
{ tcpAttemptFails.0		tcpAttemptFails		}
{ tcpEstabResets.0		tcpEstabResets		}
{ tcpCurrEstab.0		tcpCurrEstab		}
{ tcpInSegs.0			tcpInSegs		}
{ tcpOutSegs.0			tcpOutSegs		}
{ tcpRetransSegs.0		tcpRetransSegs		}
{ tcpInErrs.0			tcpInErrs		}
{ tcpOutRsts.0			tcpOutRsts		}
} 120} \
				-label "TCP scalars"


#
# TCP connections 
#

$menu		add command	-command {snmptcl_proc_browsetable $snmptcl_agent 1 "TCP connections for $snmptcl_context" tcpconn {
{ tcpConnState			State			}
{ unixTcpConnSendQ		SendQ			}
{ unixTcpConnRecvQ		RecvQ			}
{ identStatus			IStatus			}
{ identOpSys			OS			}
{ identCharset			Charset			}
{ identUserid			UserId			}
{ identMisc			Miscellany		}
} 300 {snmptcl_browsetable_command "          local                remote" snmptcl_print_connection}} \
				-label "TCP connections"

proc	snmptcl_print_connection	{inst} {
    switch [llength $inst] {
	2 {
	    return	[format "%15s/%-5s" \
			     [lindex $inst 0] [lindex $inst 1]]
	}

	4 {
	    return	[format "%15s/%-5s %15s/%-5s" \
			     [lindex $inst 0] [lindex $inst 1] \
			     [lindex $inst 2] [lindex $inst 3]]
	}

	default {
	    return	$inst
	}
    }
}

$menu		add separator


#
# UDP scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "UDP scalars for $snmptcl_context" udp {
{ udpInDatagrams.0		udpInDatagrams		}
{ udpNoPorts.0			udpNoPorts		}
{ udpInErrors.0			udpInErrors		}
{ udpOutDatagrams.0		udpOutDatagrams		}
} 120} \
				-label "UDP scalars"


#
# UDP listeners
#

$menu		add command	-command {snmptcl_proc_browsetable $snmptcl_agent 1 "UDP listeners for $snmptcl_context" udplisten {
{ unixUdpRemAddress		remote			}
{ unixUdpRemPort		port			}
{ unixUdpSendQ			SendQ			}
{ unixUdpRecvQ			RecvQ			}
} 300  {snmptcl_browsetable_command "          local" snmptcl_print_connection}} \
				-label "UDP listeners"

$menu		add separator


#
# SNMPv1 input scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "SNMPv1 input scalars for $snmptcl_context" snmpv1in {
{ snmpInPkts.0			snmpInPkts		}
{ snmpInBadVersions.0		snmpInBadVersions	}
{ snmpInBadCommunityNames.0	snmpInBadCommunityNames	}
{ snmpInBadCommunityUses.0	snmpInBadCommunityUses	}
{ snmpInASNParseErrs.0		snmpInASNParseErrs	}
{ snmpInTooBigs.0		snmpInTooBigs		}
{ snmpInNoSuchNames.0		snmpInNoSuchNames	}
{ snmpInBadValues.0		snmpInBadValues		}
{ snmpInReadOnlys.0		snmpInReadOnlys		}
{ snmpInGenErrs.0		snmpInGenErrs		}
{ snmpInTotalReqVars.0		snmpInTotalReqVars	}
{ snmpInTotalSetVars.0		snmpInTotalSetVars	}
{ snmpInGetRequests.0		snmpInGetRequests	}
{ snmpInGetNexts.0		snmpInGetNexts		}
{ snmpInSetRequests.0		snmpInSetRequests	}
{ snmpInGetResponses.0		snmpInGetResponses	}
{ snmpInTraps.0			snmpInTraps		}
} 120} \
				-label "SNMPv1 input scalars"


#
# SNMPv1 output scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "SNMPv1 output scalars for $snmptcl_context" snmpv1out {
{ snmpOutPkts.0			snmpOutPkts		}
{ snmpOutTooBigs.0		snmpOutTooBigs		}
{ snmpOutNoSuchNames.0		snmpOutNoSuchNames	}
{ snmpOutBadValues.0		snmpOutBadValues	}
{ snmpOutGenErrs.0		snmpOutGenErrs		}
{ snmpOutGetRequests.0		snmpOutGetRequests	}
{ snmpOutGetNexts.0		snmpOutGetNexts		}
{ snmpOutSetRequests.0		snmpOutSetRequests	}
{ snmpOutGetResponses.0		snmpOutGetResponses	}
{ snmpOutTraps.0		snmpOutTraps		}
} 120} \
				-label "SNMPv1 output scalars"


#
# SNMPv2 scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "SNMPv2 scalars for $snmptcl_context" snmpv2 {
{ snmpStatsPackets.0		snmpStatsPackets	}
{ snmpStats30Something.0	snmpStats30Something	}
{ snmpStatsEncodingErrors.0	snmpStatsEncodingErrors	}
{ snmpStatsUnknownDstParties.0	snmpStatsUnknownDstParties }
{ snmpStatsDstPartyMismatches.0	snmpStatsDstPartyMismatches }
{ snmpStatsUnknownSrcParties.0	snmpStatsUnknownSrcParties }
{ snmpStatsBadAuths.0		snmpStatsBadAuths	}
{ snmpStatsNotInLifetimes.0	snmpStatsNotInLifetimes	}
{ snmpStatsWrongDigestValues.0	snmpStatsWrongDigestValues }
{ snmpStatsUnknownContexts.0	snmpStatsUnknownContexts }
{ snmpStatsBadOperations.0	snmpStatsBadOperations	}
{ snmpStatsSilentDrops.0	snmpStatsSilentDrops	}
{ snmpV1BadCommunityNames.0	snmpV1BadCommunityNames	}
{ snmpV1BadCommunityUses.0	snmpV1BadCommunityUses	}
} 120} \
				-label "SNMPv2 scalars"

$menu		add cascade	-label "Graph SNMPv2 scalars" \
				-menu [set snmpV2 $menu.snmpV2]

menu		$snmpV2
$snmpV2		add command	-command {snmptcl_proc_browse $snmptcl_agent -1 "SNMPv2 packets/second for $snmptcl_context" v2pps {
snmpStatsPackets.0
} 10 {snmptcl_graph_command "" ""}} \
				-label "packets/second"

$menu		add separator


#
# MBUF scalars
#

$menu		add command	-command {snmptcl_proc_browse $snmptcl_agent 1 "MBUF scalars for $snmptcl_context" mbuf {
{ mbufs.0			mbufs			}
{ mbufNewClusters.0		mbufNewClusters		}
{ mbufNewFreeClusters.0		mbufNewFreeClusters	}
{ mbufDrops.0			mbufDrops		}
{ mbufWaits.0			mbufWaits		}
{ mbufDrains.0			mbufDrains		}
{ mbufFrees.0			mbufFrees		}
{ mbufNewSpaces.0		mbufNewSpaces		}
} 120} \
				-label "MBUF scalars"


#
# MBUF allocations
#

$menu		add command	-command {snmptcl_proc_browsetable $snmptcl_agent 1 "MBUF allocations for $snmptcl_context" mbufs {
{ mbufNewAllocates		Allocated		}
} 120 {snmptcl_browsetable_command Type snmptcl_print_mbuf}} \
				-label "MBUF allocations"

set	snmptcl_mbuf_type(0)	    "free list"
set	snmptcl_mbuf_type(1)	    "dynamic (data) allocation"
set	snmptcl_mbuf_type(2)	    "packet headers"
set	snmptcl_mbuf_type(3)	    "socket structures"
set	snmptcl_mbuf_type(4)	    "protocol control blocks"
set	snmptcl_mbuf_type(5)	    "routing tables"
set	snmptcl_mbuf_type(6)	    "IMP host tables"
set	snmptcl_mbuf_type(7)	    "address resolution tables"
set	snmptcl_mbuf_type(8)	    "socket names"
set	snmptcl_mbuf_type(9)	    "zombine proc status"
set	snmptcl_mbuf_type(10)	    "socket options"
set	snmptcl_mbuf_type(11)	    "fragmentation ressembly headers"
set	snmptcl_mbuf_type(12)	    "access rights"
set	snmptcl_mbuf_type(13)	    "interface addresses"

proc	snmptcl_print_mbuf	{inst} 	{
    global	snmptcl_mbuf_type

    if {[info exists snmptcl_mbuf_type($inst)]} {
	return		$snmptcl_mbuf_type($inst)
    }
    return	$inst
}


