#!/usr/local/bin/snmptcl -f

if {![info exists snmptcl_loaded]} {
    if {[file readable $snmptcl_myname.tcl]} {
	source			$snmptcl_myname.tcl
    } else {
	source			$snmptcl_library/$snmptcl_myname.tcl
    }
}

if {[info exists snmptcl_loaded(polling.tcl)]} {
    return
}
set	snmptcl_loaded(polling.tcl)	1


set	snmptcl_poll_cookie	0


#
# poll instances
#
#     pHandle - context procedure
#	    w - window to create
#       title - title for window
#       iname - name for icon
#     ibitmap - bitmap for icon
#        objs - list of instances to poll, each element may be a list,
#		if so, first is descriptor/OID and second is label
#    interval - polling interval (in seconds)
#     command - command to call on each iteration
#		(see snmptcl_poll_callback for calling procedure)
#

proc	snmptcl_proc_poll \
	{pHandle w title iname ibitmap objs interval command} {
    global	snmptcl_info snmptcl_var_sysUpTime

    catch	{ destroy $w }
    toplevel	$w

    wm title	$w		$title
    wm iconname	$w		$iname
    wm iconbitmap \
		$w		$ibitmap

    frame	$w.frame	-borderwidth 2 \
				-relief sunken

    set		vars		""
    set		odescr		""
    foreach	obj		$objs {
	if {[llength $obj] > 1} {
	    set		ovar	[lindex $obj 0]
	    set		obj	[lindex $obj 1]
	} else {
	    set		ovar	$obj
	}
	if {[catch { set full [snmpinfo mibprop $ovar full] } result]} {
	    destroy	$w
	    snmptcl_dialog	$title $result questhead 0 Abort
	    return
	}
	lappend		odescr	$obj
	set		ovar	[lindex $full 2]
	if {[set inst [lindex $full 3]] != ""} {
	    append	ovar	.$inst
	}
	lappend	vars		$ovar
    }

    set		array		snmptcl_poll_data[join [split $w "."] "_"]

    global	${array}_info
    upvar #0	${array}_info	ainfo
    set		ainfo(toplevel)	$w
    set		ainfo(frame)	$w.frame
    set		ainfo(title)	$title
    set		ainfo(objs)	$odescr
    set		ainfo(vars)	$vars
    set		ainfo(interval)	$interval
    set		ainfo(iter)	0
    set		ainfo(segs)	0
    set		ainfo(startime)	""
    set		ainfo(lastime)	""
    set		ainfo(command)	$command

    frame	$w.bot		-borderwidth 1
    pack append	$w.bot		[label $w.bot.elapsed \
					-width 60] \
				    {top padx 5 pady 5 fillx} \
				[scale $w.bot.polling \
					-command "snmptcl_poll_reschedule ${array}" \
					-from 10 -to 900 \
					-label "Polling In Progress..." \
					-length 200 \
					-orient horizontal] \
				    {left padx 5 pady 5 expand} \
				[button $w.bot.done \
					-command "snmptcl_poll_destroy $w $array" \
					-text Done] \
				    {left padx 5 pady 5 expand}
    $w.bot.polling \
		set		$interval
    $w.bot.polling \
		configure	-state disabled

    pack append	$w		$w.frame \
				    {top expand fillx filly} \
				$w.bot \
				    {bottom fill}

    set		vars		[concat $snmptcl_var_sysUpTime.0 $vars]
    set		iHandle		[$pHandle get snmptcl_poll_callback $vars]
    incr	ainfo(segs)
    set		snmptcl_info($iHandle) \
				[list $array [lrange $vars 1 end]]
}

#
# callback when polling instances
#     snmptcl_info($iHandle) contains
#	  0: name of global array containing additional information
#	  1: list of instances we asked for
#
#     array_info(*) contains
#  toplevel: top-level window
#     frame: frame window
#     title: top-level window title
#      objs: list of labels for instances
#      vars: list of OIDs for instances
#  interval: polling interval (in seconds)
#      iter: iteration number
#      segs: number of partial sweeps for this iteration
#  startime: sysUpTime of initial retrieval
#   lastime: sysUpTime of previous retrieval
#   command: command to invoke
#
#     array_var($var) contains
#	  0: current iteration:  iteration number
#	  1: previous iteration: value
#	  2:			 sysUpTime
#	  3: initial iteration:  value
#	  4: current iteration:  value
#	  5:			 sysUpTime
#	  6:			 syntax
#

proc	snmptcl_poll_callback \
	{pHandle iHandle eStatus eIndex vBinds} {
    global	snmptcl_info snmptcl_var_sysUpTime

    set		data		$snmptcl_info($iHandle)
    unset	snmptcl_info($iHandle)
    set		array		[lindex $data 0]
    set		lvars		[lindex $data 1]

    global	${array}_info
    upvar #0	${array}_info	ainfo
    if {![info exists ainfo(toplevel)]} {
	snmptcl_poll_destroy	"" $array
	return
    }
    set		w		$ainfo(toplevel)
    set		iter		$ainfo(iter)
    incr	ainfo(segs)	-1

    global	${array}_var
    upvar #0	${array}_var	avar

    if {($eStatus == "timeout") && ([llength $objs] == [llength $lvars])} {
	set	eStatus		tooBig
    }
    switch -- $eStatus {
	noSuchName {
	    if {$eIndex > 1} {
		incr	eIndex	-2
		set	vars	    [concat $snmptcl_var_sysUpTime.0 \
				         [lreplace $lvars $eIndex $eIndex]]
		set	iHandle	    [$pHandle get snmptcl_poll_callback $vars]
		incr	ainfo(segs)
		set	snmptcl_info($iHandle) \
				    [list $array [lrange $vars 1 end]]
		return
	    }
	}

	tooBig {
	    if {[set eIndex [llength $lvars]] > 1} {
		set	hIndex	    [expr $eIndex/2]

		set	vars	    [concat $snmptcl_var_sysUpTime.0 \
					[lrange $lvars 0 [expr $hIndex-1]]]
		set	iHandle	    [$pHandle get snmptcl_poll_callback $vars]
		incr	ainfo(segs)
		set	snmptcl_info($iHandle) \
				    [list $array [lrange $vars 1 end]]

		set	vars	    [concat $snmptcl_var_sysUpTime.0 \
					[lrange $lvars $hIndex end]]
		set	iHandle	    [$pHandle get snmptcl_poll_callback $vars]
		incr	ainfo(segs)
		set	snmptcl_info($iHandle) \
				    [list $array [lrange $vars 1 end]]
		return
	    }
	}
    }
	
    switch -- $eStatus {
	noError {
	}

	destroyed {
	    snmptcl_poll_destroy \
				$w.bot.polling $array
	    destroy		$w.bot.elapsed
	    return   
	}

	default {
	    switch [snmptcl_dialog "Agent Error for $ainfo(title)" \
		        $eStatus question 0 Retry Orphan Abort] {
		0 {
		    set	    vars	[concat $snmptcl_var_sysUpTime.0 \
					     $lvars]
		    set	    iHandle	[$pHandle get snmptcl_poll_callback \
					     $vars]
		    incr    ainfo(segs)
		    set	    snmptcl_info($iHandle) \
					[list $array [lrange $vars 1 end]]
		    return
		}

		1 {
		    snmptcl_poll_destroy \
					$w.bot.polling $array
		    destroy		$w.bot.elapsed
		    return   
		}
	    }
	}
    }
    if {$eStatus != "noError"} {
	snmptcl_poll_destroy	$w $array
	return
    }

    set		vb	[lindex $vBinds 0]
    if {"$snmptcl_var_sysUpTime.0" != [lindex $vb 0]} {
	snmptcl_dialog	"Agent Error for $ainfo(title)" \
			"instead of sysUpTime.0, got [snmptcl_proc_oid2ode [lindex $vb 0]]" \
				warning 0 Abort
	snmptcl_poll_destroy \
			$w $array
	return
    }
    if {([set syntax [lindex $vb 1]] == "exception") \
	    || ($syntax == "error")} {
	set	sysUpTime	4294967295.0
	set	ainfo(startime)	""
	set	ainfo(lastime)	""
	$w.bot.elapsed		configure -text "sysUpTime not available, so no information on counters" -anchor c
    } else {
	set	sysUpTime	[lindex $vb 2]
	if {[set startime $ainfo(startime)] == ""} {
	    set	    ainfo(startime)	$sysUpTime
	    $w.bot.elapsed	configure -text "just started, so information on counters is unavailable" -anchor c
	} elseif {$ainfo(lastime) > $sysUpTime} {
	    set	    ainfo(startime)	$sysUpTime
	    $w.bot.elapsed	configure -text "agent re-started, so information on counters is unavailable" -anchor c
	} elseif {$iter} {
	    $w.bot.elapsed	configure -text "Elapsed time: [snmptcl_print_TimeTicks [expr $sysUpTime-$startime]]" -anchor w
	}
	set	ainfo(lastime)	$sysUpTime
    }

    set		ivars	0
    set		nvars	[llength $lvars]

    foreach	vb	[lrange $vBinds 1 end] {
	set	var	[lindex $vb 0]

	if {$ivars >= $nvars} {
	    snmptcl_dialog	"Agent error for $ainfo(title)" \
				"retrieval botched, got back too many variables" \
				warning 0 Abort
	    snmptcl_poll_destroy \
				$w $array
	    return
	}

	set	lvar	[lindex $lvars $ivars]
	incr	ivars

	if {([set syntax [lindex $vb 1]] == "exception") \
		|| ($syntax == "error")} {
	    continue
	}

	if {[snmptcl_proc_oidcmp $lvar $var] != 0} {
	    snmptcl_dialog	"Agent Error for $ainfo(title)" \
				"retrieval botched, expecting [snmptcl_proc_oid2ode $lvar], got [snmptcl_proc_oid2ode $var]" \
				warning 0 Abort
	    snmptcl_poll_destroy \
				$w $array
	    return	    
	}

	set	value	[lindex $vb 2]
	if {![info exists avar($var)]} {
	    set		avar($var)	[list $iter "" 0 $value \
					     $value $sysUpTime $syntax]
	} elseif {[lindex [set data $avar($var)] 2] > $sysUpTime} {
	    set		avar($var)	[lreplace $data 0 6 $iter "" 0 $value \
					     $value $sysUpTime $syntax]
	} else {
	    set		avar($var)	[lreplace $data 0 5 $iter \
					     [lindex $data 4] [lindex $data 5] \
					     [lindex $data 3] $value $sysUpTime]
	}
    }
    if {$ivars < $nvars} {
	snmptcl_dialog		"Agent error for $ainfo(title)" \
				"retrieval botched, got back too few variables" \
				warning 0 Abort
	snmptcl_poll_destroy	$w $array
	return
    }

    if {$ainfo(segs)} {
	return
    }

    if {[info exists avar]} {
	foreach element [array names avar] {
	    if {[lindex $avar($element) 0] != $iter} {
		unset	avar($element)
	    }
	}
    }

    eval	$ainfo(command)		[list $w $array]

    incr	ainfo(iter)

    set		vars		[concat $snmptcl_var_sysUpTime.0 $lvars]

    $w.bot.polling \
		configure		-label "Polling Interval (seconds)" \
					-state normal
    set		ainfo(nextpoll)	[after [expr $ainfo(interval)*1000] \
				     snmptcl_poll_callback_aux \
				         [list $pHandle $vars $array]]
    update	idletasks
}

proc	snmptcl_poll_callback_aux {pHandle vars array} {
    global	snmptcl_info

    global	${array}_info
    upvar #0	${array}_info	ainfo
    if {![info exists ainfo(toplevel)]} {
	snmptcl_poll_destroy	"" $array
	return
    }
    if {[catch { $pHandle config active }]} {
	destroy			$ainfo(toplevel).bot.elapsed
	snmptcl_poll_destroy	$ainfo(toplevel).bot.polling $array
	return
    }

    $ainfo(toplevel).bot.polling \
		configure	-label "Polling In Progress..." \
				-state disabled

    set		iHandle		[$pHandle get snmptcl_poll_callback $vars]
    incr	ainfo(segs)
    set		snmptcl_info($iHandle) \
				[list $array [lrange $vars 1 end]]
}

proc	snmptcl_poll_reschedule {array value} {
    global	${array}_info
    upvar #0	${array}_info	ainfo

    set		ainfo(interval)	$value
    if {![info exists ainfo(nextpoll)]} {
	return
    }

    set		command		[cancel $ainfo(nextpoll)]
    set		ainfo(nextpoll)	[after [expr $ainfo(interval)*1000] $command]
}

proc	snmptcl_poll_destroy	{w array}	{
    global	${array}_info ${array}_var

    if {$w != ""} {
	catch { destroy $w }
    }
    catch { unset ${array}_info }
    catch { unset ${array}_var }
}


#
# poll a table
#
#     pHandle - context procedure
#	    w - window to create
#	title - title for window
#	iname - name for icon
#     ibitmap - bitmap for icon
#	 cols - list of columnars to poll, each element may be a list,
#		if so, first is descriptor/OID and second is label
#    interval - polling interval (in seconds)
#     command - command to call on each iteration
#		(see snmptcl_polltable_callback for calling procedure)
#

proc	snmptcl_proc_polltable \
	{pHandle w title iname ibitmap cols interval command} {
    global	snmptcl_info snmptcl_var_sysUpTime

    catch	{ destroy $w }
    toplevel	$w

    wm title	$w		$title
    wm iconname	$w		$iname
    wm iconbitmap \
		$w		$ibitmap

    frame	$w.frame	-borderwidth 2 \
				-relief sunken

    set		cvars		""
    set		cdescr		""
    foreach	col		$cols {
	if {[set nelem [llength $col]] > 1} {
	    set		cvar	[lindex $col 0]
	    set		col	[lindex $col 1]
	} else {
	    set		cvar	$col 
	}
	if {[catch { set full [snmpinfo mibprop $cvar full] } result]} {
	    destroy	$w
	    snmptcl_dialog	$title $result questhead 0 Abort
	    return
	}
	lappend		cdescr	$col
	set		cvar	[lindex $full 2]
	if {[set inst [lindex $full 3]] != ""} {
	    append	cvar	.$inst
	}
	lappend	cvars		$cvar
    }

    set		array		snmptcl_polltable_data[join [split $w "."] "_"]

    global	${array}_info
    upvar #0	${array}_info	ainfo
    set		ainfo(toplevel)	$w
    set		ainfo(frame)	$w.frame
    set		ainfo(title)	$title
    set		ainfo(cols)	$cdescr
    set		ainfo(cvars)	$cvars
    set		ainfo(interval)	$interval
    set		ainfo(iter)	0
    set		ainfo(segs)	0
    set		ainfo(startime)	""
    set		ainfo(lastime)	""
    set		ainfo(command)	$command

    global	${array}_ncol
    upvar #0	${array}_ncol	ancol
    set		ancol("")	""

    frame	$w.bot		-borderwidth 1
    pack append	$w.bot		[label $w.bot.elapsed \
					-width 60] \
				    {top padx 5 pady 5 fillx} \
				[scale $w.bot.polling \
					-command "snmptcl_poll_reschedule ${array}" \
					-from 10 -to 900 \
					-label "Polling In Progress..." \
					-length 200 \
					-orient horizontal] \
				    {left padx 5 pady 5 expand} \
				[button $w.bot.done \
					-command "snmptcl_polltable_destroy $w $array" \
					-text Done] \
				    {left padx 5 pady 5 expand}
    $w.bot.polling \
		set		$interval
    $w.bot.polling \
		configure	-state disabled

    pack append	$w		$w.frame \
				    {top expand filly} \
				$w.bot \
				    {bottom fill}

    set		vars		[concat $snmptcl_var_sysUpTime $cvars]
    set		maxrept		10
    set		iHandle		[$pHandle bulk snmptcl_polltable_callback \
				     $vars -maxrept $maxrept -nonrptr 1]
    incr	ainfo(segs)
    set		snmptcl_info($iHandle) \
				[list $array [lrange $vars 1 end] \
				     [lrange $vars 1 end] $maxrept]
}


#
# callback when polling a table
#     snmptcl_info($iHandle) contains
#	  0: name of global array containing additional information
#	  1: list of repeaters we asked for
#	  2: OID prefixes of corresponding columns
#	  3: max-repetitions
#
#     array_info(*) contains
#  toplevel: top-level window
#     frame: frame window
#     title: top-level window title
#      cols: list of column labels
#     cvars: list of corresponding OIDs (possibly with subtree)
#  interval: polling interval (in seconds)
#      iter: iteration number
#      segs: number of partial sweeps for this iteration
#  startime: sysUpTime of initial retrieval
#   lastime: sysUpTime of previous retrieval
#   command: command to invoke
#
#     array_var($var) contains
#	  0: current iteration:  iteration number
#	  1: previous iteration: value
#	  2:			 sysUpTime
#	  3: initial iteration:  value
#	  4: current iteration:  value
#	  5:			 sysUpTime
#	  6:			 syntax
#	  7:			 offset in cvars
#
#	array_ncol($row) contains number of existing cells in $row
#

proc	snmptcl_polltable_callback \
	{pHandle iHandle eStatus eIndex vBinds} {
    global	snmptcl_info snmptcl_var_sysUpTime

    set		data		$snmptcl_info($iHandle)
    unset	snmptcl_info($iHandle)
    set		array		[lindex $data 0]
    set		lvars		[lindex $data 1]
    set		xvars		[lindex $data 2]
    set		maxrept		[lindex $data 3]

    global	${array}_info
    upvar #0	${array}_info	ainfo
    if {![info exists ainfo(toplevel)]} {
	snmptcl_polltable_destroy \
				"" $array
	return
    }
    set		w		$ainfo(toplevel)
    set		iter		$ainfo(iter)
    set		cvars		$ainfo(cvars)
    incr	ainfo(segs)	-1

    global	${array}_var
    upvar #0	${array}_var	avar

    global	${array}_ncol
    upvar #0	${array}_ncol	ancol

    if {($eStatus == "timeout") && ([llength $cvars] == [llength $lvars])} {
	set	eStatus		tooBig
    }
    switch -- $eStatus {
	noSuchName {
	    if {$eIndex > 1} {
		incr	eIndex	-2
		set	vars	    [concat $snmptcl_var_sysUpTime \
				         [lreplace $lvars $eIndex $eIndex]]
		set	iHandle	    [$pHandle next snmptcl_polltable_callback \
					 $vars]
		incr	ainfo(segs)
		set	snmptcl_info($iHandle) \
				    [list $array [lrange $vars 1 end] \
				         [lreplace $xvars $eIndex $eIndex] \
					 $maxrept]
		return
	    }
	}

	tooBig {
	    if {[set eIndex [llength $lvars]] > 1} {
		set	hIndex	    [expr $eIndex/2]

		set	vars	    [concat $snmptcl_var_sysUpTime \
					[lrange $lvars 0 [expr $hIndex-1]]]
		set	iHandle	    [$pHandle next snmptcl_polltable_callback \
					 $vars]
		incr	ainfo(segs)
		set	snmptcl_info($iHandle) \
				    [list $array [lrange $vars 1 end] \
				         [lrange $xvars 0 [expr $hIndex-1]] \
					 $maxrept]

		set	vars	    [concat $snmptcl_var_sysUpTime \
					[lrange $lvars $hIndex end]]
		set	iHandle	    [$pHandle next snmptcl_polltable_callback \
					 $vars]
		incr	ainfo(segs)
		set	snmptcl_info($iHandle) \
				    [list $array [lrange $vars 1 end] \
					 [lrange $xvars $hIndex end] \
				         $maxrept]
		return
	    }
	}
    }
	
    switch -- $eStatus {
	noError {
	}

	destroyed {
	    snmptcl_poll_destroy \
				$w.bot.polling $array
	    destroy		$w.bot.elapsed
	    return   
	}

	default {
	    switch [snmptcl_dialog "Agent Error for $ainfo(title)" \
		        $eStatus question 0 Retry Orphan Abort] {
		0 {
		    set	    vars	[concat $snmptcl_var_sysUpTime $lvars]
		    set	    iHandle	[$pHandle bulk \
					     snmptcl_polltable_callback \
					     $vars -maxrept $maxrept \
					     -nonrptr 1]
		    incr    ainfo(segs)
		    set	    snmptcl_info($iHandle) \
					[list $array [lrange $vars 1 end] \
					     $xvars $maxrept]
		    return
		}

		1 {
		    snmptcl_polltable_destroy \
					$w.bot.polling $array
		    destroy		$w.bot.elapsed
		    return   
		}
	    }
	}
    }
    if {$eStatus != "noError"} {
	snmptcl_polltable_destroy	$w $array
	return
    }

    set		vb	[lindex $vBinds 0]
    if {"$snmptcl_var_sysUpTime.0" != [lindex $vb 0]} {
	snmptcl_dialog		"Agent Error for $ainfo(title)" \
				"instead of sysUpTime, got [snmptcl_proc_oid2ode [lindex $vb 0]]" \
				warning 0 Abort
	snmptcl_polltable_destroy \
			$w $array
	return
    }
    if {([set syntax [lindex $vb 1]] == "exception") \
	    || ($syntax == "error")} {
	set	sysUpTime	4294967295.0
	set	ainfo(startime)	""
	set	ainfo(lastime)	""
	$w.bot.elapsed		configure -text "sysUpTime not available, so no information on counters" -anchor c
    } else {
	set	sysUpTime	[lindex $vb 2]
	if {[set startime $ainfo(startime)] == ""} {
	    set	    ainfo(startime)	$sysUpTime
	    $w.bot.elapsed	configure -text "just started, so information on counters is unavailable" -anchor c
	} elseif {$ainfo(lastime) > $sysUpTime} {
	    set	    ainfo(startime)	$sysUpTime
	    $w.bot.elapsed	configure -text "agent re-started, so information on counters is unavailable" -anchor c
	} elseif {$iter} {
	    $w.bot.elapsed	configure -text "Elapsed time: [snmptcl_print_TimeTicks [expr $sysUpTime-$startime]]" -anchor w
	}
	set	ainfo(lastime)	$sysUpTime
    }

    set		ivars	0
    set		nvars	[llength [set mvars $lvars]]
    set		evars	""

    foreach	vb	[lrange $vBinds 1 end] {
	set	var	[lindex $vb 0]

	set	mvar	[lindex $mvars [set icol $ivars]]
	set	mvars	[lreplace $mvars $ivars $ivars $var]
	if {[incr ivars] >= $nvars} {
	    set		ivars	0
	}

	if {([set syntax [lindex $vb 1]] == "exception") \
		|| ($syntax == "error")} {
	    lappend	evars	$var
	    continue
	}

	if {[snmptcl_proc_oidcmp $mvar $var] >= 0} {
	    snmptcl_dialog	"Agent Error for $ainfo(title)" \
				"lexicographic ordering botched, instead of something >= [snmptcl_proc_oid2ode $mvar], got [snmptcl_proc_oid2ode $var]" \
				warning 0 Abort
	    snmptcl_polltable_destroy \
				$w $array
	    return	    
	}

	if {![snmptcl_proc_oidcontains [set xvar [lindex $xvars $icol]] $var]} {
	    continue
	}

	set	icol	0
	foreach	cvar	$cvars {
	    if {$xvar == $cvar} {
		break
	    }
	    incr	icol
	}

	set	value	[lindex $vb 2]
	if {![info exists avar($var)]} {
	    set		avar($var)	[list $iter "" 0 $value \
					     $value $sysUpTime $syntax $icol]
	} elseif {[lindex [set data $avar($var)] 2] > $sysUpTime} {
	    set		avar($var)	[lreplace $data 0 7 $iter "" 0 $value \
					     $value $sysUpTime $syntax $icol]
	} elseif {[lindex $data 0] == $iter} {
	    continue
	} else {
	    set		avar($var)	[lreplace $data 0 5 $iter \
					     [lindex $data 4] [lindex $data 5] \
					     [lindex $data 3] $value $sysUpTime]
	}
    }

    set		vars	$snmptcl_var_sysUpTime
    set		ivars	0
    set		yvars	""
    foreach	lvar	$lvars {
	set	mvar	[lindex $mvars [set icol $ivars]]
	incr	ivars

	if {([lsearch -exact $evars $mvar] >= 0) \
		|| ![snmptcl_proc_oidcontains [set xvar [lindex $xvars $icol]] \
		         $mvar]} {
	    continue
	}

	lappend	vars	$mvar
	lappend	yvars	$xvar
    }

    if {[llength $vars] > 1} {
	set	maxrept		[expr [array size ancol]]
	set	iHandle		[$pHandle bulk snmptcl_polltable_callback \
				     $vars -maxrept $maxrept -nonrptr 1]
	incr    ainfo(segs)
	set	snmptcl_info($iHandle)	\
				[list $array [lrange $vars 1 end] $yvars \
				     $maxrept]
	return
    }

    if {$ainfo(segs)} {
	return
    }

    eval	$ainfo(command)		[list $w $array]

    incr	ainfo(iter)

    set		vars		[concat $snmptcl_var_sysUpTime $cvars]
    if {[info exists ancol]} {
	set	maxrept		[array size ancol]
    } else {
	set	maxrept		1
    }

    $w.bot.polling \
		configure	-label "Polling Interval (seconds)" \
				-state normal
    set		ainfo(nextpoll)	[after [expr $ainfo(interval)*1000] \
				     snmptcl_polltable_callback_aux \
				         [list $pHandle $vars $maxrept $array]]
    update	idletasks
}

proc	snmptcl_polltable_callback_aux {pHandle vars maxrept array} {
    global	snmptcl_info

    global	${array}_info
    upvar #0	${array}_info	ainfo
    if {![info exists ainfo(toplevel)]} {
	snmptcl_polltable_destroy \
				"" $array
	return
    }
    if {[catch { $pHandle config active }]} {
	destroy			$ainfo(toplevel).bot.elapsed
	snmptcl_poll_destroy	$ainfo(toplevel).bot.polling $array
	return
    }

    $ainfo(toplevel).bot.polling \
		configure	-label "Polling In Progress..." \
				-state disabled

    set		iHandle		[$pHandle bulk snmptcl_polltable_callback \
				     $vars -maxrept $maxrept -nonrptr 1]
    incr	ainfo(segs)
    set		snmptcl_info($iHandle) \
				[list $array [lrange $vars 1 end] \
				     [lrange $vars 1 end] $maxrept]
}

proc	snmptcl_polltable_destroy	{w array}	{
    global	${array}_info ${array}_var ${array}_ncol

    if {$w != ""} {
	catch { destroy $w }
    }
    catch { unset ${array}_info }
    catch { unset ${array}_var }
    catch { unset ${array}_ncol }
}
