#! /usr/local/bin/snmptcl -f

if {[info exists snmptcl_loaded]} {
    return
}
set	snmptcl_loaded($snmptcl_myname.tcl)	1

if {![file readable [set snmptcl_icon snmptcl.icon]]} {
    set	snmptcl_icon	$snmptcl_library/snmptcl.icon
}

set	snmptcl_help(help)	\
"A management application is selected using the \"Applications\" menu.\n\nIf you want to tear-off this menu, click-middle-hold over the \"Applications\" button, drag the menu to wherever you want, and then release-middle.  To release the menu, click-left-hold on the \"Applications\" button, move the mouse off of the button, and then release-left.\n\nA management context is selected or destroyed by using the \"Contexts\" menu."


wm title \
	.		$snmptcl_myname
wm iconbitmap \
	.		@$snmptcl_icon

frame	.title		-borderwidth 2 \
			-relief raised
pack append \
	.title		[menubutton .title.apps \
				-menu .title.apps.menu \
				-relief raised \
				-state disabled \
				-text Applications] \
			    {left} \
			[button .title.quit \
				-command exit \
				-text Quit] \
			    {right} \
			[button .title.help \
				-command "snmptcl_callback_help .help \"Help with $snmptcl_myname\" help help" \
				-text Help] \
			    {right} \
			[button .title.options \
				-command snmptcl_callback_options \
				-text Options] \
			    {right} \
			[menubutton .title.contexts \
				-menu .title.contexts.menu \
				-relief raised \
				-text Contexts] \
			    {right}

menu	.title.apps.menu

menu	.title.contexts.menu
.title.contexts.menu \
	add command	-command snmptcl_callback_select_contexts \
			-label Select...
.title.contexts.menu \
	add cascade	-label Destroy \
			-menu .title.contexts.menu.destroy
menu	.title.contexts.menu.destroy \
			-post snmptcl_callback_destroy_contexts

frame	.subtitle	-borderwidth 2

pack append \
	.subtitle	[label .subtitle.msg \
				-text ""] \
			    {top expand}


pack append \
	.		.title \
			    {top pady 5 fill} \
			.subtitle \
			    {top fill}


#
# Callback: Select a Context
#

set	snmptcl_agent	agent
set	snmptcl_agent_	agent
set	snmptcl_agent_cookie \
			0
set	snmptcl_context	""
set	snmptcl_rtrqos	exponential
set	snmptcl_retries	5
set	snmptcl_timeout	3
set	snmptcl_secqos	none

set	snmptcl_help(select)	\
"Enter the name of a management context.  If the name is an IP-address or domain-name, then $snmptcl_myname will use SNMPv2 without authentication or privacy to access the management context.  If the name is of the form:\n\t<IP-address>\"/\"<community>\nor\n\t<domain-name>\"/\"<community>\nthen, $snmptcl_myname will use SNMPv1.\n\nFor example, both\n\t192.103.140.1\nand\n\tdbc.mtview.ca.us\nselect a management context accessible by SNMPv2, whilst both\n\t192.103.140.1/read/write\nand\n\tdbc.mtview.ca.us/read/write\nselect a management context accessible by SNMPv1 with the community \"read/write\".\n\nThe \"Security QOS\" buttons determine the minimal security requirements for an operation.\n\nFor retrieval operations, the \"Retransmission QOS\" buttons determine the retransmission algorithm, whilst \"Retries\" and \"Timeout\" determine the maximum number of retries and the time before the first retry."

proc	snmptcl_callback_select_contexts	{{w .contexts}} {
    global	snmptcl_myname snmptcl_icon
    global	snmptcl_agent_ snmptcl_agent_cookie snmptcl_context
    global	snmptcl_retries snmptcl_rtrqos snmptcl_secqos snmptcl_timeout

    snmptcl_proc_mkmsg	""

    catch { destroy $w }
    toplevel	$w

    wm title	$w		"select a management context"
    wm iconname	$w		"select"
    wm iconbitmap \
		$w		@$snmptcl_icon

    if {![catch { $snmptcl_agent config active }]} {
	set	snmptcl_context	[$snmptcl_agent config name]
	set	snmptcl_retries	[$snmptcl_agent config retries]
	set	snmptcl_rtrqos	[$snmptcl_agent config rtrqos]
	set	snmptcl_secqos	[$snmptcl_agent config secqos]
	set	snmptcl_timeout	[$snmptcl_agent config timeout]
    }

    set		snmptcl_agent_	agent
    while {![catch { $snmptcl_agent_$snmptcl_agent_cookie config active }]} {
	incr	snmptcl_agent_cookie
    }
    append	snmptcl_agent_	$snmptcl_agent_cookie
    incr	snmptcl_agent_cookie

    frame	$w.frame	-borderwidth 10
    if {[llength [set l [snmpinfo ctxlist]]] < 1} {
	label	$w.frame.context	-text "Context: "
    } else {
	menubutton \
		$w.frame.context	-menu $w.frame.context.menu \
					-text "Context: "
	menu	$w.frame.context.menu
    }
    pack append	$w.frame	$w.frame.context \
				    {left} \
				[entry $w.frame.entry \
					-borderwidth 2 \
					-font -*-courier-medium-r-normal-*-120-* \
					-relief sunken \
					-textvariable snmptcl_context \
					-width 60] \
				    {left}

    foreach	k		$l {
	if {![lindex $k 2]} {
	    continue
	}
	set	label		[lindex $k 0]
	set	suffix		""
	if {[lindex $k 1] == ""} {
	    set	suffix		"*"
	}

	$w.frame.context.menu add command \
				-command "set snmptcl_context [lindex $k 0]; snmptcl_callback_select_contexts_aux; destroy $w" \
				-label $label$suffix
	if {$label == $snmptcl_context} {
	    $w.frame.context.menu entryconfigure last -underline 0
	}
    }

    bind	$w.frame.entry	<Return>	"$w.bot.select flash; $w.bot.select configure -state active; $w.bot.select invoke"
    bind	$w		<Any-Enter>	"focus $w.frame.entry"

    frame	$w.secqos	-borderwidth 3
    pack append	$w.secqos	[label $w.secqos.lb \
					-text "Security QOS"] \
					{top fillx} \
				[radiobutton $w.secqos.none \
					-relief flat \
					-text "None       " \
					-value none \
					-variable snmptcl_secqos] \
				    {top pady 4 expand} \
				[radiobutton $w.secqos.auth \
					-relief flat \
					-text "Auth only" \
					-value auth \
					-variable snmptcl_secqos] \
				    {top pady 4 expand} \
				[radiobutton $w.secqos.priv \
					-relief flat \
					-text "Auth+Priv" \
					-value authpriv \
					-variable snmptcl_secqos] \
				    {top pady 4 expand}

    frame	$w.rtrqos	-borderwidth 3
    pack append	$w.rtrqos	[label $w.rtrqos.lb \
					-text "Retransmission QOS"] \
					{top filly} \
				[radiobutton $w.rtrqos.exponential \
					-relief flat \
					-text "Exponential" \
					-value exponential \
					-variable snmptcl_rtrqos] \
				    {top pady 4 expand} \
				[radiobutton $w.rtrqos.superlinear \
					-relief flat \
					-text "Superlinear" \
					-value superlinear \
					-variable snmptcl_rtrqos] \
				    {top pady 4 expand}

    frame	$w.params	-borderwidth 3
    pack append	$w.params	[scale $w.params.retries \
					-command "set snmptcl_retries" \
					-from 0 -to 10 \
					-label Retries \
					-orient horizontal] \
				    {top pady 4 expand} \
				[scale $w.params.timeout \
					-command "set snmptcl_timeout" \
					-from 1 -to 30 \
					-label "Timeout (secs)" \
					-orient horizontal] \
				    {top pady 4 expand}
    $w.params.retries set $snmptcl_retries
    $w.params.timeout set $snmptcl_timeout

    frame	$w.bot		-borderwidth 5
    pack append	$w.bot		[button	$w.bot.select \
					-command "snmptcl_callback_select_contexts_aux; destroy $w" \
					-text Select] \
				    {left expand padx 20 pady 5} \
				[button	$w.bot.help \
					-command "snmptcl_callback_help .helpsel \"Help Selecting a Management Context\" \"select help\" select" \
					-text Help] \
				    {left expand padx 20 pady 5} \
				[button	$w.bot.cancel \
					-command "destroy $w" \
					-text Cancel] \
				    {left expand padx 20 pady 5}

    pack append	$w		$w.frame \
				    {top expand filly} \
				$w.rtrqos \
				    {left filly} \
				$w.params \
				    {left filly} \
				$w.secqos \
				    {left filly} \
				$w.bot \
				    {right fill}
}

set	snmptcl_var_sysObjectID	[snmpinfo mibprop sysObjectID value]
set	snmptcl_var_sysUpTime	[snmpinfo mibprop sysUpTime value]
set	snmptcl_var_snmpORID	[snmpinfo mibprop snmpORID value]

proc	snmptcl_callback_select_contexts_aux	{{agent {}}} {
    global	snmptcl_agent snmptcl_agent_ snmptcl_context
    global	snmptcl_retries snmptcl_rtrqos snmptcl_secqos snmptcl_timeout
    global	snmptcl_var_sysObjectID snmptcl_var_snmpORID

    snmptcl_proc_mkmsg	""

    if {$agent != ""} {
# puts stdout "pre-defined $agent"
	if {![catch { $agent destroy }]} {
# puts stdout "    destroyed"
	    snmptcl_callback_destroy_contexts
	}

	set	snmptcl_agent	$agent
    } else {
	foreach	pHandle	[snmpinfo proclist] {
	    if {("$snmptcl_context" == [$pHandle config name]) \
		    && ($snmptcl_retries == [$pHandle config retries]) \
		    && ($snmptcl_rtrqos == [$pHandle config rtrqos]) \
		    && ($snmptcl_secqos == [$pHandle config secqos])} {
		set	agent	$pHandle
		break
	    }
	}
	if {$agent != ""} {
# puts stdout "specification matches $agent"
	    set		snmptcl_agent	$agent
	    if {[$agent config resources] == ""} {
# puts stdout "    which needs resources"
		snmptcl_proc_getresources	$agent
	    }
	    return
	}

	set	snmptcl_agent	[set agent $snmptcl_agent_]
# puts stdout "new agent $snmptcl_agent_"
    }

    if {[catch { newsnmp $agent $snmptcl_context \
		    -retries $snmptcl_retries \
		    -rtrqos $snmptcl_rtrqos \
		    -secqos $snmptcl_secqos \
		    -timeout $snmptcl_timeout } result]} {
	snmptcl_dialog	Error $result error 0 OK
	return
    }
    snmptcl_callback_destroy_contexts
    snmptcl_proc_getresources	$agent
}

proc	snmptcl_proc_getresources	{agent} {
    global	snmptcl_var_sysObjectID snmptcl_var_snmpORID

    $agent	config		resources ""

    $agent	bulk		snmptcl_callback_getresources \
				    [list [list $snmptcl_var_sysObjectID] \
			      	    [list $snmptcl_var_snmpORID]] \
				    -maxrept 10 -nonrptr 1
}

proc	snmptcl_callback_getresources {pHandle iHandle eStatus eIndex vBinds} {
    global	snmptcl_var_sysObjectID snmptcl_var_snmpORID
    global	snmptcl_var_sysUpTime

    switch -- $eStatus {
	noError {
	}

	noSuchName {
	    if {$eIndex > 1} {
		$pHandle	next	snmptcl_callback_getresources \
					[list [list $snmptcl_var_sysObjectID] \
				              [list $snmptcl_var_sysUpTime]]
	    }

	    return
	}

	default {
	    return
	}
    }
    if {$eStatus != "noError"} {
	return
    }

    set		reslist		""
    set		lastvar		""
    foreach	vb		$vBinds {
	if {([set syntax [lindex $vb 1]] == "exception") \
		|| ($syntax == "error")} {
	    continue
	}

	set	var		[lindex $vb 0]
	if {("$snmptcl_var_sysObjectID.0" == $var) \
		|| ([snmptcl_proc_oidcontains $snmptcl_var_snmpORID $var])} {
	    lappend	reslist	[lindex $vb 2]
	} else {
	    set		lastvar	""
	    break
	}
	set	lastvar		$var
    }

    set		resources	[$pHandle config resources]
    foreach	res		$reslist {
	if {[lsearch -exact $resources $res] < 0} {
	    lappend	resources	$res
	}
    }
    $pHandle	config		resources $resources

    if {$lastvar != ""} {
	$pHandle	bulk	snmptcl_callback_getresources \
				    [list [list $lastvar]] \
				    -maxrept 10 -nonrptr 0
    }
}

proc	snmptcl_callback_destroy_contexts	{} {
    set		menu		.title.contexts.menu.destroy

    if {[info commands $menu] == ""} {
# puts stdout "$menu doesn't exist"
	return
    }

    $menu	delete		0 last
    foreach	agent		[lsort [snmpinfo proclist]] {
# puts stdout "adding $agent to destruction list"
	$menu	add command	-label [$agent config name] \
				-command "$agent destroy; snmptcl_callback_destroy_contexts"
    }
    if {[$menu index last] == "none"} {
	set	state		disabled
    } elseif {[set state [lindex [.title.contexts.menu entryconfigure 1 \
				      -state] 4]] == "disabled"} {
	set	state		normal
    }
# puts stdout "setting state to $state"
    .title.contexts.menu \
		entryconfigure	1 -state $state
}


#
# Callback: Help with snmptcl
#

proc	snmptcl_callback_help	{w title icon msg} {
    global	snmptcl_myname snmptcl_icon
    global	snmptcl_help

    snmptcl_proc_mkmsg	""

    catch { destroy $w }
    toplevel	$w

    wm title	$w		"$title"
    wm iconname	$w		"$icon"
    wm iconbitmap \
		$w		@$snmptcl_icon

    frame	$w.frame	-borderwidth 10
    pack append	$w.frame	[message $w.frame.msg \
					-font -*-courier-medium-r-normal-*-120-* \
					-text $snmptcl_help($msg)] \
				    {expand}

    frame	$w.bot		-borderwidth 1
    pack append	$w.bot		[button	$w.bot.button \
					-command "destroy $w" \
					-text Dismiss] \
				    {top padx 5 pady 5 expand}

    pack append	$w		$w.frame \
				    {top expand filly} \
				$w.bot \
				    {bottom fill}
}


#
# Callback: Options using snmptcl
#

set	snmptcl_help(options)	\
"The \"Miscellany\" options are used for debugging purposes.\n\nThe \"Verbose\" option is specific to each application.\n\nThe \"Watch\" option controls whether PDU traffic is displayed on your terminal.\n\nThe \"Debug\" option is of interest only to the implementor of $snmptcl_myname."

proc	snmptcl_callback_options		{{w .options}} {
    global	snmptcl_myname snmptcl_icon

    snmptcl_proc_mkmsg	""

    catch { destroy $w }
    toplevel	$w

    wm title	$w		"options using $snmptcl_myname"
    wm iconname	$w		"options"

    wm iconbitmap \
		$w		@$snmptcl_icon

    frame	$w.frame	-borderwidth 10

    pack append $w.frame	[frame $w.frame.misc -borderwidth 5] \
				    {top pady 4 expand frame w}

    pack append	$w.frame.misc \
				[label $w.frame.misc.lb \
					-text Miscellany \
					-width 12] \
				    {left filly} \
				[checkbutton $w.frame.misc.verbose \
					-relief flat \
					-text "Verbose       " \
					-variable snmptcl_verbose] \
				    {left padx 4 expand} \
				[checkbutton $w.frame.misc.watch \
					-relief flat \
					-text "Watch         " \
					-variable snmptcl_watch] \
				    {left padx 4 expand} \
				[checkbutton $w.frame.misc.debug \
					-relief flat \
					-text Debug \
					-variable snmptcl_debug] \
				    {left padx 4 expand}

    frame	$w.bot		-borderwidth 1
    pack append	$w.bot		[button	$w.bot.help \
					-command "snmptcl_callback_help .helpopt \"Help with $snmptcl_myname Options\" \"option help\" options" \
					-text Help] \
				    {left expand padx 20 pady 5} \
				[button	$w.bot.dismiss \
					-command "destroy $w" \
					-text Dismiss] \
				    {left expand padx 20 pady 5}

    pack append	$w		$w.frame \
				    {top expand filly} \
				$w.bot \
				    {bottom fill}
}


#
# displays a one-line message in the title-bar
#
#	    msg - message to display
#

set	snmptcl_armed	0
set	snmptcl_busy	0
set	snmptcl_ignore	0

proc	snmptcl_proc_mkmsg		{msg} {
    global	snmptcl_armed snmptcl_ignore

    if {$snmptcl_ignore == 1} {
	set	snmptcl_ignore	0
	return
    }
    incr	snmptcl_armed
    if {$msg == ""} {
	catch { .subtitle.msg configure -background bisque1 }
	.subtitle.msg configure	-text ""
    } else {
	catch { .subtitle.msg configure -background bisque3 }
	.subtitle.msg configure	-text $msg
	after	30000		"snmptcl_proc_mkmsg_aux $snmptcl_armed"
    }
}

proc	snmptcl_proc_mkmsg_aux	{serial} {
    global	snmptcl_armed snmptcl_busy

    if {$serial == $snmptcl_armed} {
	if {$snmptcl_busy} {
	    after 30000	"snmptcl_proc_mkmsg_aux $snmptcl_armed"
	} else {
	    snmptcl_proc_mkmsg	""
	}
    }
}


#
# just like tk_dialog, but creates a unique window...
#
# use of bitmaps for tk_dialog
#
#	  error - Tcl/Tk error
#	   info - fyi
#      question - user-input required
#     questhead - application error
#       warning - agent error
#  

set	snmptcl_dialog_cookie	0

proc	snmptcl_dialog {title text bitmap default args} {
    global	snmptcl_dialog_cookie

    set		w	.error
    while {[winfo exists $w$snmptcl_dialog_cookie]} {
	incr	snmptcl_dialog_cookie
    }
    append	w	$snmptcl_dialog_cookie
    incr	snmptcl_dialog_cookie
    
    return	[eval tk_dialog [list $w $title $text $bitmap $default] $args]
}


#
# compares two OIDs lexicographically, returning
#
#	a < b: -1
#	a = b:  0
#	a > b:  1
#

if {[info commands snmptcl_proc_oidcmp] == ""} {
proc	snmptcl_proc_oidcmp	{a b} {
    set		i	[llength [set a [split $a "."]]]
    set		j	[llength [set b [split $b "."]]]

    for {set k 0} {$i>0} {incr k} {
	if {$j == 0} {
	    return	1
	}
	if {[set al [string length [set ap [lindex $a $k]]]] > \
		[set bl [string length [set bp [lindex $b $k]]]]} {
	    return	1
	} elseif {$al < $bl} {
	    return     -1
	} elseif {[set c [string compare $ap $bp]] != 0} {
	    return	$c
	}

	incr	i      -1
	incr	j      -1
    }
    if {$j == 0} {
	return	0
    }
    return     -1
}}


#
# sees if one OID is a proper subtree of another
#
#

if {[info commands snmptcl_proc_oidcontains] == ""} {
proc	snmptcl_proc_oidcontains	{parent child} {
    if {[string match $parent.* $child]} {
	return	1
    }
    if {"$parent" == "$child"} {
	return	1
    }
    return	0
}}


#
# tries to make sense of an OID
#

if {[info commands snmptcl_proc_oid2ode] == ""} {
proc	snmptcl_proc_oid2ode	{oid} {
    if {[catch { set full [snmpinfo mibprop $oid full] }]} {
	return		$oid
    }

    set		i	[llength [set elemi [split $oid "."]]]
    set		j	[llength [set elemj [split [lindex $full 2] "."]]]

    set		descr	[lindex $full 0]
    for {} {$j < $i} {incr j} {
	append	descr	"."	[lindex $elemi $j]
    }

    return	$descr
}}


#
# Finally...
#

if {$argc > 0} {
    set		snmptcl_context		[lindex $argv 0]
    snmptcl_callback_select_contexts_aux
    
    if {[catch { $snmptcl_agent config active }]} {
	exit	1
    }
#    $snmptcl_agent	wait

    incr	argc	-1
    set		argv	[lrange $argv 1 end]
} else {
    snmptcl_callback_destroy_contexts
}


return
