/* unixd.c - daemon for UNIX MIB */

#ifndef	lint
static char *rcsid = "$Header$";
#endif

/* 
 * $Header$
 *
 * Originally contributed by NYSERNet Inc.  This work was partially supported
 * by the U.S. Defense Advanced Research Projects Agency and the Rome Air
 * Development Center of the U.S. Air Force Systems Command under contract
 * number F30602-88-C-0016.
 *
 *
 * $Log$
 */

/*
 *			  ISODE 8.0 NOTICE
 *
 *   Acquisition, use, and distribution of this module and related
 *   materials are subject to the restrictions of a license agreement.
 *   Consult the Preface in the User's Manual for the full terms of
 *   this agreement.
 *
 *
 *			4BSD/ISODE SNMPv2 NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions given in the file
 *    SNMPv2-READ-ME.
 *
 */


#include <signal.h>
#include <stdio.h>
#include <varargs.h>
#include "objects.h"
#include "smux.h"
#include "snmp-g.h"
#include "sys.file.h"
#include "tailor.h"

/*    DATA */

	int	debug = 0;
static	int	ddebug = 0;
static	int	nbits = FD_SETSIZE;

static	char	sfile[BUFSIZ];
static	char   *spath;
static	char  **sargv;


static LLog	_pgm_log = {
    "unixd.log", NULLCP, NULLCP, LLOG_FATAL | LLOG_EXCEPTIONS | LLOG_NOTICE,
    LLOG_FATAL, -1, LLOGCLS | LLOGCRT | LLOGZER, NOTOK
};
static	LLog   *pgm_log = &_pgm_log;

static	char   *myname = "unixd";
static	char   *myobjectID = "fourBSDunixd.4";


static	int	snmp_fd = NOTOK;
static	int	bflag = 0;
static	char    *aParty = NULL;
static	char	*pParty = NULL;


int	init_users (), sync_users ();		/* users group */
int	init_print (), sync_print ();		/* print group */
int	init_unix_fs (), sync_unix_fs ();	/* file-system group */

static struct triple {
    char   *t_tree;
    OID	    t_name;
    IFP	    t_init;
    IFP	    t_sync;
}	triples[] = {
    "users",	NULL,	init_users,	sync_users,
    "print",	NULL,	init_print,	sync_print,
    "unixFs",	NULL,	init_unix_fs,	sync_unix_fs,

    NULL
};


void	adios (), advise ();

/*    MAIN */

/* ARGSUSED */

main (argc, argv, envp)
int	argc;
char  **argv,
      **envp;
{
    int	    nfds;
    fd_set  ifds,
	    ofds;

    arginit (argv);
    envinit ();

    FD_ZERO (&ifds);
    FD_ZERO (&ofds);
    nfds = 0;

/* set fd's for other purposes here... */

    for (;;) {
	int	action,
		secs;
	fd_set	rfds,
		wfds;

	secs = NOTOK;

	rfds = ifds;	/* struct copy */
	wfds = ofds;	/*   .. */

	if (snmp_fd != NOTOK) {
	    FD_SET (snmp_fd, &rfds);
	    if (snmp_fd >= nfds)
		nfds = snmp_fd + 1;
	}
	

	if (xselect (nfds, &rfds, &wfds, NULLFD, secs) == NOTOK)
	    adios ("failed", "xselect");

/* check fd's for other purposes here... */

	if (snmp_fd != NOTOK && FD_ISSET (snmp_fd, &rfds))
	    doit_snmp ();

	switch (action = *((integer *) *agentAction)) {
	    case AGENT_OTHER:
	        break;

	    case AGENT_COLDSTART:
	    case AGENT_TERMINATE:
		advise (LLOG_FATAL, NULLCP, "%s...",
			action == AGENT_COLDSTART ? "initiating coldStart"
						  : "I'll be back");

/* do program-related release of resources here... */

		(void) snmp_close ();
		if (action == AGENT_TERMINATE)
		    exit (0);
		(void) execv (spath, sargv);
		adios (spath, "unable to execv");
		/* NOTREACHED */

	    case AGENT_WARMSTART:
		switch (snmp_trap (NULLOID, int_SNMP_generic__trap_warmStart,
				   0, (struct type_SNMP_VarBindList *) 0)) {
		    case OK:
		        break;

		    case DONE:
			snmp_fd = NOTOK;
			/* and fall... */

		    case NOTOK:
		    default:
			advise (LLOG_EXCEPTIONS, NULLCP, "snmp_trap: %s",
				PY_pepy);
			break;
		}
		goto reset_action;

	    case AGENT_DEBUGON:
		debug = 1;
		pgm_log -> ll_events |= LLOG_ALL;
		pgm_log -> ll_stat |= LLOGTTY;
		goto reset_action;

	    case AGENT_DEBUGOFF:
		debug = 0;
		pgm_log -> ll_events = LLOG_FATAL | LLOG_EXCEPTIONS
					    | LLOG_NOTICE;
		pgm_log -> ll_stat = LLOGCLS | LLOGCRT | LLOGZER | LLOGHDR;
		goto reset_action;

	    default:
		advise (LLOG_NOTICE, NULLCP, "agentAction reset from %d",
			*((integer *) *agentAction));
reset_action: ;
		*((integer *) *agentAction) = AGENT_OTHER;
		break;
	}
    }
}

/*    MISCELLANY */

static	arginit (vec)
char	**vec;
{
    register char  *ap;

    if (myname = rindex (*vec, '/'))
	myname++;
    if (myname == NULL || *myname == NULL)
	myname = *vec;
    if (strncmp (myname, "snmp.", 5) == 0 && myname[5] != NULL)
	myname += 5;

    _tailor_logging_only = 1;
    isodetailor (myname, 0);
    ll_hdinit (pgm_log, myname);

    sargv = vec;
    if (*sargv[0] == '/')
	spath = sargv[0];
    else {
	int	i = strlen (sargv[0]);
	
	if (getcwd (sfile, sizeof sfile - (i + 1))) {
	    ap = sfile + strlen (sfile);
	    (void) sprintf (ap, "%s%s", ap > sfile + 1 ? "/" : "", sargv[0]);
	}
	else
	    (void) strcpy (sfile, _isodefile (isodesbinpath, myname));

	spath = sfile;
    }	

    for (vec++; ap = *vec; vec++) {
	if (*ap == '-')
	    switch (*++ap) {
		case 'a':
		    if ((aParty = *++vec) == NULL || *aParty == '-')
			adios (NULLCP, "usage: %s -a party", myname);
		    continue;

		case 'b':
		    if ((ap = *++vec) == NULL
			    || *ap == '-'
			    || (bflag = atoi (ap)) <= 0)
			adios (NULLCP, "usage: %s -b size", myname);
		    continue;

		case 'p':
                    if ((pParty = *++vec) == NULL || *pParty == '-')
                        adios (NULLCP, "usage: %s -p party-file-prefix",
			       myname);
		    continue;

		case 'd':
		    debug++;
		    /* and fall... */
		case 'D':
		    ddebug++;
		    continue;

		default: 
		    adios (NULLCP, "-%s: unknown switch", ap);
	    }

	adios (NULLCP, "usage: %s [switches]", myname);
    }
}

/*  */

static  envinit () {
    int     i,
            sd;
    char    file[BUFSIZ];
    FILE   *fp;

    nbits = getdtablesize ();

    if (debug == 0 && !(debug = isatty (2))) {
	for (i = 0; i < 5; i++) {
	    switch (fork ()) {
		case NOTOK: 
		    sleep (5);
		    continue;

		case OK: 
		    break;

		default: 
		    _exit (0);
	    }
	    break;
	}

	if (!ddebug)
	    (void) chdir ("/");

	if ((sd = open ("/dev/null", O_RDWR)) == NOTOK)
	    adios ("/dev/null", "unable to read");
	if (sd != 0)
	    (void) dup2 (sd, 0), (void) close (sd);
	(void) dup2 (0, 1);
	(void) dup2 (0, 2);

#ifdef	SETSID
	if (setsid () == NOTOK)
	    advise (LLOG_EXCEPTIONS, "failed", "setsid");
#endif
#ifdef	TIOCNOTTY
	if ((sd = open ("/dev/tty", O_RDWR)) != NOTOK) {
	    (void) ioctl (sd, TIOCNOTTY, NULLCP);
	    (void) close (sd);
	}
#else
#ifdef	SYS5
	(void) setpgrp ();
	(void) signal (SIGINT, SIG_IGN);
	(void) signal (SIGQUIT, SIG_IGN);
#endif
#endif
    }
    else
	ll_dbinit (pgm_log, myname);

#ifndef	sun		/* damn YP... */
    for (sd = 3; sd < nbits; sd++)
	if (pgm_log -> ll_fd != sd)
	    (void) close (sd);
#endif

    (void) signal (SIGPIPE, SIG_IGN);

    ll_hdinit (pgm_log, myname);

    o_advise = (IFP) advise;

    mibinit ();

    (void) sprintf (file, "/etc/%s.pid", myname);
    if (fp = fopen (file, "w")) {
	(void) fprintf (fp, "%d\n", getpid ());
	(void) fclose (fp);
    }
    
    advise (LLOG_NOTICE, NULLCP, "starting");
}

/*    MIB */

static  mibinit () {
    register struct triple *tc;
    OT	    ot;

    if (readobjects ("unixd.defs") == NOTOK)
	adios (NULLCP, "readobjects: %s", PY_pepy);

    for (tc = triples; tc -> t_tree; tc++)
	if (ot = text2obj (tc -> t_tree)) {
	    tc -> t_name = ot -> ot_name;
	    (void) (*tc -> t_init) ();
	}
	else
	    advise (LLOG_EXCEPTIONS, NULLCP, "text2obj (\"%s\") fails",
		    tc -> t_tree);

    if ((snmp_fd = snmp_init (myname, myobjectID, aParty, pParty, debug, 1,
			      bflag)) == NOTOK)
	advise (LLOG_EXCEPTIONS, NULLCP, "snmp_init: %s", PY_pepy);
    else
	switch (snmp_trap (NULLOID, int_SNMP_generic__trap_coldStart, 0,
			   (struct type_SNMP_VarBindList *) 0)) {
	    case OK:
	        break;

	    case DONE:
		snmp_fd = NOTOK;
		/* and fall... */
	    case NOTOK:
	    default:
		advise (LLOG_EXCEPTIONS, NULLCP, "snmp_trap: %s", PY_pepy);
		break;
	}
}

/*  */

static	doit_snmp () {
    struct type_SNMP_PDUs *event;

    switch (snmp_wait (&event, NOTOK)) {
	case OK:
	    if (event == NULL)
		return;
	    break;
	    
	case DONE:
	    snmp_fd = NOTOK;
	    /* and fall... */
	case NOTOK:
	default:
	    advise (LLOG_EXCEPTIONS, NULLCP, "snmp_wait: %s", PY_pepy);
	    return;
    }

    switch (event -> offset) {
	case type_SNMP_PDUs_get__request:
	case type_SNMP_PDUs_get__next__request:
	case type_SNMP_PDUs_set__request:
	    do_snmp (event -> un.get__request, event -> offset);
	    break;

	case type_SNMP_PDUs_response:
	case type_SNMP_PDUs_trap:
	    advise (LLOG_EXCEPTIONS, NULLCP, "unexpectedOperation: %d",
		    event -> offset);
	    goto losing;

	default:
	    advise (LLOG_EXCEPTIONS, NULLCP, "badOperation: %d",
		    event -> offset);
losing: ;
	    (void) snmp_close ();
	    snmp_fd = NOTOK;
	    break;
    }
}

/*  */

static	do_snmp (pdu, offset)
register struct type_SNMP_GetRequest__PDU *pdu;
int	offset;
{
    int	    idx,
	    status;
    object_instance ois;
    register struct type_SNMP_VarBindList *vp;
    IFP	    method;

    quantum++;
    idx = 0;
    for (vp = pdu -> variable__bindings; vp; vp = vp -> next) {
	register OI	oi;
	register OT	ot;
	register struct type_SNMP_VarBind *v = vp -> VarBind;

	idx++;

	if (offset == type_SNMP_PDUs_get__next__request) {
	    if ((oi = name2inst (v -> name)) == NULLOI
		    && (oi = next2inst (v -> name)) == NULLOI)
		goto no_name;

	    if ((ot = oi -> oi_type) -> ot_getfnx == NULLIFP)
		goto get_next;
	}
	else {
	    if ((oi = name2inst (v -> name)) == NULLOI)
		goto no_name;
	    ot = oi -> oi_type;
	    if ((offset == type_SNMP_PDUs_get__request
			    ? ot -> ot_getfnx : ot -> ot_setfnx) == NULLIFP) {
no_name: ;
		pdu -> error__status = int_SNMP_error__status_noSuchName;
		goto out;
	    }
	}

try_again: ;
	switch (offset) {
	    case type_SNMP_PDUs_get__request:
		if (!(method = ot -> ot_getfnx))
		    goto no_name;
		break;
		
	    case type_SNMP_PDUs_get__next__request:
		if (!(method = ot -> ot_getfnx))
		    goto get_next;
		break;
		
	    case type_SNMP_PDUs_set__request:
		if (!(method = ot -> ot_setfnx))
		    goto no_name;
		break;

	    default:
		goto no_name;
	}
		
	switch (status = (*method) (oi, v, offset)) {
	    case NOTOK:	    /* get-next wants a bump */
get_next: ;
		oi = &ois;
		for (;;) {
		    if ((ot = ot -> ot_next) == NULLOT) {
			pdu -> error__status =
					    int_SNMP_error__status_noSuchName;
			goto out;
		    }
		    oi -> oi_name = (oi -> oi_type = ot) -> ot_name;
		    if (ot -> ot_getfnx)
			goto try_again;
		}

	    case int_SNMP_error__status_noError:
		break;

	    default:
		pdu -> error__status = status;
		goto out;
	}
    }
    idx = 0;

out: ;
    pdu -> error__index = idx;

    if (offset == type_SNMP_PDUs_set__request) {
	int	cor = pdu -> error__status ? type_SNMP_PDUs_rollback
					   : type_SNMP_PDUs_commit;
	register struct triple *tc;

	for (vp = pdu -> variable__bindings; vp; vp = vp -> next) {
	    register OI	    oi;
	    register struct type_SNMP_VarBind *v = vp -> VarBind;

	    if ((oi = name2inst (v -> name))
		    && (method = oi -> oi_type -> ot_setfnx) == s_generic)
		(void) (*method) (oi, v, cor);
	}

	for (tc = triples; tc -> t_tree; tc++)
	    if (tc -> t_name)
		(void) (*tc -> t_sync) (cor);
    }

    switch (snmp_response ()) {
	case OK:
	    break;
	    
	case DONE:
	    snmp_fd = NOTOK;
	    /* and fall... */
	case NOTOK:
	default:
	    advise (LLOG_EXCEPTIONS, NULLCP, "snmp_response: %s", PY_pepy);
	    break;
    }
}

/*    ERRORS */

#ifndef	lint
void	adios (va_alist)
va_dcl
{
    va_list ap;

    va_start (ap);
    
    _ll_log (pgm_log, LLOG_FATAL, ap);

    va_end (ap);

    _exit (1);
}
#else
/* VARARGS */

void	adios (what, fmt)
char   *what,
       *fmt;
{
    adios (what, fmt);
}
#endif


#ifndef	lint
void	advise (va_alist)
va_dcl
{
    int	    code;
    va_list ap;

    va_start (ap);
    
    code = va_arg (ap, int);

    _ll_log (pgm_log, code, ap);

    va_end (ap);
}
#else
/* VARARGS */

void	advise (code, what, fmt)
char   *what,
       *fmt;
int	code;
{
    advise (code, what, fmt);
}
#endif
